##
 # Makefile
 # Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 # 
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
 # the Free Software Foundation, either version 3 of the License, or
 # (at your option) any later version.
 # 
 # This program is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 # GNU General Public License for more details.
 # 
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ##
include config.mak
include $(SRCDIR)/in_ffsox/version.mak

###############################################################################
.PHONY: all

###############################################################################
SFURL=http://sourceforge.net/projects/$1/files/$2/download

define DOWNLOAD
  mkdir -p $(@D)
  wget -O $@ $1
	touch $@
endef

###############################################################################
#MODES+=full
MODES+=tiny

###############################################################################
### LIB1770_TARGETS
LIB1770_TARGETS+=$(PFX)/lib/lib1770.a
LIB1770_TARGETS+=$(PFX)/include/bs1770.h

### LIBPB_TARGETS
LIBPB_TARGETS+=$(PFX)/lib/libpb.a
LIBPB_TARGETS+=$(PFX)/include/pb.h

### FFMPEG_TARGETS
AVUTIL=avutil-$(AVUTIL_VER)
SWRESAMPLE=swresample-$(SWRESAMPLE_VER)
SWSCALE=swscale-$(SWSCALE_VER)
POSTPROC=postproc-$(POSTPROC_VER)
AVCODEC=avcodec-$(AVCODEC_VER)
AVFORMAT=avformat-$(AVFORMAT_VER)
AVFILTER=avfilter-$(AVFILTER_VER)
FFMPEG_DLLS+=$(AVUTIL).dll
FFMPEG_DLLS+=$(SWRESAMPLE).dll
FFMPEG_DLLS+=$(SWSCALE).dll
FFMPEG_DLLS+=$(POSTPROC).dll
FFMPEG_DLLS+=$(AVCODEC).dll
FFMPEG_DLLS+=$(AVFORMAT).dll
FFMPEG_DLLS+=$(AVFILTER).dll
FFMPEG_TARGETS_full+=$(PFX)/full/bin
FFMPEG_TARGETS_full+=$(PFX)/full/include/libswresample/swresample.h
FFMPEG_TARGETS_full+=$(PFX)/full/include/libswscale/swscale.h
FFMPEG_TARGETS_full+=$(PFX)/full/include/libpostproc/postprocess.h
FFMPEG_TARGETS_full+=$(PFX)/full/include/libavcodec/avcodec.h
FFMPEG_TARGETS_full+=$(PFX)/full/include/libavformat/avformat.h
FFMPEG_TARGETS_full+=$(PFX)/full/include/libavfilter/avfilter.h
FFMPEG_TARGETS_full+=$(patsubst %.dll,$(PFX)/full/lib/%.def,$(FFMPEG_DLLS))
FFMPEG_TARGETS_full+=$(patsubst %,$(PFX)/full/bin/%,$(FFMPEG_DLLS))
FFMPEG_TARGETS_tiny=$(patsubst $(PFX)/full/%,$(PFX)/tiny/%,\
		$(FFMPEG_TARGETS_full))
#
FFMPEG_DELAYLOAD_full+=$(patsubst %,$(PFX)/full/bin/delayload/%.a,\
		$(FFMPEG_DLLS))
#FFMPEG_DELAYLOAD_full+=$(patsubst %,$(PFX)/full/bin/delayload/%,\
#		$(FFMPEG_DLLS))
FFMPEG_DELAYLOAD_tiny=$(patsubst $(PFX)/full/%,$(PFX)/tiny/%,\
		$(FFMPEG_DELAYLOAD_full))

### OPUS_TARGETS
OPUS_TARGETS+=$(PFX)/lib/libopus.a
OPUS_TARGETS+=$(PFX)/include/opus/opus.h
OPUS_TARGETS+=$(PFX)/include/opus/opus_defines.h
OPUS_TARGETS+=$(PFX)/include/opus/opus_types.h
OPUS_TARGETS+=$(PFX)/include/opus/opus_multistream.h

### SOX_TARGETS
SOX_TARGETS=$(PFX)/lib/libsox.a
SOX_TARGETS=$(PFX)/include/sox.h
SOX_DELAYLOAD+=$(PFX)/bin/delayload/libsox-$(LIBSOX_VER).dll
SOX_DELAYLOAD+=$(PFX)/bin/delayload/libsox-$(LIBSOX_VER).dll.a

###############################################################################
all: $(LIB1770_TARGETS)
all: $(LIBPB_TARGETS)
ifneq (,$(filter full,$(MODES))) # [
all: $(FFMPEG_DELAYLOAD_full)
all: $(FFMPEG_TARGETS_full)
endif # ]
ifneq (,$(filter tiny,$(MODES))) # [
all: $(FFMPEG_DELAYLOAD_tiny)
all: $(FFMPEG_TARGETS_tiny)
endif # ]
all: $(SOX_DELAYLOAD)
all: $(SOX_TARGETS)

### LIB1770 ###################################################################
LIB1770=lib1770
LIB1770_DEPS+=$(SRCDIR)/lib1770/source.mak
LIB1770_DEPS+=$(LIBPB_TARGETS)
LIB1770_OPTS+=--prefix=$(PWD)

$(LIB1770_TARGETS): $(patsubst %,$(SRCDIR)/lib1770/%,$(LIB1770_SRCS))
$(LIB1770_TARGETS): $(PFX)/build/$(LIB1770)/Makefile
	cd $(<D) && $(MAKE) && $(MAKE) install
	touch -r $< $@
$(PFX)/build/$(LIB1770)/Makefile: $(LIB1770_DEPS)
$(PFX)/build/$(LIB1770)/Makefile: $(SRCDIR)/lib1770/Makefile
$(PFX)/build/$(LIB1770)/Makefile: OPTS=$(LIB1770_OPTS)

### LIBPB #####################################################################
LIBPB=libpb
LIBPB_DEPS+=$(SRCDIR)/$(LIBPB)/source.mak
LIBPB_OPTS+=--prefix=$(PWD)

$(LIBPB_TARGETS): $(patsubst %,$(SRCDIR)/$(LIBPB)/%,$(LIBPB_SRCS))
$(LIBPB_TARGETS): $(PFX)/build/$(LIBPB)/Makefile
	cd $(<D) && $(MAKE) && $(MAKE) install
	touch -r $< $@
$(PFX)/build/$(LIBPB)/Makefile: $(LIBPB_DEPS)
$(PFX)/build/$(LIBPB)/Makefile: $(SRCDIR)/$(LIBPB)/Makefile
$(PFX)/build/$(LIBPB)/Makefile: OPTS=$(LIBPB_OPTS)

### FFMPEG ####################################################################
FFMPEG=ffmpeg
FFMPEG_DEPS+=$(PFX)/lib/libpthread.a
FFMPEG_DEPS+=$(PFX)/lib/libiconv.a
FFMPEG_DEPS+=$(PFX)/lib/libbz2.a
FFMPEG_DEPS+=$(PFX)/lib/libz.a
FFMPEG_DEPS+=$(PFX)/lib/liblzma.a
FFMPEG_DEPS+=$(OPUS_TARGETS)

########
FFMPEG_OPTS_full+=--prefix=$(PFX)/full
FFMPEG_OPTS_full+=--enable-gpl
FFMPEG_OPTS_full+=--enable-version3
FFMPEG_OPTS_full+=--enable-shared
FFMPEG_OPTS_full+=--disable-doc
FFMPEG_OPTS_full+=--disable-w32threads
FFMPEG_OPTS_full+=--disable-os2threads
FFMPEG_OPTS_full+=--enable-libopus
FFMPEG_OPTS_full+='--extra-ldflags=-L$(PFX)/lib'

########
FFMPEG_OPTS_tiny+=$(patsubst --prefix=$(PFX)/full,--prefix=$(PFX)/tiny,\
		$(FFMPEG_OPTS_full))
FFMPEG_OPTS_tiny+=--disable-encoders
FFMPEG_OPTS_tiny+=--disable-decoders
FFMPEG_OPTS_tiny+=--disable-muxers
FFMPEG_OPTS_tiny+=--disable-demuxers
FFMPEG_OPTS_tiny+=--disable-hwaccels
FFMPEG_OPTS_tiny+=--disable-parsers
#FFMPEG_OPTS_tiny+=--disable-filter=elbg
#FFMPEG_OPTS_tiny+=--disable-parser=hevc
# ???
#FFMPEG_OPTS_tiny+=--enable-dct
FFMPEG_OPTS_tiny+=--disable-dxva2
# decoders
FFMPEG_OPTS_tiny+=--enable-decoder=pcm_s16le
FFMPEG_OPTS_tiny+=--enable-decoder=pcm_s24le
FFMPEG_OPTS_tiny+=--enable-decoder=pcm_dvd
FFMPEG_OPTS_tiny+=--enable-decoder=flac
FFMPEG_OPTS_tiny+=--enable-decoder=wavpack
FFMPEG_OPTS_tiny+=--enable-decoder=vorbis
FFMPEG_OPTS_tiny+=--enable-decoder=libopus
FFMPEG_OPTS_tiny+=--enable-decoder=vp3
FFMPEG_OPTS_tiny+=--enable-decoder=vp5
FFMPEG_OPTS_tiny+=--enable-decoder=vp6
FFMPEG_OPTS_tiny+=--enable-decoder=vp6a
FFMPEG_OPTS_tiny+=--enable-decoder=vp6f
FFMPEG_OPTS_tiny+=--enable-decoder=vp7
FFMPEG_OPTS_tiny+=--enable-decoder=vp8
FFMPEG_OPTS_tiny+=--enable-decoder=vp9
FFMPEG_OPTS_tiny+=--enable-decoder=theora
# demuxers
FFMPEG_OPTS_tiny+=--enable-demuxer=pcm_s16le
FFMPEG_OPTS_tiny+=--enable-demuxer=pcm_s24le
FFMPEG_OPTS_tiny+=--enable-demuxer=flac
FFMPEG_OPTS_tiny+=--enable-demuxer=wav
FFMPEG_OPTS_tiny+=--enable-demuxer=ogg
FFMPEG_OPTS_tiny+=--enable-demuxer=matroska
# parsers
FFMPEG_OPTS_tiny+=--enable-parser=png
FFMPEG_OPTS_tiny+=--enable-parser=vorbis
FFMPEG_OPTS_tiny+=--enable-parser=vp3
FFMPEG_OPTS_tiny+=--enable-parser=vp8
FFMPEG_OPTS_tiny+=--enable-parser=vp9

#.PRECIOUS: $(PFX)/$(1)/bin/delayload/%.dll
#$(PFX)/$(1)/bin/delayload/%.dll: $(PFX)/$(1)/lib/%.def
#	lib=`echo $$(<F) | sed -e "s/\(.*\)-.*.def/lib\1/g"`; \
#	gcc -shared -o $$@ $< $(PFX)/$(1)/build/ffmpeg/$$$${lib}/*.o $(PFX)/$(1)/build/ffmpeg/$$$${lib}/x86/*.o
define FFMPEG_BUILD # [
.PRECIOUS: $(PFX)/$(1)/bin/delayload/%.dll.a
$(PFX)/$(1)/bin/delayload/%.dll.a: $(PFX)/$(1)/bin/delayload/%.def
	dlltool -d $$< --output-delaylib $$@
.PRECIOUS: $(PFX)/$(1)/bin/delayload/%.def
$(PFX)/$(1)/bin/delayload/%.def: $(PFX)/$(1)/lib/%.def
	mkdir -p $$(@D)
	echo "LIBRARY `basename $$< .def`.dll" > $$@
	cat $$< >> $$@
$(FFMPEG_TARGETS_$(1)): $(PFX)/$(1)/build/$(FFMPEG)/Makefile
	cd $$(<D) && $(MAKE) && $(MAKE) install
$(PFX)/$(1)/build/$(FFMPEG)/Makefile: $(FFMPEG_DEPS)
$(PFX)/$(1)/build/$(FFMPEG)/Makefile: OPTS=$(FFMPEG_OPTS_$(1))
endef # ]

$(foreach mode,$(MODES),$(eval $(call FFMPEG_BUILD,$(mode))))

$(TOOLDIR)/ffmpeg-snapshot.tar.bz2:
	$(call DOWNLOAD,http://ffmpeg.org/releases/$(@F))

###############################################################################
OPUS:=opus-1.2.1

OPUS_DEPS=

OPUS_OPTS+=--prefix=$(PWD)
OPUS_OPTS+=--disable-shared

$(OPUS_TARGETS): $(PFX)/build/$(OPUS)/Makefile
	cd $(<D) && $(MAKE) && $(MAKE) install
$(PFX)/build/$(OPUS)/Makefile: $(OPUS_DEPS)
$(PFX)/build/$(OPUS)/Makefile: OPTS=$(OPUS_OPTS)

$(TOOLDIR)/$(OPUS).tar.gz:
	$(call DOWNLOAD,http://downloads.xiph.org/releases/opus/$(@F))

### SOX #######################################################################
SOXVER:=14.4.2
SOX:=sox-$(SOXVER)
SOX_DEPS:=$(PFX)/lib/libpthread.a
SOX_DEPS+=$(PFX)/lib/libgomp.a
SOX_DEPS+=$(PFX)/lib/libgcc.a

SOX_CFLAGS+=-static-libgcc
SOX_CFLAGS+=-shared
SOX_LDFLAGS+=-L$(PFX)/lib
SOX_LDLIBS+=-lgomp
SOX_LDLIBS+=-lpthread
SOX_LDLIBS+=-lwinmm

SOX_OPTS:=--prefix=$(PWD)
SOX_OPTS+=--without-ladspa
SOX_OPTS+=--without-png
SOX_OPTS+=--with-ffmpeg=no
SOX_OPTS+='LDFLAGS=-static-libgcc -L$(PWD)/lib'

SOX_BUILD_TARGETS+=$(PFX)/build/$(SOX)/src/.libs/libsox-$(LIBSOX_VER).dll.def
SOX_BUILD_TARGETS+=$(PFX)/build/$(SOX)/src/.libs/libsox.a

$(PFX)/bin/delayload/libsox-$(LIBSOX_VER).dll: $(SOX_BUILD_TARGETS)
	mkdir -p $(@D)
	gcc $(SOX_CFLAGS) -o $@ $^ $(SOX_LDFLAGS) $(SOX_LDLIBS)
$(PFX)/bin/delayload/libsox-$(LIBSOX_VER).dll.a: \
		$(PFX)/bin/delayload/libsox-$(LIBSOX_VER).dll.def
	dlltool -d $< --output-delaylib $@
$(PFX)/bin/delayload/libsox-$(LIBSOX_VER).dll.def: \
		$(PFX)/build/$(SOX)/src/.libs/libsox-$(LIBSOX_VER).dll.def
	mkdir -p $(@D)
	echo "LIBRARY `basename $(@F) .def`" > $@
	cat $< >> $@
$(SOX_TARGETS): $(SOX_BUILD_TARGETS)
$(SOX_TARGETS): $(PFX)/build/$(SOX)/Makefile
	cd $(<D) && $(MAKE) install
$(SOX_BUILD_TARGETS): $(PFX)/build/$(SOX)/Makefile
	cd $(<D) && $(MAKE)
$(PFX)/build/$(SOX)/Makefile: $(SOX_DEPS)
$(PFX)/build/$(SOX)/Makefile: OPTS=$(SOX_OPTS)

$(TOOLDIR)/$(SOX).tar.bz2:
	$(call DOWNLOAD,$(call SFURL,sox,sox/$(SOXVER)/$(@F)))

### CONFIGURE #################################################################
PFX_PATH:=$(PFX)/lib/pkgconfig
PFX_PATH_full:=$(PFX)/full/lib/pkgconfig:$(PFX_PATH)
PFX_PATH_tiny:=$(PFX)/tiny/lib/pkgconfig:$(PFX_PATH)

define CONFIGURE
  mkdir -p $(@D)
  export PKG_CONFIG_PATH=$1 && cd $(@D) && $< $(OPTS)
  touch $@
endef

.PRECIOUS: $(PFX)/build/%/Makefile
$(PFX)/build/%/Makefile: $(SRCDIR)/%/configure
	$(call CONFIGURE,$(PFX_PATH))
$(PFX)/full/build/%/Makefile: $(SRCDIR)/%/configure
	$(call CONFIGURE,$(PFX_PATH_full))
$(PFX)/tiny/build/%/Makefile: $(SRCDIR)/%/configure
	$(call CONFIGURE,$(PFX_PATH_tiny))
$(PFX)/build/%/Makefile: $(PFX)/unpack/%/configure
	$(call CONFIGURE,$(PFX_PATH))
$(PFX)/full/build/%/Makefile: $(PFX)/unpack/%/configure
	$(call CONFIGURE,$(PFX_PATH_full))
$(PFX)/tiny/build/%/Makefile: $(PFX)/unpack/%/configure
	$(call CONFIGURE,$(PFX_PATH_tiny))

### UNPACK ####################################################################
define UNPACK
  mkdir -p $(PFX)/unpack
  tar xfv$1 $< -C $(PFX)/unpack
  if test -e $(SRCDIR)/patch/$*.patch; then \
    cd $(PFX)/unpack; \
    patch -p0 -i $(SRCDIR)/patch/$*.patch; \
  fi
  touch $@
endef

# $(1) - mode
define REMOVE_MODE
	ffmpeg=$(filter ffmpeg%,$*) && if [ 0 -lt $${#ffmpeg} ]; then \
		rm -rf $(PFX)/$(1)/bin/delayload/*; \
	fi
	if [ -e $(PFX)/$(1)/build/$*/Makefile ]; then \
		cd $(PFX)/$(1)/build/$*; \
		make uninstall; \
	fi
	rm -rf $(PFX)/$(1)/build/$*/*
endef

define REMOVE
	sox=$(filter sox%,$*) && if [ 0 -lt $${#sox} ]; then \
		rm -rf $(PFX)/bin/delayload/*; \
	fi
	if [ -e $(PFX)/build/$*/Makefile ]; then \
		cd $(PFX)/build/$*; \
		make uninstall; \
	fi
	rm -rf $(PFX)/build/$*/*
endef

.PRECIOUS: $(PFX)/unpack/%/configure
$(PFX)/unpack/%/configure: $(TOOLDIR)/%-snapshot.tar.bz2
	$(foreach mode,$(MODES),$(call REMOVE_MODE,$(mode)))
	rm -rf $(PFX)/$(1)/unpack/$*/*
	$(call UNPACK,j)
$(PFX)/unpack/%/configure: $(TOOLDIR)/%.tar.bz2
	$(call REMOVE)
	rm -rf $(PFX)/unpack/$*/*
	$(call UNPACK,j)
$(PFX)/unpack/%/configure: $(TOOLDIR)/%.tar.gz
	$(call REMOVE)
	rm -rf $(PFX)/unpack/$*/*
	$(call UNPACK,z)

###############################################################################
.PRECIOUS: $(PFX)/lib/lib%.a
# We copy some static libraries local in order that during linking they will
# be found before their shared counterparts. In order to make this work we
# need to set up the linker's library search path accordingly.
$(PFX)/lib/lib%.a:
	mkdir -p $(@D)
	cp -p $(shell gcc -print-file-name=$(@F)) $(@D)
