/*
 * engine.c
 *
 * Copyright (C) 2019 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

#if defined (THREAD) // [
void EngineCreate(void *pe, const void *ps)
{
  ((SEngine *)pe)->pStateMin=ps;
  ((SEngine *)pe)->pState=ps;
}

void EngineDestroy(void *pe)
{
  EngineDown(pe,0);
}

int EngineState(void *pe)
{
  SEngine *pEngine=pe;

  return pe?pEngine->pState-pEngine->pStateMin:0;
}

const char *EngineStateIdEx(void *pe, int nState)
{
  SEngine *pEngine=pe;
  IEngineState *pStateMin=pEngine->pStateMin;
  IEngineState *pState=pEngine->pStateMin+nState;

  return pState<=pStateMin?"eEngineStart":(pState-1)->pId;
}

const char *EngineStateId(void *pe)
{
  SEngine *pEngine=pe;
  IEngineState *pState=pEngine->pState;
  IEngineState *pStateMin=pEngine->pStateMin;

  return EngineStateIdEx(pEngine,pState-pStateMin);
}

int EngineUp(void *pe, int nState)
{
  SEngine *pEngine=pe;
  IEngineState *pStateTarget=pEngine->pStateMin+nState;

  if (pEngine->pState<pStateTarget) {
    do {
      DVWRITELN("entering %s ...",pEngine->pState->pId);

      if (pEngine->pState->Enter(pEngine)<0) {
        DVMESSAGE("entering %s",pEngine->pState->pId);
        goto exit;
      }
    } while (pEngine->pState<pStateTarget);
  }

  return nState;
exit:
  return -1;
}

void EngineDown(void *pe, int nState)
{
  SEngine *pEngine=pe;
  IEngineState *pStateTarget=pEngine->pStateMin+nState;

  if (pStateTarget<pEngine->pState) {
    do {
      DVWRITELN("leaving %s ...",(pEngine->pState-1)->pId);
      (pEngine->pState-1)->Leave(pEngine);
    } while (pStateTarget<pEngine->pState);
  }
}
#endif // ]
