/*
 * plugin.h
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef PLUGIN_H
#define PLUGIN_H

#include <stdio.h>
#include <windows.h>
#include <tchar.h>
#include <Winamp/in2.h>
#include <ffsox.h>
#if defined (LIBFFSOX_BS1770) // [
//#define IN_FFSOX_BS1770
#endif // ]
//#define THREAD_VISTOGGLE
#include "vout.h"
#if defined (FFSOX_EXPERIMENTAL) // [
#define YASAPI
#endif // ]
#if defined (YASAPI) // [
#include "yasapi_nt_api.h"
#endif // ]
#if defined (FFSOX_EXPERIMENTAL) // [
//#define YAS
#endif // ]
#if defined (YAS) // [
#include "yas_api.h"
#endif // ]

//#if defined (FFSOX_DEBUG) // [
// dont't touch, instead define DEBUG in <pb.h>:
//#define DEBUG
//#endif // ]
//#if ! defined (DEBUG) // [
////#define DEBUG
//#endif // ]
//#define TABS_INIT_ADAPT
//#define LOCAL_AR
//#define MINGW_LINK
//#define FFSOX_32BIT
//#define ENABLE_THEME_DIALOG_TEXTURE_INDIRECT
//#define TOP_ON_SWITCH
#define SOX
#define VIDEO_PRORITY
#if defined (FFSOX_DEPRECATED) // [
//#define VIDEO_START_PTS
#endif // ]

#if defined (DEBUG) // {
//#define DEBUG_TAGS
#endif // }

/*
 * +---------+                   +-------+
 * |         |                   |       |
 * | FFmpeg  |                   |  SoX  |
 * |         |                   |       |
 * +----+----+                   +---+---+
 *      |                            |
 *      |                            |
 * +----+----+     +-------+     +---+---+     +---------+     +-------+
 * |         |     |       |     |       |     |         |     |       |
 * |         +---->+ queue +<----+ audio +---->+ samples +<----+ main  |
 * |         |     |       |     |       |     |         |     |       |
 * |         |     +-------+     +-------+     +---------+     +-------+
 * |         |                     loop                          loop
 * | plugin  |
 * |         |     +-------+     +-------+     +---------+
 * |         |     |       |     |       |     |         |
 * |         +---->+ queue +<----+ video +---->+ screen  |
 * |         |     |       |     |       |     |         |
 * +---------+     +-------+     +-------+     +---------+
 *    loop                         loop   
 */

///////////////////////////////////////////////////////////////////////////////
#define _STRA(x) #x
#define STRA(x) _STRA(x)

#define _STRW(x) L ## #x
#define STRW(x) _STRW(x)

#if defined (FFSOX_MAP_MEMORY) // [
///////////////////////////////////////////////////////////////////////////////
#define MAP_SIZE 4096

#if defined (YAS) // [
#define PREFETCH
#endif // ]
#endif // ]

///////////////////////////////////////////////////////////////////////////////
#define MODULE "in_ffsox.dll"

#if ! defined (LABEL)
  //#define LABEL "FFSoX Player 2"
  //#define LABEL "FFSoX 2a Input"
#if defined (SOX) // [
  #define LABEL "FF (mpeg and) SoX 2a Input"
#else // ] [
  #define LABEL "FF (mpeg and formerly SoX) 2a Input"
#endif // ]
#endif

#if ! defined (VERSION)
  #define VERSION "0.0"
#endif

#if ! defined (CAPTION)
  #define CAPTION LABEL " v" VERSION
#endif

#define ZERO(p) \
  memset(p,0,sizeof(*p))
#define Q2DB(q) \
  (20.0*log10(q))
#define DB2Q(db) \
  (pow(10.0,0.05*(db)))

#if 1 // {
#define round(x) \
  floor((x)+0.5)
#define stricmp(s1,s2) \
  _stricmp(s1,s2)
#endif // }

#if 0 // [
#if defined (DEBUG) // {
  #define DPUTS(s) fputs(s,stderr)
  #define DPRINTF(fmt,...) fprintf(stderr,fmt,__VA_ARGS__)
  #define DWPUTS(s) fwputs(s,stderr)
  #define DWPRINTF(fmt,...) fwprintf(stderr,fmt,__VA_ARGS__)
  #define DMARKLN() fprintf(stderr,"[%s:%d:%s]\n",\
      pb_basenamea(__FILE__),__LINE__,__func__)
  #define DWRITELN(s) fprintf(stderr,s " [%s:%d:%s]\n",\
      pb_basenamea(__FILE__),__LINE__,__func__)
  #define DVWRITELN(fmt,...) fprintf(stderr,fmt " [%s:%d:%s]\n",__VA_ARGS__,\
      pb_basenamea(__FILE__),__LINE__,__func__)
  #define DMESSAGE(cs) av_log(NULL,AV_LOG_ERROR,\
      "ERROR: " cs " [%s:%d].\n",pb_basenamea(__FILE__),__LINE__)
  #define DVMESSAGE(fmt,...) av_log(NULL,AV_LOG_ERROR,\
      "ERROR: " fmt " [%s:%d].\n",__VA_ARGS__,pb_basenamea(__FILE__),__LINE__)
  #define DGETC() fgetc(stdin)
#else // } {
  #define DPUTS(s)
  #define DPRINTF(fmt,...)
  #define DWPUTS(s)
  #define DWPRINTF(fmt,...)
  #define DMARKLN()
  #define DWRITELN(s)
  #define DVWRITELN(fmt,...)
  #define DMESSAGE(cs)
  #define DVMESSAGE(fmt,...)
  #define DGETC()
#endif // }
#endif // ]

///////////////////////////////////////////////////////////////////////////////
int ffsox_load_ffmpeg(pb_str_t *home);
#if defined (SOX) // [
int ffsox_load_sox(pb_str_t *home);
#endif // ]
int ffsox_load_dlgitemurl(pb_str_t *home);

///////////////////////////////////////////////////////////////////////////////
BOOL CALLBACK AboutDlgProc(HWND hDlg, UINT message, WPARAM wParam,
    LPARAM lParam);

///////////////////////////////////////////////////////////////////////////////
typedef enum state {
  STATE_RUN=0,
  STATE_PAUSE,
  STATE_EOF,
  STATE_KILL,
  STATE_MAX,
} state_t;

typedef struct ap {
  int rate;
  int nch;
  int bps;
} ap_t;

typedef struct decoder decoder_t;
typedef struct plugin plugin_t;
#if defined (LOCAL_AR) // {
typedef struct ar ar_t;
#endif // }

///////////////////////////////////////////////////////////////////////////////
typedef struct buf {
  char *rep;
  char *wp;
  char *mp;
} buf_t;

buf_t *buf_init(buf_t *buf, char *rep, size_t size);
buf_t *buf_cleanup(buf_t *buf);

buf_t *buf_realloc(buf_t *buf, size_t n);
buf_t *buf_append_n(buf_t *buf, char *cs, size_t n);
buf_t *buf_append(buf_t *buf, char *cs);

///////////////////////////////////////////////////////////////////////////////
typedef struct wbuf {
  wchar_t *rep;
  wchar_t *wp;
  wchar_t *mp;
} wbuf_t;

wbuf_t *wbuf_init(wbuf_t *wb, wchar_t *rep, size_t n);

wbuf_t *wbuf_add(wbuf_t *wb, wchar_t *wcs);
wbuf_t *wbuf_add_wva(wbuf_t *wb, wchar_t *fmt, ...);
wbuf_t *wbuf_add_va(wbuf_t *wb, char *fmt, ...);

///////////////////////////////////////////////////////////////////////////////
#define SOX_DITHER_OFF \
    "off"
#define SOX_DITHER_FILTER \
    "-f "
#define RG_WEIGHT_Q \
    100.0
#if defined (FFSOX_AUDIO_DRC) // [
#define PREFS_AUDIO_DRC
#define AUDIO_DRC_Q \
    0.1
#endif // ]
#define VIS_DELAY_MS_Q \
    100.0
#define PREFS_YADIF

typedef struct prefs {
  pb_str_t *path;

  struct {
    buf_t display;
    buf_t plugin;
  } exts;

  struct {
    int enable_plugin;
    int bps;
    struct {
      int enabled;

      struct {
        int enabled;
        double ms;
      } delay;
    } vis;

    struct {
      int enabled;
#if defined (VIDEO_PRORITY) // [
      int prio;
#endif // ]
    } vid;
  } general;

  struct {
#if ! defined (RG_WEIGHT_BOX) // {
    int off;
#endif // }
    double weight;
    int comment;
  } rg;

  struct {
    int queue_size;
    double preamp;
#if defined (PREFS_AUDIO_DRC) // [
    int force_stereo;

    struct {
      int enabled;
      double scale;
    } drc;
#endif // ]
  } audio;

  struct {
    char mode[16];
    int rate;
    int rate_min;
    int rate_mul;
    char rate_q[16];
    char dither[32];
  } sox;

  struct {
    int queue_size;
    double min_drift;
    double max_drift;

#if defined (PREFS_YADIF) // [
    struct {
      int disabled;
      int mode;
      int parity;
      int deint;
    } yadif;
#endif // ]
  } video;

#if defined (IN_FFSOX_BS1770) // {
  struct {
    int off;
    double gate;
  } adapt;
#endif // }

#if defined (YASAPI) // [
#define PREFS_YASAPI
#endif // ]
#if defined (YAS) // [
#define PREFS_YAS
#endif // ]
#if defined (PREFS_YASAPI) || defined (PREFS_YAS) // [
#define PREFS_INTEGRATE
#if defined (PREFS_INTEGRATE) && defined (PREFS_YASAPI) // [
#define PREFS_INTEGRATE_YASAPI
#define YASAPI_OFF 0
#define YASAPI_INTERNAL 1
#define YASAPI_EXTERNAL 2
#endif // ]
#endif // ]

#if defined (PREFS_INTEGRATE) // [
  struct {
#if defined (PREFS_YASAPI) // [
    struct {
      int enabled;
    } yasapi;
#endif // ]

#if defined (PREFS_YAS) // [
    struct {
      int enabled;
    } yas;
#endif // ]
  } integrate;
#endif // ]
} prefs_t;

HRESULT PrefsEnableThemeDialogTexture(HWND hwnd, DWORD dwFlags);

void prefs_load(prefs_t *play, prefs_t *conv, pb_str_t *home,
    HWND hMainWindow);
void prefs_edit(prefs_t *play, prefs_t *conv, HWND hParent);

#define ps_sox_mode(ps,val) \
  (0==strcmp(#val,(ps)->sox.mode))
#define ps_sox_mode_off(ps) \
  ps_sox_mode(ps,off)
#define prefs_sox_mode_auto(ps) \
  ps_sox_mode(ps,auto)
#define ps_sox_mode_on(ps) \
  ps_sox_mode(ps,on)

///////////////////////////////////////////////////////////////////////////////
typedef struct track {
  struct {
#define INFO_BEGIN(id) \
    pb_str_t *id;
#define INFO_ALT(id,id2)
#define INFO_END()
#include "info.h"
#undef INFO_END
#undef INFO_ALT
#undef INFO_BEGIN
  } info;

  pb_str_t *fn;
  prefs_t *ps;
  ffi_t ffi;
} track_t;

#if defined (FFSOX_MAP_MEMORY) // [
track_t *track_init(track_t *t, const in_char *fn, size_t size, prefs_t *ps,
    int full);
#else // ] [
track_t *track_init(track_t *t, const in_char *fn, prefs_t *ps, int full);
#endif // ]
void track_cleanup(track_t *t);

#if defined (FFSOX_MAP_MEMORY) // [
track_t *track_reopen(track_t *t, const in_char *fn, size_t size,
    prefs_t *ps, int full);
#else // ] [
track_t *track_reopen(track_t *t, const in_char *fn, prefs_t *ps, int full);
#endif // ]
void track_getfileinfo(track_t *t, in_char *title, int *length_in_ms);
in_char *track_getfileinfo_ext(track_t *t, char *metadata, wchar_t *ret,
    int retlen);

int track_sox_needed(track_t *t, ap_t *in, ap_t *out);

char *track_get_tag(track_t *t, char *key);
double track_get_q(track_t *t);
int track_get_length(track_t *t);
in_char *track_get_overwrite_comment(track_t *t, in_char *buf, size_t n);
in_char *track_get_formatinformation(track_t *t, in_char *buf, size_t n);

#define INFO_BEGIN(id) \
in_char *track_get_##id(track_t *t, in_char *buf, size_t n);
#define INFO_ALT(id,id2)
#define INFO_END()
#include "info.h"
#undef INFO_END
#undef INFO_ALT
#undef INFO_BEGIN

///////////////////////////////////////////////////////////////////////////////
#if defined (NO_WAITABLE_TIMER) // {
typedef struct timer {
  volatile state_t state;
  HANDLE hMutex;
  HANDLE hEvent;
  HANDLE hTimerQueue;
} timer_t;
#else // } {
typedef struct timer {
  volatile state_t state;
  HANDLE hTimer;
} timer_t;
#endif // }

timer_t *timer_init(timer_t *t);
void timer_cleanup(timer_t *t);

void timer_kill(timer_t *t);
void timer_sleep(timer_t *t, DWORD msInterval);

///////////////////////////////////////////////////////////////////////////////
typedef struct node {
  struct node *next;
  AVPacket pkt;
} node_t;

///////////////////////////////////////////////////////////////////////////////
typedef struct list {
  node_t *head;
  node_t *tail;
} list_t;

node_t *list_write(list_t *l, node_t *n);
node_t *list_read(list_t *l);

///////////////////////////////////////////////////////////////////////////////
#define QUEUE_SIZE 4

#define queue_overflow(q) \
  ((q)->size<=(q)->count)

typedef struct queue {
  decoder_t *dec;
  char *name;
  HANDLE hMutex;
  HANDLE hEventWrite;
  HANDLE hEventRead;

  int size;
  volatile int count;

  list_t active;
#if defined (QUEUE_FREE) // {
  list_t free;
#else // } {
  list_t *free;
#endif // }
} queue_t;

queue_t *queue_init(queue_t *q, decoder_t *dec, char *name, int size);
queue_t *queue_cleanup(queue_t *q);

void queue_reset(queue_t *q, int size);
void queue_flush(queue_t *q);

// non-blocking.
int queue_write(queue_t *q, AVPacket *pkt);
// blocking.
int queue_read(queue_t *q, AVPacket *pkt);

///////////////////////////////////////////////////////////////////////////////
typedef struct audio audio_t;

typedef struct samples {
  HANDLE hMutex;
  HANDLE hEventWrite;
  HANDLE hEventRead;

  volatile state_t state;
  volatile double time;
  track_t *t;
  int bps;
  double q;

  size_t size;

  struct {
    char *rep;
    volatile char *wp;
    char *mp;
  } in;

  struct {
    char *rep;
  } out;

#if defined (IN_FFSOX_BS1770) // {
  int nch;
  double rate;
#if defined (DEBUG) // {
  unsigned long long adaptive;
#else // } {
  int adaptive;
#endif // }
  bs1770_lufs_t lufs;
#endif // }
} samples_t;

samples_t *samples_init(samples_t *s);
samples_t *samples_cleanup(samples_t *s);

void samples_kill(samples_t *s);
void samples_clear(samples_t *s);
#if defined (IN_FFSOX_BS1770) // {
samples_t *samples_realloc(samples_t *s, track_t *t, size_t nsam,
    int nch, int bps, double rate);
void samples_adapt(samples_t *s, double *q);
#else // } {
samples_t *samples_realloc(samples_t *s, track_t *t, size_t nsam,
    int nch, int bps);
#endif // }

void samples_wait_lock(samples_t *s);
void samples_notify_unlock(samples_t *s);

#if defined (LOCAL_AR) // {
void samples_copy(samples_t *s, ar_t *ar);
#else // } {
void samples_copy(samples_t *s, ffsox_ar_t *ar);
#endif // }

///////////////////////////////////////////////////////////////////////////////
typedef struct screen {
  In_Module *mod;
  vout_t *vo;
  
  int open;
  int width;
  int height;
  double ar;
} screen_t;

screen_t *screen_init(screen_t *s, In_Module *mod);
screen_t *screen_cleanup(screen_t *s);

void screen_open(screen_t *s, int width, int height, double ar);
#if defined (FFSOX_DEPRECATED_PICTURE_CROP) // [
void screen_draw(screen_t *s, AVFrame *frame);
#else // ] [
void screen_draw(screen_t *s, AVPicture *pict);
#endif // ]
void screen_close(screen_t *s);
void screen_done(screen_t *s);

#if defined (LOCAL_AR) // {
///////////////////////////////////////////////////////////////////////////////
#define ar_valid(ar) \
  (AR_STATE_READ==(ar)->state||AR_STATE_FLUSH==(ar)->state)

enum {
  AR_STATE_READ,
  AR_STATE_FLUSH,
  AR_STATE_DONE,
  AR_STATE_KILL,
  AR_STATE_ERROR,
};

struct ar {
  int state;
  audio_t *a;
  queue_t *q;
  plugin_t *p;
  AVCodecContext *dc;
  int touched;
  AVFrame *frame;
  AVPacket pkt;
};

void ar_init(ar_t *ar, audio_t *a);
void ar_start(ar_t *ar);
void ar_next(ar_t *ar);
#endif // }

///////////////////////////////////////////////////////////////////////////////
typedef struct audio {
  decoder_t *dec;
#if defined (LOCAL_AR) // {
  AVFrame *frame;
#endif // }
  queue_t q;
  samples_t s;

#if defined (LOCAL_AR) // {
  ar_t ar;
#else // } {
  ffsox_ar_t ar;
#endif // }
#if defined (SOX) // [
  sox_effects_chain_t *chain;
#endif // ]
  ffi_st_t *st;

  ap_t in;
  ap_t out;

  HANDLE hThread;
} audio_t;

audio_t *audio_init(audio_t *a, decoder_t *dec);
audio_t *audio_cleanup(audio_t *a);

audio_t *audio_realloc(audio_t *a, track_t *t, size_t nsam);
void audio_reset(audio_t *a);
audio_t *audio_run(audio_t *a);
void audio_wait(audio_t *a);
void audio_stop(audio_t *a);
void audio_kill(audio_t *a);
#if defined (IN_FFSOX_BS1770) // {
void audio_adapt(audio_t *a);
#endif // }

DWORD WINAPI audio_thread(LPVOID lpParam);

///////////////////////////////////////////////////////////////////////////////
// https://ffmpeg.org/ffmpeg-filters.html#yadif-1
// "yadif=<mode>:<parity>:<deint>"
// mode:
//   0, send_frame: Output one frame for each frame. 
//   1, send_field: Output one frame for each field. 
//   2, send_frame_nospatial: Like send_frame, but it skips the spatial
//        interlacing check. 
//   3, send_field_nospatial: Like send_field, but it skips the spatial
//        interlacing check. 
//   The default value is send_frame.
// parity:
//   0, tff: Assume the top field is first. 
//   1, bff: Assume the bottom field is first. 
//   -1, auto: Enable automatic detection of field parity. 
//   The default value is auto. If the interlacing is unknown or the decoder
//   does not export this information, top field first will be assumed.
// deint:
//   0, all: Deinterlace all frames. 
//   1, interlaced: Only deinterlace frames marked as interlaced. 
//   The default value is all.
typedef enum _EYadifMode {
  // send_frame: Output one frame for each frame.
  eYadifModeSendFrame=0,
  // send_field: Output one frame for each field. 
  eYadifModeSendField=1,
  // send_frame_nospatial: Like send_frame, but it skips the spatial
  // interlacing check. 
  eYadifModeSendFrameNoSpatial=2,
  // send_field_nospatial: Like send_field, but it skips the spatial
  // interlacing check. 
  eYadifModeSendFieldNoSpatial=3,
} EYadifMode;

typedef enum _EYadifParity {
  // tff: Assume the top field is first. 
  eYadifParityTff=0,
  // bff: Assume the bottom field is first. 
  eYadifParityBff=1,
  // auto: Enable automatic detection of field parity. 
  eYadifParityAuto=-1,
} EYadifParity;

typedef enum _EYadifDeint {
  // all: Deinterlace all frames. 
  eYadifDeintAll=0,
  // interlaced: Only deinterlace frames marked as interlaced. 
  eYadifDeintInterlaced=1,
} EYadifDeint;

///////////////////////////////////////////////////////////////////////////////
typedef struct video_filter {
  AVFilterGraph *filter_graph;
  AVFilterContext *buffersink_ctx;
  AVFilterContext *buffersrc_ctx;
  AVFrame *frame;
} video_filter_t;

int video_filter_init(video_filter_t *vf, AVFormatContext *pFormatContext,
    int vi, AVCodecContext *pCodecContext, const char *filters_descr);
void video_filter_cleanup(video_filter_t *vf);

int video_filter_push_frame(video_filter_t *vf, AVFrame *frame);
int video_filter_pull_frame(video_filter_t *vf);

///////////////////////////////////////////////////////////////////////////////
#define VIDEO_MIN_DRIFT (-200.0)
#define VIDEO_MAX_DRIFT (8000.0)

typedef struct video {
#if defined (VIDEO_START_PTS) // [
  struct {
    int64_t pts;
    int touched;
  } start;
#endif // ]
  decoder_t *dec;
  In_Module *mod;
  AVFrame *frame;
  timer_t timer;
  queue_t q;
  screen_t s;

  ffi_st_t *st;
  double time_base;
  video_filter_t vf;
  HANDLE hThread;

  int crop;
  int width;
  int height;
  int x;
  int y;
  int64_t next_pts;
} video_t;

video_t *video_init(video_t *v, decoder_t *dec, In_Module *mod);
video_t *video_cleanup(video_t *v);

video_t *video_realloc(video_t *v, track_t *t);
void video_reset(video_t *v);
video_t *video_run(video_t *v);
void video_wait(video_t *v);
void video_stop(video_t *v);
void video_kill(video_t *v);

DWORD WINAPI video_thread(LPVOID lpParam);

#if defined (YASAPI) // [
///////////////////////////////////////////////////////////////////////////////
typedef struct _Cache Cache;

//#define CACHE_CRITICAL_SECTION
struct _Cache {
#if defined (CACHE_CRITICAL_SECTION) // [
  CRITICAL_SECTION cs;
#endif // ]
  size_t size;
  size_t step;
  const char *rp;
  const char *mp;
};

void CacheCreate(Cache *pCache);
void CacheDestroy(Cache *pCache);

void CacheSet(Cache *pCache, const char *samples, size_t nSizeFFSoX,
    size_t nSizeYasapi);
void CacheReset(Cache *pCache);
int CacheValid(Cache *pCache);
size_t CacheSizeStep(Cache *pCache);
size_t CacheSizeLatency(Cache *pCache);
void CacheNext(Cache *pCache);
#endif // ]

///////////////////////////////////////////////////////////////////////////////
struct decoder {
#if defined (YASAPI) // [
	const In_Module *mod;
  const YasapiNtApi *yasapi;
  int frames;
  Cache cache;
#endif // ]
#if defined (YAS) // [
  const YasApi *yas;
#endif // ]
#if 0 && defined (FFSOX_DEPRECATED) // [
  int64_t pos_ms;
#endif // ]
  prefs_t *ps;
  volatile state_t state;
  HANDLE hMutex;
  HANDLE hEventWrite;
#if ! defined (QUEUE_FREE) // {
  list_t free;
#endif // }
  audio_t a;
  video_t v;
  track_t t;
  int is_video;
  HANDLE hThread;
};

decoder_t *decoder_init(decoder_t *dec, prefs_t *ps, In_Module *mod);
decoder_t *decoder_cleanup(decoder_t *dec);

#if 0 // [
double decoder_get_time(decoder_t *dec, In_Module *mod);
#else // ] [
int decoder_get_time(decoder_t *dec, In_Module *mod);
#endif // ]
#if defined (FFSOX_MAP_MEMORY) // [
#if defined (YASAPI) && defined (YAS) // [
decoder_t *decoder_start(decoder_t *dec, const in_char *fn, size_t size,
    size_t nsam, const YasapiNtApi *yasapi, const YasApi *yas);
#elif defined (YASAPI) // ] [
decoder_t *decoder_start(decoder_t *dec, const in_char *fn, size_t size,
    size_t nsam, const YasapiNtApi *yasapi);
#elif defined (YAS) // ] [
decoder_t *decoder_start(decoder_t *dec, const in_char *fn, size_t size,
    size_t nsam, const YasApi *yas);
#else // ] [
decoder_t *decoder_start(decoder_t *dec, const in_char *fn, size_t size,
    size_t nsam);
#endif // ]
#else // ] [
#if defined (YASAPI) && defined (YAS) // [
decoder_t *decoder_start(decoder_t *dec, const in_char *fn, size_t nsam,
    const YasapiNtApi *yasapi, const YasApi *yas);
#elif defined (YASAPI) // ] [
decoder_t *decoder_start(decoder_t *dec, const in_char *fn, size_t nsam,
    const YasapiNtApi *yasapi);
#elif defined (YAS) // ] [
decoder_t *decoder_start(decoder_t *dec, const in_char *fn, size_t nsam,
    const YasApi *yas);
#else // ] [
decoder_t *decoder_start(decoder_t *dec, const in_char *fn, size_t nsam);
#endif // ]
#endif // ]
void decoder_stop(decoder_t *dec, void (*killed)(void *), void *data);
#if 0 && defined (FFSOX_DEPRECATED) // [
void decoder_reset(decoder_t *dec, int64_t pos_ms);
decoder_t *decoder_run(decoder_t *dec, int64_t pos_ms);
#else // ] [
void decoder_reset(decoder_t *dec);
decoder_t *decoder_run(decoder_t *dec);
#endif // ]
void decoder_pause(decoder_t *dec);
void decoder_unpause(decoder_t *dec);
void decoder_wait(decoder_t *dec);
void decoder_kill(decoder_t *dec);
void decoder_eof(decoder_t *dec);
int decoder_seek(decoder_t *dec, int64_t pos_ms);
void decoder_write(decoder_t *dec, queue_t *q, AVPacket *pkt);

DWORD WINAPI decoder_thread(LPVOID lpParam);
#if defined (YASAPI) // [
int decoder_write_yasapi(decoder_t *dec, const char *samples, size_t size,
    volatile double *decode_pos_ms);
#endif // ]

#if defined (PREFETCH) // [
///////////////////////////////////////////////////////////////////////////////
typedef enum prefetch_state {
  prefetch_state_null,
  prefetch_state_running,
  prefetch_state_open,
} prefetch_state_t;

typedef enum prefetch_command {
  prefetch_command_null,
  prefetch_command_open,
  prefetch_command_exit,
} prefetch_command_t;

typedef struct prefetch {
  plugin_t *p;
  CRITICAL_SECTION cs;
  HANDLE hProducer;
  HANDLE hConsumer;
  HANDLE hThread;
  prefetch_state_t state;
  prefetch_command_t command;

  union {
    struct {
      const in_char *fn;
/*
      size_t size;
      size_t nsam;
#if defined (YASAPI) // [
      const YasapiNtApi *yasapi;
#endif // ]
#if defined (YAS) // [
      const YasApi *yas;
#endif // ]
*/
    } open;
  } args;

  decoder_t physical[2];
  decoder_t *foreground;
  decoder_t *background;
} prefetch_t;

prefetch_t *prefetch_init(prefetch_t *pf, plugin_t *p);
void prefetch_destroy(prefetch_t *pf);

decoder_t *prefetch_swap(prefetch_t *pf, const wchar_t *path);
#endif // ]

///////////////////////////////////////////////////////////////////////////////
struct plugin {
  In_Module *mod;
#if defined (YAS) // [
  HMODULE hYas;
#endif // ]
  pb_str_t *root;
  pb_str_t *home;

  struct {
    HANDLE hThread;
    DWORD dwThreadId;
  } delay_toggle_vis;

  struct {
    prefs_t play;
    prefs_t conv;
  } ps;

  track_t list;
#if defined (PREFETCH) // [
  prefetch_t pf;
#else // ] [
  decoder_t dec;
#endif // ]
};

#if defined (PREFETCH) // [
#if 0 && defined (FFSOX_DEPRECATED) // [
#define plugin_run(p) \
  (NULL==decoder_run((p)->pf.foreground,0ll)?NULL:p)
#else // ] [
#define plugin_run(p) \
  (NULL==decoder_run((p)->pf.foreground)?NULL:p)
#endif // ]
#define plugin_pause(p) \
  (decoder_pause((p)->pf.foreground))
#define plugin_unpause(p) \
  (decoder_unpause((p)->pf.foreground))
#define plugin_wait(p) \
  (decoder_wait((p)->pf.foreground))
#define plugin_kill(p) \
  (decoder_kill((p)->pf.foreground))
#define plugin_seek(p,pos_ms) \
  (decoder_seek((p)->pf.foreground,pos_ms))
#else // ] [
#if 0 && defined (FFSOX_DEPRECATED) // [
#define plugin_run(p) \
  (NULL==decoder_run(&(p)->dec,0ll)?NULL:p)
#else // ] [
#define plugin_run(p) \
  (NULL==decoder_run(&(p)->dec)?NULL:p)
#endif // ]
#define plugin_pause(p) \
  (decoder_pause(&(p)->dec))
#define plugin_unpause(p) \
  (decoder_unpause(&(p)->dec))
#define plugin_wait(p) \
  (decoder_wait(&(p)->dec))
#define plugin_kill(p) \
  (decoder_kill(&(p)->dec))
#define plugin_seek(p,pos_ms) \
  (decoder_seek(&(p)->dec,pos_ms))
#endif // ]

plugin_t *plugin_init(plugin_t *p, In_Module *mod, pb_str_t *root,
    pb_str_t *home);
plugin_t *plugin_cleanup(plugin_t *p);

#if defined (FFSOX_MAP_MEMORY) // [
plugin_t *plugin_start(plugin_t *p, const in_char *fn, size_t size);
#else // ] [
plugin_t *plugin_start(plugin_t *p, const in_char *fn);
#endif // ]
void plugin_stop(plugin_t *p, int killed);
void plugin_killed(void *data);

#if defined (SOX) // [
///////////////////////////////////////////////////////////////////////////////
sox_effects_chain_t *ffsox_chain(audio_t *a);
sox_effect_handler_t const *ffsox_reader();
sox_effect_handler_t const *ffsox_writer();
#endif // ]

//#define THREAD
#if defined (THREAD) // [
///////////////////////////////////////////////////////////////////////////////
typedef struct _SEngine SEngine;
typedef const struct _IEngineState IEngineState;
typedef enum _EProducer EProducer;
typedef struct _SProducer SProducer;
typedef const struct _IProducerState IProducerState;

////////////////
struct _SEngine {
  IEngineState *pStateMin;
  IEngineState *pState;
};

void EngineCreate(void *pe, const void *ps);
void EngineDestroy(void *pe);

int EngineState(void *pe);
const char *EngineStateIdEx(void *pe, int nState);
const char *EngineStateId(void *pe);
int EngineUp(void *pe, int nState);
void EngineDown(void *pe, int nState);

////////////////
struct _IEngineState {
  const char *pId;
  int (*Enter)(void *pEngine);
  void (*Leave)(void *pEngine);
};

////////////////
enum _EProducer {
  eProducerNull,
#define PRODUCER_STATE(x) eProducer ## x, // [
#include "producer_state.h"
#undef PRODUCER_STATE // ]
  eProducerMax,
  // Minus eSthreaNull:
  eProducerCount=eProducerMax-1,
};

////////////////
struct _SProducer {
  IProducerState *pStateMin;
  IProducerState *pState;
  In_Module *mod;
  decoder_t *dec;
  int seek_needed;
  int killDecodeProducer;
  CRITICAL_SECTION cs;
  HANDLE hEvent;
  HANDLE hThread;
};

int ProducerCreate(SProducer *pProducer, In_Module *mod, decoder_t *dec);
void ProducerDestroy(SProducer *pProducer);

////////////////
struct _IProducerState {
  const char *pId;
  int (*Enter)(SProducer *pProducer);
  void (*Leave)(SProducer *pProducer);
};

extern IProducerState gcaProducerState[eProducerCount];

////////////////
DWORD WINAPI ConsumerThreadProc(LPVOID lpParameter);
#endif // ]

#if defined (YASAPI) // [
///////////////////////////////////////////////////////////////////////////////
const YasapiNtApi *plugin_yasapi(plugin_t *p);
#endif // ]
#if defined (YAS) // [
///////////////////////////////////////////////////////////////////////////////
const YasApi *plugin_yas(plugin_t *p);
#endif // ]

#endif // PLUGIN_H
