/*
 * prefetch.c
 * Copyright (C) 2010 - 2018 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

#if defined (PREFETCH) // [
static DWORD WINAPI prefetch_command_proc(LPVOID param);

prefetch_t *prefetch_init(prefetch_t *pf, plugin_t *p)
{
  decoder_t *dec=pf->physical;

  ////////////////
  //memset(pf,0,sizeof *pf);
  pf->p=p;

  ////////////////
  InitializeCriticalSection(&pf->cs);

  ////////////////
  pf->hProducer=CreateEventW(NULL,FALSE,FALSE,NULL);

  if (!pf->hProducer) {
    DMESSAGE("creating producer event");
    goto error;
  }

  ////////////////
  pf->hConsumer=CreateEventW(NULL,FALSE,FALSE,NULL);

  if (!pf->hConsumer) {
    DMESSAGE("creating consumer event");
    goto error;
  }

  ////////////////
  pf->hThread=CreateThread(
    NULL,                   // LPSECURITY_ATTRIBUTES   lpThreadAttributes,
    0ul,                    // SIZE_T                  dwStackSize,
    prefetch_command_proc,  // LPTHREAD_START_ROUTINE  lpStartAddress,
    pf,                     // __drv_aliasesMem LPVOID lpParameter,
    0ul,                    // DWORD                   dwCreationFlags,
    NULL                    // LPDWORD                 lpThreadId
  );

  if (!pf->hThread) {
    DMESSAGE("starting thread");
    goto error;
  }

  ////////////////
  EnterCriticalSection(&pf->cs); // {

  while (prefetch_state_running!=pf->state) {
    LeaveCriticalSection(&pf->cs); // {
    WaitForSingleObject(pf->hProducer,INFINITE);
    EnterCriticalSection(&pf->cs); // }
  }

  LeaveCriticalSection(&pf->cs); // }

  ////////////////
  if (!decoder_init(dec,&p->ps.play,p->mod))
    goto error;

  pf->foreground=dec;
  ++dec;

  ////////////////
  if (!decoder_init(dec,&p->ps.play,p->mod))
    goto error;

  pf->background=dec;
  ++dec;

  return pf;
error:
  prefetch_destroy(pf);
  return NULL;
}

void prefetch_destroy(prefetch_t *pf)
{
  decoder_t *dec=pf->physical+2;

  while (pf->physical<dec)
    decoder_cleanup(--dec);

  if (pf->hThread) {
    pf->command=prefetch_command_exit;
    SetEvent(pf->hConsumer);
    WaitForSingleObject(pf->hThread,INFINITE);
    CloseHandle(pf->hThread);
  }

  if (pf->hConsumer)
    CloseHandle(pf->hConsumer);

  if (pf->hProducer)
    CloseHandle(pf->hProducer);

  DeleteCriticalSection(&pf->cs);
}

decoder_t *prefetch_swap(prefetch_t *pf, const wchar_t *fn)
{
  decoder_t *dec=pf->background;

  EnterCriticalSection(&pf->cs); // {

  while (prefetch_state_open!=pf->state) {
    LeaveCriticalSection(&pf->cs); // {
    WaitForSingleObject(pf->hProducer,INFINITE);
    EnterCriticalSection(&pf->cs); // }
  }

  dec=pf->background;
  pf->background=pf->foreground;
  pf->foreground=dec;

  pf->state=prefetch_state_null;
  pf->args.open.fn=fn;
  pf->command=prefetch_command_open;
  SetEvent(pf->hConsumer);
  LeaveCriticalSection(&pf->cs); // }

  return dec;
}

///////////////////////////////////////////////////////////////////////////////
static DWORD WINAPI prefetch_command_proc(LPVOID param)
{
  prefetch_t *pf=param;

  EnterCriticalSection(&pf->cs); // {
  pf->state=prefetch_state_running;
  SetEvent(pf->hProducer);

  for (;;) {
    LeaveCriticalSection(&pf->cs); // {
    WaitForSingleObject(pf->hConsumer,INFINITE);
    EnterCriticalSection(&pf->cs); // }

    switch (pf->command) {
    case prefetch_command_exit:
      goto exit;
    case prefetch_command_open:
      break;
    default:
      break;
    }

    pf->command=prefetch_command_null;
    SetEvent(pf->hProducer);
  }
exit:
  LeaveCriticalSection(&pf->cs); // }
  ExitThread(0ul);
  return 0ul;
}
#endif // ]
