/*
 * prefs.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <windows.h>
#include <commctrl.h>
#include <uxtheme.h>
#include <Winamp/wa_ipc.h>
#include "prefdef.h"

#if defined (TABS_INIT_ADAPT) // [
enum {
  WM_PREFS_POS=WM_APP
};
#endif // ]

///////////////////////////////////////////////////////////////////////////////
#define DEFAULT_PAGE 1

static prefs_t play_dps={
  // general
  .general.enable_plugin=1,
  .general.bps=16,
  .general.vis.enabled=1,
  .general.vis.delay.enabled=1,
  .general.vis.delay.ms=20.0/VIS_DELAY_MS_Q,
  .general.vid.enabled=1,

  // rg
  .rg.weight=0.5,
  .rg.comment=1,

  // audio
  .audio.queue_size=QUEUE_SIZE,
  .audio.preamp=0.0,
#if defined (PREFS_AUDIO_DRC) // [
  .audio.force_stereo=1,
  .audio.drc.enabled=0,
  .audio.drc.scale=0.0,
#endif // ]

  // sox
  .sox.mode="auto",
  .sox.rate=44100,
  .sox.rate_min=1,
  .sox.rate_mul=1,
  .sox.rate_q="-v",
  .sox.dither="-s",

  // video
  .video.queue_size=QUEUE_SIZE,
  .video.min_drift=-800.0,
#define OLD_MAX_DRIFT
#if defined (OLD_MAX_DRIFT) // {
  .video.max_drift=8000.0,
#else // } {
  .video.max_drift=800.0,
#endif // }
#if defined (PREFS_YADIF) // [
  .video.yadif.disabled=0,
  .video.yadif.mode=eYadifModeSendFrame,
  .video.yadif.parity=eYadifParityAuto,
  .video.yadif.deint=eYadifDeintInterlaced,
#endif // ]

#if defined (IN_FFSOX_BS1770) // {
  // adapt
  .adapt.off=0,
  .adapt.gate=-23.0,
#endif // }

#if defined (PREFS_YASAPI) // [
  // yasapi
  .integrate.yasapi.enabled=1,
#endif // ]
#if defined (PREFS_YAS) // [
  // yas
  .integrate.yas.enabled=1,
#endif // ]
};

///////////////////////////////////////////////////////////////////////////////
typedef struct page {
  prefs_general_t general;
  prefs_audio_t audio;
  prefs_video_t video;
  prefs_conv_t conv;
#if defined (IN_FFSOX_BS1770) // {
  prefs_adapt_t adapt;
#endif // }
#if defined (PREFS_INTEGRATE) // [
  prefs_yasapi_t integrate;
#endif // ]
  prefs_help_t help;
} page_t;

static tcp_t tcp[]={
#if defined (VIDEO_PRORITY) // [
  { L"General",IDD_PREFS_GENERAL_VIDEO_PRORITY,GeneralDlgProc,
      offsetof(page_t,general) },
#else // ] [
  { L"General",IDD_PREFS_GENERAL,GeneralDlgProc,offsetof(page_t,general) },
#endif // ]
  { L"Audio",IDD_PREFS_AUDIO,AudioDlgProc,offsetof(page_t,audio) },
  { L"Video",IDD_PREFS_VIDEO,VideoDlgProc,offsetof(page_t,video) },
  { L"Convert",IDD_PREFS_CONV,ConvDlgProc,offsetof(page_t,conv) },
#if defined (IN_FFSOX_BS1770) // {
  { L"Adapt",IDD_PREFS_ADAPT,AdaptDlgProc,offsetof(page_t,adapt) },
#endif // }
#if defined (PREFS_INTEGRATE_YASAPI) // [
  { L"Integration",IDD_INTEGRATE_YASAPI,IntegrateDlgProc,offsetof(page_t,
      integrate) },
#elif defined (PREFS_INTEGRATE) // [
  { L"Integration",IDD_INTEGRATE,IntegrateDlgProc,offsetof(page_t,integrate) },
#endif // ]
  { L"Help",IDD_PREFS_HELP,HelpDlgProc,offsetof(page_t,help) },
  { NULL,0,NULL }
};

enum { PREFS_NPAGES=ARRAY_SIZE(tcp)-1 };

typedef struct priv {
  prefs_t *ps;
  page_t page;
  tabs_t tc;
  HWND hPages[PREFS_NPAGES];
#if defined (TABS_INIT_ADAPT) // [
  int bSendPos;
#endif // ]
} priv_t;

///////////////////////////////////////////////////////////////////////////////
void prefs_read(prefs_t *ps, prefdef_t *pd, prefs_t *dps)
{
  char *path=PS_UTF8(ps->path);
  char buf[128];

  switch (pd->type) {
  case PREF_INT:
    pref_int(pd,ps)=GetPrivateProfileIntA(pd->group,pd->name,
        pref_int(pd,dps),path);
    break;
  case PREF_DOUBLE:
    sprintf(buf,"%f",pref_double(pd,dps));
    GetPrivateProfileStringA(pd->group,pd->name,buf,buf,sizeof buf,path);
    pref_double(pd,ps)=atof(buf);
    break;
  case PREF_STRING:
    GetPrivateProfileStringA(pd->group,pd->name,pref_string(pd,dps),
        pref_string(pd,ps),pd->size,path);
    break;
  default:
    break;
  }
}

void prefs_read_play(prefs_t *ps, prefdef_t *pd)
{
  prefs_read(ps,pd,&play_dps);
}

void prefs_write(prefs_t *ps, prefdef_t *pd)
{
  char *path=PS_UTF8(ps->path);
  char buf[128];

  switch (pd->type) {
  case PREF_INT:
    sprintf(buf,"%d",pref_int(pd,ps));
    WritePrivateProfileStringA(pd->group,pd->name,buf,path);
    break;
  case PREF_DOUBLE:
    sprintf(buf,"%f",pref_double(pd,ps));
    WritePrivateProfileStringA(pd->group,pd->name,buf,path);
    break;
  case PREF_STRING:
    WritePrivateProfileStringA(pd->group,pd->name,pref_string(pd,ps),path);
    break;
  default:
    break;
  }
}

///////////////////////////////////////////////////////////////////////////////
void prefs_load(prefs_t *play, prefs_t *conv, pb_str_t *home,
    HWND hMainWindow)
{
  char *dir=(char *)SendMessageW(hMainWindow,WM_WA_IPC,0,IPC_GETINIDIRECTORY);
  pb_str_t *p1=NULL;
  pb_str_t *p2=NULL;
  pb_str_t *p3=NULL;
  pb_str_t *p4=NULL;

  if (NULL==dir||0==*dir) {
    if (NULL==(p3=pb_str_cat_ps(home,_PS("\\in_ffsox.ini"),NULL)))
      goto cleanup;
  }
  else {
    if (NULL==(p1=pb_str_cre_cs(dir)))
      goto cleanup;

#if defined (PLUGIN_DIR) // {
    if (NULL==(p2=pb_str_cat_ps(p1,_PS("\\Plugins\\in_ffsox"),NULL)))
      goto cleanup;
#else // } {
    if (NULL==(p2=pb_str_cat_ps(p1,_PS("\\Plugins"),NULL)))
      goto cleanup;
#endif // }

    CreateDirectoryW(
      PS_TCS(p2),   // __in      LPCTSTR lpPathName,
      NULL          // __in_opt  LPSECURITY_ATTRIBUTES lpSecurityAttributes
    );

    if (NULL==(p3=pb_str_cat_ps(p2,_PS("\\in_ffsox.ini"),NULL)))
      goto cleanup;
  }

  if (NULL==(p4=pb_str_dup(p3)))
    goto cleanup;

  play->path=p3;
  p3=NULL;
  conv->path=p4;
  p4=NULL;

  prefs_general_load(play);
  prefs_audio_load(play);
  prefs_video_load(play);
#if defined (PREFS_INTEGRATE) // [
  prefs_integrate_load(play);
#endif // ]

  prefs_conv_general_load(conv);
  prefs_conv_audio_load(conv);

#if defined (IN_FFSOX_BS1770) // {
  prefs_adapt_load(play);
#endif // }
  prefs_help_load(play);
cleanup:
  if (NULL!=p4)
    free(p4);

  if (NULL!=p3)
    free(p3);

  if (NULL!=p2)
    free(p2);

  if (NULL!=p1)
    free(p1);
}

///////////////////////////////////////////////////////////////////////////////
static void PrefsDlgDispatch(priv_t *p, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  int i;

  for (i=0;i<PREFS_NPAGES;++i)
    SendMessage(p->hPages[i],uMsg,wParam,lParam);
}

BOOL CALLBACK PrefsDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  priv_t *p=WM_INITDIALOG==uMsg?(void *)lParam
      :(void *)GetWindowLongPtr(hDlg,GWLP_USERDATA);
  int n;
  char buf[32];
#if 0 && defined (DEBUG) // [
  int i;
#endif // ]

  switch(uMsg) {
#if defined (TABS_INIT_ADAPT) // [
  // Raymond Chen rocks!
  // https://blogs.msdn.microsoft.com/oldnewthing/20060925-02/?p=29603
  case WM_PREFS_POS:
    n=GetPrivateProfileIntA(PLAY_GENERAL,"page",DEFAULT_PAGE,
        PS_UTF8(p->ps->path));
    tabs_init(&p->tc,hDlg,IDC_TAB_PREFS,n-1,tcp,p->hPages,(char *)&p->page);
    break;
  case WM_WINDOWPOSCHANGED:
    if ((SWP_SHOWWINDOW&((WINDOWPOS*)lParam)->flags)&&p->bSendPos) {
      p->bSendPos=0;
      PostMessage(hDlg,WM_PREFS_POS,0,0);
    }

    break;
#endif // ]
  case WM_INITDIALOG:
#if 0
    PrefsEnableThemeDialogTexture(
      hDlg,             // _In_  HWND hwnd,
      ETDT_ENABLETAB    // _In_  DWORD dwFlags
    );
#endif
    SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);
    SetWindowTextA(hDlg,CAPTION " Preferences");
#if defined (TABS_INIT_ADAPT) // [
    p->bSendPos=1;
#else // ] [
    n=GetPrivateProfileIntA(PLAY_GENERAL,"page",DEFAULT_PAGE,
        PS_UTF8(p->ps->path));
#if 0 && defined (DEBUG) // [
DVWRITELN("offsetof(page_t,audio): %d",offsetof(page_t,audio));
DVWRITELN("offsetof(page_t,yasapi): %d",offsetof(page_t,yasapi));
for (i=0;i<PREFS_NPAGES;++i) DVWRITELN("tcp[%d].offs: %d",i,tcp[i].offs);
DVWRITELN("eOffsYasapi: %d",eOffsYasapi);
#endif // ]
    tabs_init(&p->tc,hDlg,IDC_TAB_PREFS,n-1,tcp,p->hPages,(char *)&p->page);
#endif // ]
    return TRUE;
  case WM_NOTIFY:
    if (IDC_TAB_PREFS==wParam&&TCN_SELCHANGE==((NMHDR *)lParam)->code) {
      tabs_change(&p->tc,hDlg);
      return TRUE;
    }

    break;
  case WM_COMMAND:
    switch(LOWORD(wParam)) {
    case IDOK:
//DPRINTF("%s:%d\n",__func__,__LINE__);
#if defined (TABS_INIT_ADAPT) // [
      n=TabCtrl_GetCurSel(p->tc.hTab); 
#else // ] [
      n=TabCtrl_GetCurSel(GetDlgItem(hDlg,p->tc.id)); 
#endif // ]
//DPRINTF("%s:%d\n",__func__,__LINE__);
      sprintf(buf,"%d",1+n);
//DPRINTF("%s:%d\n",__func__,__LINE__);
      WritePrivateProfileStringA(PLAY_GENERAL,"page",buf,PS_UTF8(p->ps->path));
//DPRINTF("%s:%d\n",__func__,__LINE__);
      PrefsDlgDispatch(p,WM_COMMAND,IDOK,0);
//DPRINTF("%s:%d\n",__func__,__LINE__);
      EndDialog(hDlg,IDOK);
//DPRINTF("%s:%d\n",__func__,__LINE__);
      return TRUE;
    case IDCANCEL:
      EndDialog(hDlg,IDCANCEL);
      return TRUE;
    case IDC_BUTTON_DEFAULT:
      PrefsDlgDispatch(p,WM_COMMAND,IDC_BUTTON_DEFAULT,(LPARAM)&play_dps);
      tabs_set_page(&p->tc,hDlg,DEFAULT_PAGE-1);
      return TRUE;
    default:
      break;
    }

    break;
  default:
    break;
  }

  return FALSE;
}

///////////////////////////////////////////////////////////////////////////////
void prefs_edit(prefs_t *play, prefs_t *conv, HWND hParent)
{
  priv_t p;

  ZERO(&p);
  p.ps=play;
  p.page.general.ps=play;
  p.page.audio.ps=play;
  p.page.video.ps=play;
#if defined (IN_FFSOX_BS1770) // {
  p.page.adapt.ps=play;
#endif // }
#if defined (PREFS_INTEGRATE) // [
  p.page.integrate.ps=play;
#endif // ]

  p.page.conv.ps=conv;
  p.page.conv.page.general.ps=conv;
  p.page.conv.page.audio.ps=conv;

DMARKLN();
  DialogBoxParamW(
    GetModuleHandleA(MODULE),       // __in_opt  HINSTANCE hInstance
    MAKEINTRESOURCEW(IDD_PREFS),    // __in LPCTSTR lpTemplateName
    hParent,                        // __in_opt  HWND hWndParent
    PrefsDlgProc,                   // __in_opt  DLGPROC lpDialogFunc
    (LPARAM)&p                      // __in LPARAM dwInitParam
  );
}
