/*
 * prefs_video.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "prefdef.h"
#include <uxtheme.h>

#if defined (IN_FFSOX_BS1770) // {
#define PREFS_ADAPT

///////////////////////////////////////////////////////////////////////////////
typedef prefs_video_t priv_t;

static prefdef_t pds[]=ADAPT_PDS();
static prefdef_t *mpd=pds+PREFS_ADAPT_N;

///////////////////////////////////////////////////////////////////////////////
void prefs_adapt_load(prefs_t *ps)
{
  prefdef_t *pd;

  for (pd=pds;pd<mpd;++pd)
    prefs_read_play(ps,pd);
}

///////////////////////////////////////////////////////////////////////////////
#define PREFS_SYNC_SLIDER
#include "slider.h"

static void priv_init(priv_t *p, prefs_t *ps, HWND hDlg)
{
  prefdef_t *pd;
  prefgui_t *pg;

  for (pd=pds,pg=p->pg;pd<mpd;++pd,++pg) {
    if (NULL!=pd->init) {
      pd->init(pg,pd,ps,hDlg);

      switch (pg->id) {
#define PREFS_INIT_SLIDER
#include "slider.h"
      default:
        break;
      }
    }
  }
}

static void priv_save(priv_t *p, prefs_t *ps, HWND hDlg)
{
  prefdef_t *pd;
  prefgui_t *pg;

  for (pd=pds,pg=p->pg;pd<mpd;++pd,++pg) {
    if (NULL!=pg->vmt) {
      pg->vmt->get(pg,ps,hDlg);
      prefs_write(ps,pd);
    }
  }
}

static void priv_display(priv_t *p, prefs_t *ps, HWND hDlg)
{
  prefdef_t *pd;
  prefgui_t *pg;

  for (pd=pds,pg=p->pg;pd<mpd;++pd,++pg) {
    if (NULL!=pg->vmt)
      pg->vmt->set(pg,ps,hDlg);
  }
}

///////////////////////////////////////////////////////////////////////////////
#define URL_PFX "file:///"
#define setdlgitemurl_stub_valid (NULL!=*__SetDlgItemUrl)

#if defined (PREFS_ADAPT_CONTROLS) // {
static void set_link(HWND hDlg, int id, char *buf, int size)
{
  if (setdlgitemurl_stub_valid) {
    GetDlgItemTextA(hDlg,id,buf,size);
    __SetDlgItemUrl(hDlg,id,buf);
  }
}
#endif // }

BOOL CALLBACK AdaptDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
#if defined (PREFS_ADAPT_CONTROLS) // {
  static char buf[(sizeof URL_PFX)+MAX_PATH];
#endif // }

  priv_t *p=WM_INITDIALOG==uMsg?(void *)lParam
      :(void *)GetWindowLongPtr(hDlg,GWLP_USERDATA);

  switch(uMsg) {
  case WM_INITDIALOG:
    PrefsEnableThemeDialogTexture(
      hDlg,             // _In_  HWND hwnd,
      ETDT_ENABLETAB    // _In_  DWORD dwFlags
    );

	SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);
    priv_init(p,p->ps,hDlg);
#if defined (PREFS_ADAPT_CONTROLS) // {
    set_link(hDlg,IDC_STATIC_ADAPT_WEB_PAGE_URL,buf,sizeof buf);
    set_link(hDlg,IDC_STATIC_ADAPT_WA_FORUM_URL,buf,sizeof buf);
#if 0 // {
    set_link(hDlg,IDC_STATIC_ADAPT_WAE_FORUM_URL,buf,sizeof buf);
#endif // }
    set_link(hDlg,IDC_STATIC_ADAPT_FS_PROJECT_URL,buf,sizeof buf);
    set_link(hDlg,IDC_STATIC_ADAPT_DOWNLOAD_URL,buf,sizeof buf);
#endif // }
    return TRUE;
  case WM_COMMAND:
#if 1 // {
    switch(LOWORD(wParam)) {
    case IDOK:
      priv_save(p,p->ps,hDlg);
      return TRUE;
    case IDC_BUTTON_DEFAULT:
      priv_display(p,(prefs_t *)lParam,hDlg);
      return TRUE;
    default:
	  break;
    }
#endif // }

    break;
  case WM_NOTIFY:
    switch (LOWORD(wParam)) {
#define PREFS_NOTIFY_SLIDER
#include "slider.h"
    default:
      break;
    }

    break;
  default:
    break;
  }

  return FALSE;
}
#endif // }
