/*
 * prefs_conv_general.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "prefdef.h"
#include <uxtheme.h>

#define PREFS_CONV_GENERAL

///////////////////////////////////////////////////////////////////////////////
typedef prefs_conv_general_t priv_t;

static prefdef_t pds[]=CONV_GENERAL_PDS();
static prefdef_t *mpd=pds+PREFS_CONV_GENERAL_N;

///////////////////////////////////////////////////////////////////////////////
void prefs_conv_general_load(prefs_t *ps)
{
  prefdef_t *pd;

  for (pd=pds;pd<mpd;++pd)
    prefs_read_conv(ps,pd);
}

///////////////////////////////////////////////////////////////////////////////
#define PREFS_SYNC_SLIDER
#include "slider.h"

static void priv_init(priv_t *p, prefs_t *ps, HWND hDlg)
{
  prefdef_t *pd;
  prefgui_t *pg;

  for (pd=pds,pg=p->pg;pd<mpd;++pd,++pg) {
    if (NULL!=pd->init) {
//#if defined (DEBUG) // [
//      pd->init(pg,pg->name,pd,ps,hDlg);
//#else // ] [
      pd->init(pg,pd,ps,hDlg);
//#endif // ]

      switch (pg->id) {
#define PREFS_INIT_SLIDER
#include "slider.h"
      default:
        break;
      }
    }
  }
}

static void priv_save(priv_t *p, prefs_t *ps, HWND hDlg)
{
  prefdef_t *pd;
  prefgui_t *pg;

  for (pd=pds,pg=p->pg;pd<mpd;++pd,++pg) {
    if (NULL!=pg->vmt) {
      pg->vmt->get(pg,ps,hDlg);
      prefs_write(ps,pd);
    }
  }
}

static void priv_display(priv_t *p, prefs_t *ps, HWND hDlg)
{
  prefdef_t *pd;
  prefgui_t *pg;

  for (pd=pds,pg=p->pg;pd<mpd;++pd,++pg) {
    if (NULL!=pg->vmt)
      pg->vmt->set(pg,ps,hDlg);
  }
}

///////////////////////////////////////////////////////////////////////////////
static void EnableReplayGain(HWND hDlg)
{
  int enabled=!IsDlgButtonChecked(hDlg,IDC_CHECKBOX_RG_OFF);

  // weight
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_RG_WEIGHT_ALBUM),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_RG_WEIGHT_TRACK),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_RG_WEIGHT_LABEL),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_RG_WEIGHT),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_SLIDER_RG_WEIGHT),enabled);

#if 0 // {
  // preamp
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_RG_PREAMP_LABEL),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_RG_PREAMP),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_SLIDER_RG_PREAMP),enabled);
#endif // }

  // comment
  EnableWindow(GetDlgItem(hDlg,IDC_CHECKBOX_RG_COMMENT),enabled);
}

///////////////////////////////////////////////////////////////////////////////
BOOL CALLBACK ConvGeneralDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  priv_t *p=WM_INITDIALOG==uMsg?(void *)lParam
      :(void *)GetWindowLongPtr(hDlg,GWLP_USERDATA);

  switch(uMsg) {
  case WM_INITDIALOG:
    PrefsEnableThemeDialogTexture(
      hDlg,             // _In_  HWND hwnd,
      ETDT_ENABLETAB    // _In_  DWORD dwFlags
    );

    SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);
    priv_init(p,p->ps,hDlg);
    EnableReplayGain(hDlg);
    return TRUE;
  case WM_COMMAND:
    switch(LOWORD(wParam)) {
    case IDOK:
      priv_save(p,p->ps,hDlg);
      return TRUE;
    case IDC_CHECKBOX_RG_OFF:
      EnableReplayGain(hDlg);
      return TRUE;
    case IDC_BUTTON_DEFAULT:
      priv_display(p,(prefs_t *)lParam,hDlg);
      EnableReplayGain(hDlg);
      return TRUE;
    default:
      break;
    }

    break;
  case WM_NOTIFY:
    switch (LOWORD(wParam)) {
#define PREFS_NOTIFY_SLIDER
#include "slider.h"
    default:
      break;
    }

    break;
  default:
    break;
  }

  return FALSE;
}
