/*
 * prefs_general.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "prefdef.h"
#include <uxtheme.h>

#define PREFS_GENERAL

///////////////////////////////////////////////////////////////////////////////
#if ! defined (FULL) // {
#define EXTS "FLAC;MKV;OGG;WAV;WEBM;WV"
#else // } {
#define EXTS "FLAC;MKV;OGG;WAV;WEBM;WV;MP3;MP4;M4A;FLV;WMV;MPG;VOB;AVI;M2TS"
#endif // }

///////////////////////////////////////////////////////////////////////////////
typedef prefs_general_t priv_t;

static prefdef_t pds[]=GENERAL_PDS();
static prefdef_t *mpd=pds+PREFS_GENERAL_N;

///////////////////////////////////////////////////////////////////////////////
static void display2plugin(buf_t *display, buf_t *plugin, int enable_plugin)
{
  char *rp=display->rep;
  char *mp=display->rep;

  plugin->wp=plugin->rep;

  if (enable_plugin) {
    for (;;) {
      if (';'==*mp||'\0'==*mp) {
        if (NULL==buf_append_n(plugin,rp,mp-rp))
          break;

        ++plugin->wp;

        if (NULL==buf_append_n(plugin,rp,mp-rp))
          break;

        ++plugin->wp;

        if ('\0'==*mp)
          break;
        else
          rp=++mp;
      }
      else
        ++mp;
    }
  }
  else {
    *plugin->wp++='\0';
    *plugin->wp++='\0';
  }
}

static void prefs_read_exts(prefs_t *ps)
{
  char *path=PS_UTF8(ps->path);
  int enable_plugin=ps->general.enable_plugin;
  int n;

  ps->exts.display.wp=ps->exts.display.rep;
  n=ps->exts.display.mp-ps->exts.display.wp-2;
  n=GetPrivateProfileStringA(PLAY_GENERAL,"exts",EXTS,
      ps->exts.display.wp,n,path);
  ps->exts.display.wp+=n;
  ps->exts.display.wp[0]='\0';
  ps->exts.display.wp[1]='\0';
  display2plugin(&ps->exts.display,&ps->exts.plugin,enable_plugin);
}

static void prefs_write_exts(prefs_t *ps, HWND hDlg)
{
  char *path=PS_UTF8(ps->path);
  int enable_plugin=ps->general.enable_plugin;
  int n;

  ps->exts.display.wp=ps->exts.display.rep;
  n=ps->exts.display.mp-ps->exts.display.wp-2;
  n=GetDlgItemTextA(hDlg,IDC_EDIT_WINAMP_EXTS,ps->exts.display.wp,n);
  ps->exts.display.wp+=n;
  ps->exts.display.wp[0]='\0';
  ps->exts.display.wp[1]='\0';
  WritePrivateProfileStringA(PLAY_GENERAL,"exts",ps->exts.display.rep,path);
  display2plugin(&ps->exts.display,&ps->exts.plugin,enable_plugin);
}

///////////////////////////////////////////////////////////////////////////////
#if defined (DEBUG) // {
static void prefs_list_exts()
{
#if 0 // {
  AVInputFormat  *f=NULL;
  const char *rp;
  char *wp;
  char buf[64];

  while (NULL!=(f=av_iformat_next(f))) {
    if (NULL==(rp=f->extensions))
      fprintf(stderr,"%s: NULL\n",f->name);
    else {
      wp=buf;

      for (;;) {
        if ('\0'==*rp||','==*rp) {
          *wp++='\0';
          fprintf(stderr,"%s: \"%s\"\n",f->name,buf);

          if ('\0'==*rp)
            break;

          ++rp;
          wp=buf;
        }
        else
          *wp++=*rp++;
      }
    }
  }
#endif // }
}
#endif // }

void prefs_general_load(prefs_t *ps)
{
  prefdef_t *pd;

#if defined (DEBUG) // {
  prefs_list_exts();
#endif // }
  prefs_read_exts(ps);

  for (pd=pds;pd<mpd;++pd)
    prefs_read_play(ps,pd);

  prefs_read_exts(ps);
}

///////////////////////////////////////////////////////////////////////////////
#define PREFS_SYNC_SLIDER
#include "slider.h"

static void priv_init(priv_t *p, prefs_t *ps, HWND hDlg)
{
  prefdef_t *pd;
  prefgui_t *pg;

  SetDlgItemTextA(hDlg,IDC_EDIT_WINAMP_EXTS,ps->exts.display.rep);

  for (pd=pds,pg=p->pg;pd<mpd;++pd,++pg) {
    if (NULL!=pd->init) {
//#if defined (DEBUG) // [
//      pd->init(pg,pg->name,pd,ps,hDlg);
//#else // ] [
      pd->init(pg,pd,ps,hDlg);
//#endif // ]

      switch (pg->id) {
#define PREFS_INIT_SLIDER
#include "slider.h"
      default:
        break;
      }
    }
  }
}

static void priv_save(priv_t *p, prefs_t *ps, HWND hDlg)
{
  prefdef_t *pd;
  prefgui_t *pg;

  prefs_write_exts(ps,hDlg);

  for (pd=pds,pg=p->pg;pd<mpd;++pd,++pg) {
    if (NULL!=pg->vmt) {
      pg->vmt->get(pg,ps,hDlg);
      prefs_write(ps,pd);
    }
  }

  prefs_write_exts(ps,hDlg);
}

static void priv_display(priv_t *p, prefs_t *ps, HWND hDlg)
{
  prefdef_t *pd;
  prefgui_t *pg;

  SetDlgItemTextA(hDlg,IDC_EDIT_WINAMP_EXTS,EXTS);

  for (pd=pds,pg=p->pg;pd<mpd;++pd,++pg) {
    if (NULL!=pg->vmt)
      pg->vmt->set(pg,ps,hDlg);
  }
}

///////////////////////////////////////////////////////////////////////////////
static void EnableReplayGain(HWND hDlg)
{
  int enabled=!IsDlgButtonChecked(hDlg,IDC_CHECKBOX_RG_OFF);

  // weight
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_RG_WEIGHT_ALBUM),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_RG_WEIGHT_TRACK),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_RG_WEIGHT_LABEL),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_RG_WEIGHT),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_SLIDER_RG_WEIGHT),enabled);

#if 0 // {
  // preamp
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_RG_PREAMP_LABEL),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_RG_PREAMP),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_SLIDER_RG_PREAMP),enabled);
#endif // }

  // comment
  EnableWindow(GetDlgItem(hDlg,IDC_CHECKBOX_RG_COMMENT),enabled);
}

static void EnableVis(HWND hDlg)
{
  int bVis=IsDlgButtonChecked(hDlg,IDC_CHECKBOX_WINAMP_VIS);
  int bDelay=bVis&&IsDlgButtonChecked(hDlg,IDC_CHECKBOX_WINAMP_VIS_DELAY);

  EnableWindow(GetDlgItem(hDlg,IDC_CHECKBOX_WINAMP_VIS_DELAY),bVis);
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_WINAMP_VIS_DELAY_MS),bDelay);
  EnableWindow(GetDlgItem(hDlg,IDC_SLIDER_WINAMP_VIS_DELAY_MS),bDelay);
}

///////////////////////////////////////////////////////////////////////////////
BOOL CALLBACK GeneralDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  priv_t *p=WM_INITDIALOG==uMsg?(void *)lParam
      :(void *)GetWindowLongPtr(hDlg,GWLP_USERDATA);

  switch(uMsg) {
  case WM_INITDIALOG:
    PrefsEnableThemeDialogTexture(
      hDlg,             // _In_  HWND hwnd,
      ETDT_ENABLETAB    // _In_  DWORD dwFlags
    );

    SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);
    priv_init(p,p->ps,hDlg);
    EnableReplayGain(hDlg);
    EnableVis(hDlg);
    return TRUE;
  case WM_COMMAND:
    switch(LOWORD(wParam)) {
    case IDOK:
      priv_save(p,p->ps,hDlg);
      return TRUE;
    case IDC_CHECKBOX_RG_OFF:
      EnableReplayGain(hDlg);
    case IDC_CHECKBOX_WINAMP_VIS:
    case IDC_CHECKBOX_WINAMP_VIS_DELAY:
      EnableVis(hDlg);
      return TRUE;
    case IDC_BUTTON_DEFAULT:
      priv_display(p,(prefs_t *)lParam,hDlg);
      EnableReplayGain(hDlg);
      EnableVis(hDlg);
      return TRUE;
    default:
      break;
    }

    break;
  case WM_NOTIFY:
    switch (LOWORD(wParam)) {
#define PREFS_NOTIFY_SLIDER
#include "slider.h"
    default:
      break;
    }

    break;
  default:
    break;
  }

  return FALSE;
}
