/*
 * prefs_yasapi.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "prefdef.h"
#if defined (PREFS_INTEGRATE) // [
#include <uxtheme.h>

//#define PREFS_INTEGRATE

///////////////////////////////////////////////////////////////////////////////
typedef prefs_yasapi_t priv_t;

static prefdef_t pds[]=INTEGRATE_PDS();
static prefdef_t *mpd=pds+PREFS_INTEGRATE_N;

///////////////////////////////////////////////////////////////////////////////
void prefs_integrate_load(prefs_t *ps)
{
  prefdef_t *pd;

  for (pd=pds;pd<mpd;++pd)
    prefs_read_play(ps,pd);
}

///////////////////////////////////////////////////////////////////////////////
//#define PREFS_INTEGRATE_WITH_SLIDER
#if defined (PREFS_INTEGRATE_WITH_SLIDER) // [
#define PREFS_SYNC_SLIDER
#include "slider.h"
#endif // ]

static void priv_init(priv_t *p, prefs_t *ps, HWND hDlg)
{
  prefdef_t *pd;
  prefgui_t *pg;
#if defined (DEBUG) // [
  int i=0;
#endif // ]

/*
DVWRITELN("=== offsetof(prefs_t,yasapi.enabled): %d",
    offsetof(prefs_t,yasapi.enabled));
*/
/*
DVWRITELN("=== prefs_t,yasapi.enabled: %d",
    offsetof(prefs_t,yasapi.enabled));
    */
  for (pd=pds,pg=p->pg;pd<mpd;++pd,++pg) {
    if (NULL!=pd->init) {
DVWRITELN("=== i=%d (%p %p) {",i,ps,pd->init);
      pd->init(pg,pd,ps,hDlg);
DVWRITELN("=== i=%d (%p %p) }",i,ps,pd->init);

#if defined (PREFS_INTEGRATE_WITH_SLIDER) // [
      switch (pg->id) {
#define PREFS_INIT_SLIDER
#include "slider.h"
      default:
        break;
      }
#endif // ]
#if defined (DEBUG) // [
      ++i;
#endif // ]
    }
  }
}

static void priv_save(priv_t *p, prefs_t *ps, HWND hDlg)
{
  prefdef_t *pd;
  prefgui_t *pg;

  for (pd=pds,pg=p->pg;pd<mpd;++pd,++pg) {
    if (pg->vmt) {
      pg->vmt->get(pg,ps,hDlg);
      prefs_write(ps,pd);
      DVWRITELN("field=\"%s.%s\" saved",pd->group,pd->name);
    }
    else
      DVWRITELN("field=\"%s.%s\" nop",pd->group,pd->name);
  }
}

static void priv_display(priv_t *p, prefs_t *ps, HWND hDlg)
{
  prefdef_t *pd;
  prefgui_t *pg;

  for (pd=pds,pg=p->pg;pd<mpd;++pd,++pg) {
    if (NULL!=pg->vmt)
      pg->vmt->set(pg,ps,hDlg);
  }
}

///////////////////////////////////////////////////////////////////////////////
#if 0 // [
static void EnableSox(HWND hDlg)
{
  int enabled=0<SendDlgItemMessage(hDlg,IDC_COMBO_SOX_MODE,CB_GETCURSEL,0,0);

  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_SOX_RATE_LABEL),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_COMBO_SOX_RATE),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_SOX_DITHER_LABEL),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_COMBO_SOX_DITHER),enabled);
}

static void SetForce(HWND hDlg, int enabled)
{
  int checked=!IsDlgButtonChecked(hDlg,IDC_CHECKBOX_SOX_RATE_MIN)
      &&!IsDlgButtonChecked(hDlg,IDC_CHECKBOX_SOX_RATE_MUL);

  CheckDlgButton(hDlg,IDC_CHECKBOX_SOX_RATE_FORCE,checked);
  EnableWindow(GetDlgItem(hDlg,IDC_CHECKBOX_SOX_RATE_FORCE),enabled&&!checked);
}

static void EnableSoxRate(HWND hDlg)
{
  int enabled=0<SendDlgItemMessage(hDlg,IDC_COMBO_SOX_MODE,CB_GETCURSEL,0,0)
      &&0<SendDlgItemMessage(hDlg,IDC_COMBO_SOX_RATE,CB_GETCURSEL,0,0);

  EnableWindow(GetDlgItem(hDlg,IDC_COMBO_SOX_RATE_Q),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_CHECKBOX_SOX_RATE_MIN),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_CHECKBOX_SOX_RATE_MUL),enabled);
  SetForce(hDlg,enabled);
}
#endif // ]

static void Enable(HWND hDlg)
{
#if 0 // [
  EnableSox(hDlg);
  EnableSoxRate(hDlg);
#else // ] [
#if ! defined (PREFS_YASAPI) // [
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_YASAPI_GROUP),FALSE);
  EnableWindow(GetDlgItem(hDlg,IDC_CHECKBOX_YASAPI_ENABLED),FALSE);
#endif // ]
#if ! defined (PREFS_YAS) // [
  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_YAS_GROUP),FALSE);
  EnableWindow(GetDlgItem(hDlg,IDC_CHECKBOX_YAS_ENABLED),FALSE);
#endif // ]
#endif // ]
}

#if 0 && defined (PREFS_AUDIO_DRC) // [
///////////////////////////////////////////////////////////////////////////////
static void EnableAudioDrc(HWND hDlg)
{
  int enabled=IsDlgButtonChecked(hDlg,IDC_CHECKBOX_AUDIO_DRC_ON);

  EnableWindow(GetDlgItem(hDlg,IDC_STATIC_AUDIO_AC3_DRC_SCALE),enabled);
  EnableWindow(GetDlgItem(hDlg,IDC_SLIDER_AUDIO_AC3_DRC_SCALE),enabled);
}
#endif // ]

///////////////////////////////////////////////////////////////////////////////
BOOL CALLBACK IntegrateDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  priv_t *p=WM_INITDIALOG==uMsg?(void *)lParam
      :(void *)GetWindowLongPtr(hDlg,GWLP_USERDATA);

  switch(uMsg) {
  case WM_INITDIALOG:
    PrefsEnableThemeDialogTexture(
      hDlg,             // _In_  HWND hwnd,
      ETDT_ENABLETAB    // _In_  DWORD dwFlags
    );

    SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);

#if 0 // [
    if (p->ps)
      priv_init(p,p->ps,hDlg);
#else // ] [
    priv_init(p,p->ps,hDlg);
#endif // ]

    Enable(hDlg);
#if 0 && defined (PREFS_AUDIO_DRC) // [
    EnableAudioDrc(hDlg);
#endif // ]
    return TRUE;
#if 0 // {
  // http://www.winprog.org/tutorial/dlgfaq.html
  case WM_CTLCOLORDLG:
    return (BOOL)CreateSolidBrush(RGB(255,255,255));
#endif // }
  case WM_COMMAND:
    switch(LOWORD(wParam)) {
    case IDOK:
#if 0 // [
DVWRITELN("=== yasapi.force_stero=%d"
        ", yasapi.drc.enabled=%d"
        ", yasapi.drc.scale=%0.2f ============================",
        p->ps->yasapi.force_stereo,
        p->ps->yasapi.drc.enabled,
        AUDIO_DRC_Q*p->ps->yasapi.drc.scale);
#endif // ]
#if 0 // [
      if (p->ps)
        priv_save(p,p->ps,hDlg);
#else // ] [
      priv_save(p,p->ps,hDlg);
#endif // ]
#if 0 // [
DVWRITELN("=== yasapi.force_stero=%d"
        ", yasapi.drc.enabled=%d"
        ", yasapi.drc.scale=%0.2f ============================",
        p->ps->yasapi.force_stereo,
        p->ps->yasapi.drc.enabled,
        AUDIO_DRC_Q*p->ps->yasapi.drc.scale);
#endif // ]
      return TRUE;
#if 0 // [
    case IDC_COMBO_SOX_MODE:
      Enable(hDlg);
      return TRUE;
    case IDC_COMBO_SOX_RATE:
      EnableSoxRate(hDlg);
      return TRUE;
#if 0 && defined (PREFS_AUDIO_DRC) // [
    case IDC_CHECKBOX_AUDIO_DRC_ON:
      EnableAudioDrc(hDlg);
      return TRUE;
#endif // ]
#endif // ]
    case IDC_BUTTON_DEFAULT:
      priv_display(p,(prefs_t *)lParam,hDlg);
      Enable(hDlg);
#if 0 && defined (PREFS_AUDIO_DRC) // [
      EnableAudioDrc(hDlg);
#endif // ]
      return TRUE;
#if 0 // [
    case IDC_CHECKBOX_SOX_RATE_MIN:
    case IDC_CHECKBOX_SOX_RATE_MUL:
      SetForce(hDlg,1);
      return TRUE;
    case IDC_CHECKBOX_SOX_RATE_FORCE:
      if (IsDlgButtonChecked(hDlg,IDC_CHECKBOX_SOX_RATE_FORCE)) {
        CheckDlgButton(hDlg,IDC_CHECKBOX_SOX_RATE_MIN,0);
        CheckDlgButton(hDlg,IDC_CHECKBOX_SOX_RATE_MUL,0);
        EnableWindow(GetDlgItem(hDlg,IDC_CHECKBOX_SOX_RATE_FORCE),0);
      }

      return TRUE;
#endif // ]
    default:
      break;
    }

    break;
#if defined (PREFS_INTEGRATE_WITH_SLIDER) // [
  case WM_NOTIFY:
    switch (LOWORD(wParam)) {
#define PREFS_NOTIFY_SLIDER
#include "slider.h"
    default:
      break;
    }

    break;
#endif // ]
  default:
    break;
  }

  return FALSE;
}
#endif // ]
