/*
 * prefs_tabs.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
// http://social.msdn.microsoft.com/forums/en-US/vcgeneral/thread/c06503f2-72f6-4898-b573-d6c6714fd7be
// http://forums.codeguru.com/showthread.php?102573-Setting-Background-color-of-a-Dialog-using-Win32-NO-MFC.
// http://social.msdn.microsoft.com/forums/en-US/vcgeneral/thread/e4fd1c4b-4407-4743-a74f-7d78338c17d2/
#include <windows.h>
#include <commctrl.h>
#include "prefs.h"

DLGTEMPLATE *load_template(HMODULE hMod, int id)
{
  HRSRC hRsrc;
  HGLOBAL hGlb;
  DLGTEMPLATE *res;

  if (NULL==(hRsrc=FindResource(hMod,MAKEINTRESOURCE(id),RT_DIALOG))) {
#if defined (DEBUG) // {
    fprintf(stderr,"Error finding resource.\n");
#endif // }
    return NULL;
  }

  if (NULL==(hGlb=LoadResource(hMod,hRsrc))) {
#if defined (DEBUG) // {
    fprintf(stderr,"Error loading resource.\n");
#endif // }
    return NULL;
  }

  if (NULL==(res=(DLGTEMPLATE *)LockResource(hGlb))) {
#if defined (DEBUG) // {
    fprintf(stderr,"Error locking resource.\n");
#endif // }
    return NULL;
  }

  return res;
}

#if defined (TABS_INIT_ADAPT) // [
#error TABS_INIT_ADAPT defined
// https://msdn.microsoft.com/en-us/library/windows/desktop/hh298366(v=vs.85).aspx
tabs_t *tabs_init(tabs_t *tc, HWND hDlg, int id, int n, tcp_t *tcp,
    HWND *hPages, char *page)
{
  HMODULE hMod=GetModuleHandleA(MODULE);
  TCITEMW tci;
  tcp_t *cur;
  int i;
  DLGTEMPLATE *res;
  RECT rc={ 0,0,1,1 };

  // Create the tab control. Note that g_hInst is a global 
  // instance handle. 
  tc->hTab=CreateWindowW( 
    WC_TABCONTROL,
    L"", 
    WS_CHILD|WS_CLIPSIBLINGS|WS_VISIBLE, 
    0,
    0,
    100,
    100, 
    hDlg,
    NULL,
    hMod,
    NULL 
  );

  if (NULL==tc->hTab)
    goto tabcontrol;

  // Add a tab for each of the three child dialog boxes. 
  tci.mask=TCIF_TEXT|TCIF_IMAGE; 
  tci.iImage=-1;

  for (cur=tcp,i=0;NULL!=cur->title;++cur,++i) {
    if (NULL==(res=load_template(hMod,cur->id)))
      continue;

    tci.pszText=cur->title;
    SendMessageW(tc->hTab,TCM_INSERTITEMW,cur-tcp,(LPARAM)&tci);

    hPages[i]=CreateDialogIndirectParamW(
      hMod,                     // _In_opt_  HINSTANCE hInstance,
      res,                      // _In_      LPCDLGTEMPLATE lpTemplate,
      tc->hTab,                 // _In_opt_  HWND hWndParent,
      cur->proc,                // _In_opt_  DLGPROC lpDialogFunc,
      (LPARAM)(page+cur->offs)	// _In_      LPARAM lParamInit
    );

    if (rc.right<res->cx)
      rc.right=res->cx;

    if (rc.bottom<res->cy)
      rc.bottom=res->cy;
  }

  // Map the rectangle from dialog box units to pixels.
  MapDialogRect(hDlg,&rc);

  return tc;
tabcontrol:
  return NULL;
}
#else // ] [
tabs_t *tabs_init(tabs_t *tc, HWND hDlg, int id, int n, tcp_t *tcp,
    HWND *hPages, char *page)
{
  HMODULE hMod=GetModuleHandleA(MODULE);
  HWND hTab=GetDlgItem(hDlg,id);
  tcp_t *cur;
  int i;
  DLGTEMPLATE *res;
#if defined (TABS_INIT_ADAPT) // [
  RECT rc={ 0,0,1,1 };
  TCITEMW tci;

  // Retrieves the coordinates of a window's client area. The client
  // coordinates specify the upper-left and lower-right corners of the
  // client area. Because client coordinates are relative to the
  // upper-left corner of a window's client area, the coordinates of
  // the upper-left corner are (0,0). 
  // https://msdn.microsoft.com/en-us/library/windows/desktop/ms633503(v=vs.85).aspx
  // we don't need GetClientRect() because the upper left corner is
  // (0,0) anyway.
  //GetClientRect(hTab,&rc);

  // Calculates a tab control's display area given a window rectangle,
  // or calculates the window rectangle that would correspond to a
  // specified display area.
  // https://msdn.microsoft.com/en-us/library/windows/desktop/bb760645(v=vs.85).aspx
#if 0 // [
  TabCtrl_AdjustRect(hTab,FALSE,&rc);
#else // ] [
  TabCtrl_AdjustRect(hTab,TRUE,&rc);
#endif // ]

#if 0 // [
  // Converts the specified dialog box units to screen units (pixels).
  // The function replaces the coordinates in the specified RECT structure
  // with the converted coordinates, which allows the structure to be
  // used to create a dialog box or position a control within a dialog box. 
  // https://msdn.microsoft.com/en-us/library/windows/desktop/ms645502(v=vs.85).aspx
  MapDialogRect(
    hDlg,       // _In_    HWND   hDlg,
    &rc         // _Inout_ LPRECT lpRect
  );

  // The MapWindowPoints function converts (maps) a set of points from
  // a coordinate space relative to one window to a coordinate space
  // relative to another window.
  // https://msdn.microsoft.com/en-us/library/windows/desktop/dd145046(v=vs.85).aspx
  MapWindowPoints(
    hTab,               // _In_    HWND    hWndFrom,
    hDlg,               // _In_    HWND    hWndTo,
    (LPPOINT)&rc,       // _Inout_ LPPOINT lpPoints,
    2                   // _In_    UINT    cPoints
  );
#endif // ]
#else // ] [
  TCITEM tci;
#endif // ]

  tc->id=id;
  tc->hPages=hPages;

  ZERO(&tci);
  tci.mask=TCIF_TEXT;

  for (cur=tcp,i=0;NULL!=cur->title;++cur,++i) {
    if (NULL==(res=load_template(hMod,cur->id)))
      continue;

    tci.pszText=cur->title;
#if defined (TABS_INIT_ADAPT) // [
    SendMessageW(hTab,TCM_INSERTITEMW,cur-tcp,(LPARAM)&tci);
#else // ] [
    TabCtrl_InsertItem(hTab,cur-tcp,&tci);
#endif // ]
DVWRITELN("page: %p, cur->offs: %d",page, cur->offs);

    hPages[i]=CreateDialogIndirectParamW(
      hMod,                     // _In_opt_  HINSTANCE hInstance,
      res,                      // _In_      LPCDLGTEMPLATE lpTemplate,
      hTab,                     // _In_opt_  HWND hWndParent,
      cur->proc,                // _In_opt_  DLGPROC lpDialogFunc,
      (LPARAM)(page+cur->offs)	// _In_      LPARAM lParamInit
    ); 

#if defined (TABS_INIT_ADAPT) // [
#if 1 // [
    SetWindowPos(hPages[i],HWND_TOP,rc.left,rc.top,0,0,
        SWP_NOSIZE|SWP_NOZORDER);
#else // ] [
    SetWindowPos(hPages[i],HWND_TOP,2,28,0,0,SWP_NOSIZE|SWP_NOZORDER);
#endif // ]
#else // ] [
    SetWindowPos(hPages[i],HWND_TOP,2,22,0,0,SWP_NOSIZE);
#endif // ]

    if (n==i) {
      TabCtrl_SetCurSel(hTab,n);
      tc->hCurPage=tc->hPages[n];
      ShowWindow(tc->hCurPage,SW_SHOW);
    }
    else
      ShowWindow(hPages[i],SW_HIDE);
  }

  return tc;
}
#endif // ]

void tabs_set_page(tabs_t *tc, HWND hDlg, int i)
{ 
#if defined (TABS_INIT_ADAPT) // [
  HWND hTab=tc->hTab;
#else // ] [
  HWND hTab=GetDlgItem(hDlg,tc->id);
#endif // ]
 
  if (tc->hCurPage!=NULL) 
    ShowWindow(tc->hCurPage,SW_HIDE);
 
  tc->hCurPage=tc->hPages[i];
  ShowWindow(tc->hCurPage,SW_SHOW);
  TabCtrl_SetCurSel(hTab,i);
}

void tabs_change(tabs_t *tc, HWND hDlg)
{ 
#if defined (TABS_INIT_ADAPT) // [
  HWND hTab=tc->hTab;
#else // ] [
  HWND hTab=GetDlgItem(hDlg,tc->id);
#endif // ]
  int i=TabCtrl_GetCurSel(hTab); 
 
  if (tc->hCurPage!=NULL) 
    ShowWindow(tc->hCurPage,SW_HIDE);
 
  tc->hCurPage=tc->hPages[i];
  ShowWindow(tc->hCurPage,SW_SHOW);
}
