/*
 * thread.c
 * Copyright (C) 2010 - 2019 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

#if defined (THREAD) // [
int ProducerCreate(SProducer *pProducer, In_Module *mod, decoder_t *dec)
{
  EngineCreate(pProducer,gcaProducerState);
  pProducer->pStateMin=gcaProducerState;
  pProducer->pState=gcaProducerState;
  pProducer->mod=mod;
  pProducer->dec=dec;
  pProducer->seek_needed=0;
  pProducer->killDecodeProducer=0;
  pProducer->hEvent=NULL;
  pProducer->hThread=INVALID_HANDLE_VALUE;

  return 0;
}

void ProducerDestroy(SProducer *pProducer)
{
  EngineDestroy(pProducer);
}

///////////////////////////////////////////////////////////////////////////////
static int ProducerCriticalSectionEnter(SProducer *pProducer)
{
  InitializeCriticalSection(&pProducer->cs);
  ++pProducer->pState;

  return 0;
}

static void ProducerCriticalSectionLeave(SProducer *pProducer)
{
  --pProducer->pState;
  DeleteCriticalSection(&pProducer->cs);
}

////////////////
static int ProducerEventEnter(SProducer *pProducer)
{
  // This is a MANUAL RESET event!
  if (!(pProducer->hEvent=CreateEventW(NULL,TRUE,FALSE,NULL))) {
    DMESSAGE("creating timer event");
    goto exit;
  }

  ++pProducer->pState;

  return 0;
exit:
  return -1;
}

static void ProducerEventLeave(SProducer *pProducer)
{
  --pProducer->pState;
  CloseHandle(pProducer->hEvent);
}

////////////////
static int ProducerCreateEnter(SProducer *pProducer)
{
  pProducer->hThread=CreateThread(
    NULL,                 // LPSECURITY_ATTRIBUTES   lpThreadAttributes,
    0u,                   // SIZE_T                  dwStackSize,
    ConsumerThreadProc,   // LPTHREAD_START_ROUTINE  lpStartAddress,
    pProducer,            // __drv_aliasesMem LPVOID lpParameter,
    0ul,                  // DWORD                   dwCreationFlags,
    NULL                  // LPDWORD                 lpThreadId
  );

  if (INVALID_HANDLE_VALUE==pProducer->hThread) {
    DMESSAGE("creating thread");
    goto exit;
  }

  ++pProducer->pState;

  return 0;
exit:
  return -1;
}

static void ProducerCreateLeave(SProducer *pProducer)
{
  --pProducer->pState;
  CloseHandle(pProducer->hThread);
}

////////////////
static int ProducerRunningEnter(SProducer *pProducer)
{
  return -1;
}

static void ProducerRunningLeave(SProducer *pProducer)
{
}

////////////////
IProducerState gcaProducerState[eProducerCount]={
#define PRODUCER_STATE(x) { \
  .pId=STRA(x), \
  .Enter=Producer ## x ## Enter, \
  .Leave=Producer ## x ## Leave, \
}, // [
#include "producer_state.h"
#undef PRODUCER_STATE // ]
};
#endif // ]
