/*
 * screen.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <math.h>
#include "plugin.h"

screen_t *screen_init(screen_t *s, In_Module *mod)
{
  s->mod=mod;
  s->vo=NULL;
  s->open=0;
  s->width=0;
  s->height=0;
  s->ar=0.0;

  return s;
}

screen_t *screen_cleanup(screen_t *s)
{
  screen_close(s);

  return s;
}

void screen_open(screen_t *s, int width, int height, double ar)
{
  vout_t *vo;

  vo=s->vo=vout_get(s->mod->hMainWindow);
  //DVWRITELN("vo=%p",s->vo);

  if (0!=s->open&&(s->width!=width||s->height!=height||1e-4<fabs(s->ar-ar))) {
    vout_close(s->vo);
    s->open=0;
  }

  if (0==s->open) {
    //DVWRITELN("vo=%p",s->vo); DGETC();
    vout_open(vo,width,height,0.0==ar?0.0:1.0/ar);
    //DVWRITELN("vo=%p",s->vo); DGETC();
    vout_set_caption(vo,CAPTION);
    //DVWRITELN("vo=%p",s->vo); DGETC();
    s->width=width;
    s->height=height;
    s->ar=ar;
    s->open=1;
  }
}

#if defined (FFSOX_DEPRECATED_PICTURE_CROP) // [
void screen_draw(screen_t *s, AVFrame *frame)
{
  //DVWRITELN("vo=%p",s->vo); DGETC();
  vout_draw(s->vo,
      frame->data[0],frame->linesize[0],
      frame->data[2],frame->linesize[2],
      frame->data[1],frame->linesize[1]);
  //DVWRITELN("vo=%p",s->vo); DGETC();
}
#else // ] [
void screen_draw(screen_t *s, AVPicture *pict)
{
  //DVWRITELN("vo=%p",s->vo); DGETC();
  vout_draw(s->vo,
      pict->data[0],pict->linesize[0],
      pict->data[2],pict->linesize[2],
      pict->data[1],pict->linesize[1]);
  //DVWRITELN("vo=%p",s->vo); DGETC();
}
#endif // ]

void screen_close(screen_t *s)
{
  if (0!=s->open) {
    vout_close(s->vo);
    s->open=0;
  }
}

void screen_done(screen_t *s)
{
#if 0 // {
  screen_close(s);
#endif // }
}
