/*
 * bs1770_lufs.c
 * Copyright (C) 2011, 2012 Peter Belkner <pbelkner@snafu.de>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <math.h>
#include <stdio.h>
#include <bs1770.h>

bs1770_lufs_t *bs1770_lufs_init(bs1770_lufs_t *lufs, int active)
{
  const bs1770_ps_t *ps;

  bs1770_lufs_set_active(lufs,active);
  ps=bs1770_lufs_ps_default();

  if (NULL==bs1770_hist_init(&lufs->hist,ps))
    goto hist;

  if (NULL==bs1770_aggr_init(&lufs->aggr,ps,&lufs->hist,NULL))
    goto aggr;

  if (NULL==bs1770_init(&lufs->bs1770,&lufs->aggr,NULL))
    goto bs1770;

  return lufs;
bs1770:
  bs1770_aggr_cleanup(&lufs->aggr);
aggr:
  bs1770_hist_cleanup(&lufs->hist);
hist:
  return NULL;
}

bs1770_lufs_t *bs1770_lufs_cleanup(bs1770_lufs_t *lufs)
{
  bs1770_cleanup(&lufs->bs1770);
  bs1770_aggr_cleanup(&lufs->aggr);
  bs1770_hist_cleanup(&lufs->hist);

  return lufs;
}

void bs1770_lufs_set_active(bs1770_lufs_t *lufs, int active)
{
  lufs->vmt=0==active?&bs1770_lufs_null_vmt:&bs1770_lufs_vmt;
}

#if defined (BS1770_LUFS_X) // {
void bs1770_lufs_reset(bs1770_lufs_t *lufs, int active)
#else // } {
void bs1770_lufs_reset(bs1770_lufs_t *lufs, double fs, int channels,
    int active)
#endif // }
{
#if 0
  bs1770_reset(&lufs->bs1770);
  bs1770_aggr_reset(&lufs->aggr);
  bs1770_hist_reset(&lufs->hist);
#else
  bs1770_flush(&lufs->bs1770);
  bs1770_hist_reset(&lufs->hist);
#endif
#if ! defined (BS1770_LUFS_X) // {
  lufs->fs=fs;
  lufs->channels=channels;
#endif // }
  bs1770_lufs_set_active(lufs,active);
}

///////////////////////////////////////////////////////////////////////////////
#if defined (BS1770_LUFS_X) // {
void bs1770_lufs_null_add_sample_i16(bs1770_lufs_t *lufs, double fs,
    int channels, bs1770_sample_i16_t sample)
{
}

void bs1770_lufs_null_add_sample_i32(bs1770_lufs_t *lufs, double fs,
    int channels, bs1770_sample_i32_t sample)
{
}

void bs1770_lufs_null_add_sample_f32(bs1770_lufs_t *lufs, double fs,
    int channels, bs1770_sample_f32_t sample)
{
}

void bs1770_lufs_null_add_sample_f64(bs1770_lufs_t *lufs, double fs,
    int channels, bs1770_sample_f64_t sample)
{
}
#else // } {
void bs1770_lufs_null_add_sample_i16(bs1770_lufs_t *lufs,
    bs1770_sample_i16_t sample)
{
}

void bs1770_lufs_null_add_sample_i32(bs1770_lufs_t *lufs,
    bs1770_sample_i32_t sample)
{
}

void bs1770_lufs_null_add_sample_f32(bs1770_lufs_t *lufs,
    bs1770_sample_f32_t sample)
{
}

void bs1770_lufs_null_add_sample_f64(bs1770_lufs_t *lufs,
    bs1770_sample_f64_t sample)
{
}
#endif // }

double bs1770_lufs_null_get_cur(bs1770_lufs_t *lufs, double reference)
{
  return reference;
}

double bs1770_lufs_null_get_cur_gain(bs1770_lufs_t *lufs, double reference)
{
  return 0.0;
}

void bs1770_lufs_null_get_cur_gain_db(bs1770_lufs_t *lufs, double reference,
    double *db)
{
}

void bs1770_lufs_null_get_cur_gain_q(bs1770_lufs_t *lufs, double reference,
    double *q)
{
}

double bs1770_lufs_null_get_final(bs1770_lufs_t *lufs, double reference,
    int active)
{
  return reference;
}

double bs1770_lufs_null_get_final_gain(bs1770_lufs_t *lufs, double reference,
    int active)
{
  return 0.0;
}

bs1770_lufs_vmt_t bs1770_lufs_null_vmt={
#if defined (_MSC_VER) // {
  bs1770_lufs_null_add_sample_i16,
  bs1770_lufs_null_add_sample_i32,
  bs1770_lufs_null_add_sample_f32,
  bs1770_lufs_null_add_sample_f64,

  bs1770_lufs_null_get_cur,
  bs1770_lufs_null_get_cur_gain,
  bs1770_lufs_null_get_cur_gain_db,
  bs1770_lufs_null_get_cur_gain_q,
  bs1770_lufs_null_get_final,
  bs1770_lufs_null_get_final_gain
#else // } {
  .add_sample_i16=bs1770_lufs_null_add_sample_i16,
  .add_sample_i32=bs1770_lufs_null_add_sample_i32,
  .add_sample_f32=bs1770_lufs_null_add_sample_f32,
  .add_sample_f64=bs1770_lufs_null_add_sample_f64,

  .get_cur=bs1770_lufs_null_get_cur,
  .get_cur_gain=bs1770_lufs_null_get_cur_gain,
  .get_cur_gain_db=bs1770_lufs_null_get_cur_gain_db,
  .get_cur_gain_q=bs1770_lufs_null_get_cur_gain_q,
  .get_final=bs1770_lufs_null_get_final,
  .get_final_gain=bs1770_lufs_null_get_final_gain
#endif // }
};

///////////////////////////////////////////////////////////////////////////////
#if defined (BS1770_LUFS_X) // {
void bs1770_lufs_add_sample_i16(bs1770_lufs_t *lufs, double fs,
    int channels, bs1770_sample_i16_t sample)
{
  bs1770_add_sample_i16(&lufs->bs1770,fs,channels,sample);
}

void bs1770_lufs_add_sample_i32(bs1770_lufs_t *lufs, double fs,
    int channels, bs1770_sample_i32_t sample)
{
  bs1770_add_sample_i32(&lufs->bs1770,fs,channels,sample);
}

void bs1770_lufs_add_sample_f32(bs1770_lufs_t *lufs, double fs,
    int channels, bs1770_sample_f32_t sample)
{
  bs1770_add_sample_f32(&lufs->bs1770,fs,channels,sample);
}

void bs1770_lufs_add_sample_f64(bs1770_lufs_t *lufs, double fs,
    int channels, bs1770_sample_f64_t sample)
{
  bs1770_add_sample_f64(&lufs->bs1770,fs,channels,sample);
}
#else // } {
void bs1770_lufs_add_sample_i16(bs1770_lufs_t *lufs,
    bs1770_sample_i16_t sample)
{
  bs1770_add_sample_i16(&lufs->bs1770,lufs->fs,lufs->channels,sample);
}

void bs1770_lufs_add_sample_i32(bs1770_lufs_t *lufs,
    bs1770_sample_i32_t sample)
{
  bs1770_add_sample_i32(&lufs->bs1770,lufs->fs,lufs->channels,sample);
}

void bs1770_lufs_add_sample_f32(bs1770_lufs_t *lufs,
    bs1770_sample_f32_t sample)
{
  bs1770_add_sample_f32(&lufs->bs1770,lufs->fs,lufs->channels,sample);
}

void bs1770_lufs_add_sample_f64(bs1770_lufs_t *lufs,
    bs1770_sample_f64_t sample)
{
  bs1770_add_sample_f64(&lufs->bs1770,lufs->fs,lufs->channels,sample);
}
#endif // }

double bs1770_lufs_get_cur(bs1770_lufs_t *lufs, double reference)
{
  return bs1770_hist_get_lufs(&lufs->hist,reference);
}

double bs1770_lufs_get_cur_gain(bs1770_lufs_t *lufs, double reference)
{
  return reference-bs1770_lufs_get_cur(lufs,reference);
}

void bs1770_lufs_get_cur_gain_db(bs1770_lufs_t *lufs, double reference,
    double *db)
{
  *db=bs1770_lufs_get_cur_gain(lufs,reference);
}

void bs1770_lufs_get_cur_gain_q(bs1770_lufs_t *lufs, double reference,
    double *q)
{
  *q=BS1770_DB2Q(bs1770_lufs_get_cur_gain(lufs,reference));
}

double bs1770_lufs_get_final(bs1770_lufs_t *lufs, double reference,
    int active)
{
  double final;

  bs1770_flush(&lufs->bs1770);
  final=bs1770_hist_get_lufs(&lufs->hist,reference);
  bs1770_hist_reset(&lufs->hist);
  bs1770_lufs_set_active(lufs,active);

  return final;
}

double bs1770_lufs_get_final_gain(bs1770_lufs_t *lufs, double reference,
    int active)
{
  return reference-bs1770_lufs_get_final(lufs,reference,active);
}

bs1770_lufs_vmt_t bs1770_lufs_vmt={
#if defined (_MSC_VER) // {
  bs1770_lufs_add_sample_i16,
  bs1770_lufs_add_sample_i32,
  bs1770_lufs_add_sample_f32,
  bs1770_lufs_add_sample_f64,

  bs1770_lufs_get_cur,
  bs1770_lufs_get_cur_gain,
  bs1770_lufs_get_cur_gain_db,
  bs1770_lufs_get_cur_gain_q,
  bs1770_lufs_get_final,
  bs1770_lufs_get_final_gain
#else // } {
  .add_sample_i16=bs1770_lufs_add_sample_i16,
  .add_sample_i32=bs1770_lufs_add_sample_i32,
  .add_sample_f32=bs1770_lufs_add_sample_f32,
  .add_sample_f64=bs1770_lufs_add_sample_f64,

  .get_cur=bs1770_lufs_get_cur,
  .get_cur_gain=bs1770_lufs_get_cur_gain,
  .get_cur_gain_db=bs1770_lufs_get_cur_gain_db,
  .get_cur_gain_q=bs1770_lufs_get_cur_gain_q,
  .get_final=bs1770_lufs_get_final,
  .get_final_gain=bs1770_lufs_get_final_gain
#endif // }
};
