/*
 * ffsox_ar.c
 * Copyright (C) 2011, 2012 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include "ffsox.h"

#if defined (DEBUG) // [
///////////////////////////////////////////////////////////////////////////////
//#define FFSOX_AR_STATE_MNEMONIC
#if defined (FFSOX_AR_STATE_MNEMONIC) // [
#define FFSOX_AR_STATE_MNEMONIC_OR
static const char *ffsox_ar_state_mnemonic(const ffsox_ar_t *ar)
{
  static struct { int state; const char *mnemonic; } map[]={
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
    { FFSOX_AR_READ_PACKET, "FFSOX_AR_READ_PACKET" },
    { FFSOX_AR_RECEIVE_FRAME, "FFSOX_AR_RECEIVE_FRAME" },
#if defined (FFSOX_AUDIO_STEREO) // [
    { FFSOX_AR_CONVERT_FRAME, "FFSOX_AR_CONVERT_FRAME" },
#endif // ]
    { FFSOX_AR_FLUSH, "FFSOX_AR_FLUSH" },
#else // ] [
    { FFSOX_AR_RUN, "FFSOX_AR_RUN" },
    { FFSOX_AR_FLUSH, "FFSOX_AR_FLUSH" },
#endif // ]
    { FFSOX_AR_DONE, "FFSOX_AR_DONE" },
    { FFSOX_AR_ERROR, "FFSOX_AR_ERROR" },
#if ! defined (FFSOX_AR_STATE_MNEMONIC_OR) // [
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
    { FFSOX_AR_MASK, "FFSOX_AR_MASK" },
    { FFSOX_AR_VALID, "FFSOX_AR_VALID" },
#else // ] [
    { FFSOX_AR_VALID, "FFSOX_AR_VALID" },
#endif // ]
#endif // ]
  };

  static int n=(sizeof map)/(sizeof *map);
#if 1 // [
  static char mnemonic[256];
  char *mp=mnemonic;
#endif // ]
  int i;

  for (i=0;i<n;++i) {
#if defined (FFSOX_AR_STATE_MNEMONIC_OR) // [
    if (map[i].state&ar->state) {
      if (mnemonic<mp)
        *mp++='|';

      sprintf(mp,map[i].mnemonic);
      mp+=strlen(mp);
      *mp='\0';
    }
#else // ] [
    if (map[i].state==ar->state)
      return map[i].mnemonic;
#endif // ]
  }

#if defined (FFSOX_AR_STATE_MNEMONIC_OR) // [
  return mnemonic<mp?mnemonic:"ERROR: out of range";
#else // ] [
  return "ERROR: out of range";
#endif // ]
}
#endif // ]
#endif // ]

///////////////////////////////////////////////////////////////////////////////
ffsox_ar_t *ffsox_ar_init(ffsox_ar_t *ar, double q, void *data,
    ffsox_ar_cbs_t *cbs)
{
  ar->data=data;
  ar->cbs=cbs;

#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
  ar->decode.frame=NULL;
#if defined (FFSOX_AUDIO_STEREO) // [
  ar->convert.frame=NULL;
#endif // ]

  // allocate frame only if not provided by the client.
  if (NULL==(ar->decode.frame=ffsox_ar_frame(ar))) {
    if (NULL==(ar->decode.frame=av_frame_alloc()))
      return NULL;
  }
#else // ] [
  ar->frame=NULL;

  // allocate frame only if not provided by the client.
  if (NULL==(ar->frame=ffsox_ar_frame(ar))) {
    if (NULL==(ar->frame=av_frame_alloc()))
      return NULL;
  }
#endif // ]

#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
#if defined (LIBFFSOX_BS1770) // [
  ffsox_ar_reset(ar,NULL,q,NULL);
#else // ] [
  ffsox_ar_reset(ar,NULL,q);
#endif // ]
#else // ] [
#if defined (LIBFFSOX_BS1770) // [
  ffsox_ar_reset(ar,q,NULL);
#else // ] [
  ffsox_ar_reset(ar,q);
#endif // ]
#endif // ]

  return ar;
}

ffsox_ar_t *ffsox_ar_cleanup(ffsox_ar_t *ar)
{
#if 0 // [
fprintf(stderr,"[%s:%d:%s]\n",pb_basenamea(__FILE__),__LINE__,__func__);
#endif // ]
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
#if defined (FFSOX_AUDIO_STEREO) // [
  if (ar->convert.frame)
    av_frame_free(&ar->convert.frame);

  if (ar->convert.swr_ctx)
    swr_free(&ar->convert.swr_ctx);
#endif // ]

  if (ar->decode.frame&&!ffsox_ar_frame(ar))
    av_frame_free(&ar->decode.frame);
#else // ] [
  if (NULL!=ar->frame&&NULL==ffsox_ar_frame(ar)) {
    av_free(ar->frame);
    ar->frame=NULL;
  }
#endif // ]

  return ar;
}

#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
#if defined (LIBFFSOX_BS1770) // [
void ffsox_ar_reset(ffsox_ar_t *ar, ffi_st_t *ist, double q,
    bs1770_lufs_t *lufs)
#else // ] [
void ffsox_ar_reset(ffsox_ar_t *ar, ffi_st_t *ist, double q)
#endif // ]
#else // ] [
#if defined (LIBFFSOX_BS1770) // [
void ffsox_ar_reset(ffsox_ar_t *ar, double q, bs1770_lufs_t *lufs)
#else // ] [
void ffsox_ar_reset(ffsox_ar_t *ar, double q)
#endif // ]
#endif // ]
{
#if 0 // [
fprintf(stderr,"[%s:%d:%s]\n",pb_basenamea(__FILE__),__LINE__,__func__);
#endif // ]
  ar->q=q;
  ar->count=0;
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
  ar->ist=ist;
  ar->state=FFSOX_AR_READ_PACKET;
#else // ] [
  ar->state=FFSOX_AR_RUN;
#endif // ]
#if defined (LIBFFSOX_BS1770) // {
  ar->lufs=lufs;
#endif // }
#if defined (FFSOX_DEPRECATED_FREE_PACKET) // [
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
#if defined (FFSOX_AUDIO_STEREO) // [
  if (ar->convert.frame)
    av_frame_free(&ar->convert.frame);

  if (ar->convert.swr_ctx)
    swr_free(&ar->convert.swr_ctx);
#endif // ]

  av_init_packet(&ar->pkt);
  ar->pkt.data=NULL;
  ar->pkt.size=0;
#else // ] [
  av_init_packet(&ar->pkt1);
  ar->pkt1.data=NULL;
  ar->pkt1.size=0;
  // "pkt2" should be a shallow copy of "pkt1".
  ar->pkt2=ar->pkt1;
#endif // ]
#else // ] [
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
  FFSOX_ZERO(&ar->pkt);
#else // ] [
  FFSOX_ZERO(&ar->pkt1);
  FFSOX_ZERO(&ar->pkt2);
#endif // ]
#endif // ]
#if defined (FFSOX_AR_DELAY) // [
  ar->stopwatch.start.ui64=0ull;
  ar->stopwatch.end.ui64=0ull;
#endif // ]
}

AVFrame *ffsox_ar_frame_internal(ffsox_ar_t *ar)
{
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
#if defined (FFSOX_AUDIO_STEREO) // [
  return ar->state==FFSOX_AR_CONVERT_FRAME?ar->convert.frame:ar->decode.frame;
#else // ] [
  return ar->decode.frame;
#endif // ]
#else // ] [
  return ar->frame;
#endif // ]
}

#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
void ffsox_ar_got_frame(ffsox_ar_t *ar, AVFrame *frame)
{
  ar->cbs->got_frame(ar->data,ar,frame);
  ++ar->count;
}
#endif // ]

void ffsox_ar_next(ffsox_ar_t *ar)
{
  AVStream *st=ffsox_ar_stream(ar);
  int index=st->index;
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
  AVCodecContext *dc=ar->ist->dc1;
  int fferr;
#else // ] [
  AVCodecContext *dc=st->codec;
  int got_frame=0,len;
#endif // ]

#if 0 // [
MessageBox(
  NULL,              // _In_opt_ HWND    hWnd,
  "ffsox_ar_next"   // _In_opt_ LPCTSTR lpText,
  "Message",        // _In_opt_ LPCTSTR lpCaption,
  MB_OK             // _In_     UINT    uType
);
#endif // ]
  //avcodec_get_frame_defaults(ar->frame);

#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
  switch (FFSOX_AR_MASK&ar->state) {
  case FFSOX_AR_READ_PACKET:
  read_packet:
//DWRITELN("FFSOX_AR_READ_PACKET");
    /*
     * We read packets until we've found one of "our" stream.
     */
#if defined (FFSOX_DEPRECATED_FREE_PACKET) // [
    av_packet_unref(&ar->pkt);
#else // ] [
    av_free_packet(&ar->pkt);
    FFSOX_ZERO(&ar->pkt);
#endif // ]

    fferr=ffsox_ar_read(ar,&ar->pkt);
#if defined (FFSOX_AR_STATE_MNEMONIC) // [
DVWRITELN("packet read: %d",fferr);
#endif // ]

    if (0!=fferr) {
      /*
       * Error: no more packets to read, we should flush the decoder.
       */
      FFSOX_ZERO(&ar->pkt);
      ar->state=FFSOX_AR_FLUSH|FFSOX_AR_RECEIVE_FRAME;
#if defined (FFSOX_AR_STATE_MNEMONIC) // [
      DVWRITELN("state=%s",ffsox_ar_state_mnemonic(ar));
#endif // ]
      goto receive_frame;
    }
    else if (ar->pkt.stream_index!=index) {
      /*
       * Some other stream: discard the packet and loop back???
       * No, we should never go here because the packets received
       * by ffsox_ar_read() are from exact the same audio stream.
       */
#if defined (FFSOX_DEPRECATED_FREE_PACKET) // [
      av_packet_unref(&ar->pkt);
#else // ] [
      av_free_packet(&ar->pkt);
      FFSOX_ZERO(&ar->pkt);
#endif // ]
      goto read_packet;
    }
    else {
      /*
       * We've received a packet from "our" stream.
       */
      fferr=avcodec_send_packet(dc,&ar->pkt);
#if defined (FFSOX_AR_STATE_MNEMONIC) // [
DVWRITELN("send packet: %d",fferr);
#endif // ]

      if (fferr<0) {
#if defined (FFSOX_DEPRECATED_FREE_PACKET) // [
        av_packet_unref(&ar->pkt);
#else // ] [
        av_free_packet(&ar->pkt);
        FFSOX_ZERO(&ar->pkt);
#endif // ]
        ar->state=FFSOX_AR_ERROR;
#if defined (FFSOX_AR_STATE_MNEMONIC) // [
        DVWRITELN("state=%s",ffsox_ar_state_mnemonic(ar));
#endif // ]
        return;
      }
      else {
#if defined (FFSOX_DEPRECATED_FREE_PACKET) // [
        av_packet_unref(&ar->pkt);
#else // ] [
        av_free_packet(&ar->pkt);
        FFSOX_ZERO(&ar->pkt);
#endif // ]
        ar->state=FFSOX_AR_RECEIVE_FRAME;
        goto receive_frame;
      }
    }

    break;
  case FFSOX_AR_RECEIVE_FRAME:
  receive_frame:
//DMESSAGE("FFSOX_AR_RECEIVE_FRAME");
    fferr=avcodec_receive_frame(dc,ar->decode.frame);
#if defined (FFSOX_AR_STATE_MNEMONIC) // [
DVWRITELN("FFSOX_AR_RECEIVE_FRAME: %d",fferr);
#endif // ]

    if (fferr<0) {
//DMESSAGE("FFSOX_AR_RECEIVE_FRAME: error");
      /*
       * Error: no frame received.
       */
      if (FFSOX_AR_FLUSH&ar->state) {
        /*
         * We should exit the loop.
         */
#if defined (FFSOX_AR_STATE_MNEMONIC) // [
        //DVWRITELN("state=%s",ffsox_ar_state_mnemonic(ar));
#endif // ]
        ar->state=FFSOX_AR_DONE;
#if defined (FFSOX_AR_STATE_MNEMONIC) // [
        DVWRITELN("state=%s",ffsox_ar_state_mnemonic(ar));
#endif // ]
        return;
      }
      else {
        /*
         * We should read the next packet.
         */
        ar->state=FFSOX_AR_READ_PACKET;
        goto read_packet;
      }
    }
    else {
      /*
       * We've received a valid frame.
       */
      ar->decode.frame->pts=ar->decode.frame->best_effort_timestamp;
//DVMESSAGE("Here we go: %I64d ...",ar->decode.frame->pts);
//DMESSAGE("FFSOX_AR_RECEIVE_FRAME: valid");
#if 0 // [
MessageBoxA(
  NULL,              // _In_opt_ HWND    hWnd,
  "FFSOX_AR_RECEIVE_FRAME: valid", // _In_opt_ LPCTSTR lpText,
  "Message",        // _In_opt_ LPCTSTR lpCaption,
  MB_OK|MB_TASKMODAL             // _In_     UINT    uType
);
#endif // ]
#if defined (FFSOX_AUDIO_STEREO) // [
      if (ffsox_ar_stereo(ar)
          &&AV_CH_LAYOUT_STEREO!=ar->decode.frame->channel_layout) {
        if (!ar->convert.frame) {
          if (!(ar->convert.swr_ctx=swr_alloc())) {
            DMESSAGE("allocating context");
            ar->state=FFSOX_AR_ERROR;
#if defined (FFSOX_AR_STATE_MNEMONIC) // [
          DVWRITELN("state=%s",ffsox_ar_state_mnemonic(ar));
#endif // ]
            return;
          }

#if 0 // [
// https://www.ffmpeg.org/doxygen/4.0/group__lswr.html
av_opt_set_channel_layout(ar->convert.swr_ctx, "in_channel_layout",  ar->decode.frame->channel_layout, 0);
av_opt_set_channel_layout(ar->convert.swr_ctx, "out_channel_layout", AV_CH_LAYOUT_STEREO,  0);
//av_opt_set_int(swr, "in_sample_rate",     48000,                0);
//av_opt_set_int(swr, "out_sample_rate",    44100,                0);
//av_opt_set_sample_fmt(swr, "in_sample_fmt",  AV_SAMPLE_FMT_FLTP, 0);
//av_opt_set_sample_fmt(swr, "out_sample_fmt", AV_SAMPLE_FMT_S16,  0);
#endif // ]

          if (!(ar->convert.frame=av_frame_alloc())) {
            DMESSAGE("allocating frame");
            ar->state=FFSOX_AR_ERROR;
#if defined (FFSOX_AR_STATE_MNEMONIC) // [
            DVWRITELN("state=%s",ffsox_ar_state_mnemonic(ar));
#endif // ]
            return;
          }
        }

#if 1 // [
        ar->convert.frame->format=ar->decode.frame->format;
        ar->convert.frame->sample_rate=ar->decode.frame->sample_rate;
        ar->convert.frame->channel_layout=AV_CH_LAYOUT_STEREO;
#endif // ]
#if defined (FFSOX_AR_DELAY) // [
        GetSystemTimeAsFileTime(
          &ar->stopwatch.start.ft // _Out_ LPFILETIME lpSystemTimeAsFileTime
        );
#endif // ]

        if ((fferr=swr_convert_frame(ar->convert.swr_ctx,ar->convert.frame,
            ar->decode.frame))<0) {
          DVMESSAGE("converting frame: %s",av_err2str(fferr));
          ar->state=FFSOX_AR_ERROR;
#if defined (FFSOX_AR_STATE_MNEMONIC) // [
          DVWRITELN("state=%s",ffsox_ar_state_mnemonic(ar));
#endif // ]
          return;
        }

        /*
         * We need to inform our client in order to get the right time.
         */
#if defined (FFSOX_AR_DELAY) // [
        GetSystemTimeAsFileTime(
          &ar->stopwatch.end.ft   // _Out_ LPFILETIME lpSystemTimeAsFileTime
        );
DVWRITELN("deleay: %I64u %I64u %I64u",ar->stopwatch.start.ui64,ar->stopwatch.end.ui64,ar->stopwatch.end.ui64-ar->stopwatch.start.ui64);
fprintf(stderr,"deleay: %.0f\n",(double)(ar->stopwatch.end.ui64-ar->stopwatch.start.ui64));
#endif // ]
        ar->convert.frame->pts=ar->decode.frame->best_effort_timestamp;
        ffsox_ar_got_frame(ar,ar->convert.frame);
        ar->state=FFSOX_AR_CONVERT_FRAME;
#if defined (FFSOX_AR_STATE_MNEMONIC) // [
        DVWRITELN("state=%s",ffsox_ar_state_mnemonic(ar));
#endif // ]
        return;
      }
      else {
#endif // ]
        /*
         * We need to inform our client in order to get the right time.
         */
        ffsox_ar_got_frame(ar,ar->decode.frame);
#if defined (FFSOX_AR_STATE_MNEMONIC) // [
        DVWRITELN("state=%s",ffsox_ar_state_mnemonic(ar));
#endif // ]
        return;
#if defined (FFSOX_AUDIO_STEREO) // [
      }
#endif // ]
    }

    break;
#if defined (FFSOX_AUDIO_STEREO) // [
  case FFSOX_AR_CONVERT_FRAME:
//DWRITELN("FFSOX_AR_CONVERT_FRAME");
  //convert_frame:
    /*
     * We should read the next packet.
     */
    av_frame_unref(ar->convert.frame);
    ar->state=FFSOX_AR_RECEIVE_FRAME;
    goto receive_frame;
#endif // ]
  default:
    break;
  }
#else // ] [
  while (FFSOX_AR_RUN==ar->state) {
    got_frame=0;

    if (ar->pkt2.size<=0) {
#if defined (FFSOX_DEPRECATED_FREE_PACKET) // [
      av_packet_unref(&ar->pkt1);
#else // ] [
      av_free_packet(&ar->pkt1);
      FFSOX_ZERO(&ar->pkt1);
#endif // ]

      if (0!=ffsox_ar_read(ar,&ar->pkt1)) {
#if ! defined (FFSOX_DEPRECATED_FREE_PACKET) // [
        FFSOX_ZERO(&ar->pkt1);
        FFSOX_ZERO(&ar->pkt2);
#endif // ]
        ar->state=FFSOX_AR_FLUSH;
#if defined (FFSOX_AR_STATE_MNEMONIC) // [
        DVWRITELN("state=%s",ffsox_ar_state_mnemonic(ar));
#endif // ]
        break;
      }
      else if (ar->pkt1.stream_index!=index) {
#if defined (FFSOX_DEPRECATED_FREE_PACKET) // [
        av_packet_unref(&ar->pkt1);
#else // ] [
        av_free_packet(&ar->pkt1);
        FFSOX_ZERO(&ar->pkt1);
#endif // ]
        continue;
      }
      else {
        ar->pkt2=ar->pkt1;
      }
    }

    if ((len=avcodec_decode_audio4(dc,ar->frame,&got_frame,&ar->pkt2))<0) {
      //ar->state=FFSOX_AR_ERROR;
      //goto cleanup;
      // ignore and continue
      FFSOX_ZERO(&ar->pkt2);
      continue;
    }
  cleanup:
    ar->pkt2.size-=len;
    ar->pkt2.data+=len;

    if (got_frame) {
      ffsox_ar_got_frame(ar,ar->frame);
#if 0 // [
fprintf(stderr,"stereo=%d [%s:%d:%s]\n",ffsox_ar_stereo(ar),
    pb_basenamea(__FILE__),__LINE__,__func__);
#endif // ]
      //av_frame_unref(ar->frame);
      ++ar->count;
      return;
    }
  }

  if (FFSOX_AR_FLUSH==ar->state) {
    if (avcodec_decode_audio4(dc,ar->frame,&got_frame,&ar->pkt2)<0) {
      ar->state=FFSOX_AR_ERROR;
    }
    else if (got_frame) {
      ffsox_ar_got_frame(ar,ar->frame);
#if 0 // [
fprintf(stderr,"stereo=%d [%s:%d:%s]\n",ffsox_ar_stereo(ar),
    pb_basenamea(__FILE__),__LINE__,__func__);
#endif // ]
      //av_frame_unref(ar->frame);
      ++ar->count;
    }
    else {
      ar->state=FFSOX_AR_DONE;
    }
  }
#endif // ]
}
