/*
 * ffsox_ar_ffi_cbs.c
 * Copyright (C) 2011, 2012 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include "ffsox.h"

///////////////////////////////////////////////////////////////////////////////
static ffsox_ar_cbs_t cbs;

ffsox_ar_t *ffsox_ar_init_ffi(ffsox_ar_t *ar, double q, ffi_t *ffi)
{
  return ffsox_ar_init(ar,q,ffi,&cbs);
}

static AVFrame *priv_frame(void *data)
{
  return NULL;
}

static AVStream *priv_stream(void *data)
{
  return ((ffi_t *)data)->ast->st;
}

static int priv_read(void *data, AVPacket *pkt)
{
  return av_read_frame(((ffi_t *)data)->ic,pkt);
}

static void priv_got_frame(void *data, ffsox_ar_t *ar, AVFrame *frame)
{
}

#if defined (FFSOX_AUDIO_CUST) // [
#if defined (FFSOX_AUDIO_STEREO) // [
static int priv_stereo(void *data)
{
  return 1;
}
#endif // ]
#if defined (FFSOX_AUDIO_DRC) // [
static int priv_drc_enabled(void *data)
{
  return 0;
}

static double priv_drc_scale(void *data)
{
  return 0.0;
}
#endif // ]
#endif // ]

static ffsox_ar_cbs_t cbs={
  .frame=priv_frame,
  .stream=priv_stream,
  .read=priv_read,
  .got_frame=priv_got_frame,
#if defined (FFSOX_AUDIO_CUST) // [
#if defined (FFSOX_AUDIO_STEREO) // [
  .stereo=priv_stereo,
#endif // ]
#if defined (FFSOX_AUDIO_DRC) // [
  .drc.enabled=priv_drc_enabled,
  .drc.scale=priv_drc_scale,
#endif // ]
#endif // ]
};
