/*
 * ffsox_lame.c
 * Copyright (C) 2011, 2012 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include "ffsox.h"

ffsox_buf_t *ffsox_buf_cleanup(ffsox_buf_t *buf)
{
  if (NULL!=buf->rep)
    free(buf->rep);

  return buf;
}

ffsox_buf_t *ffsox_buf_init(ffsox_buf_t *buf, size_t size)
{
  if (0<size) {
    if (NULL==(buf->rep=malloc(size)))
      return NULL;

    buf->wp=buf->rep;
    buf->mp=buf->rep+size;
  }
  else
    buf->mp=buf->wp=buf->rep=NULL;

  return buf;
}

ffsox_buf_t *ffsox_buf_realloc(ffsox_buf_t *buf, size_t len)
{
  size_t offs;
  size_t size;
  uint8_t *rep;

  if (NULL==buf->rep) {
	len<<=1;

    if (NULL==(rep=malloc(len)))
      return NULL;

    buf->rep=rep;
    buf->wp=rep;
    buf->mp=rep+len;
  }
  else {
    size=buf->mp-buf->rep;
    offs=buf->wp-buf->rep;

    if (size<offs+len) {
      do {
        size<<=1;
      } while (size<offs+len);

      if (NULL==(rep=realloc(buf->rep,size)))
        return NULL;

      buf->rep=rep;
      buf->wp=rep+offs;
      buf->mp=rep+size;
    }
  }

  return buf;
}
