/*
 * ffsox_ffi.c
 * Copyright (C) 2011, 2012 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef NDEBUG /* N.B. assert used with active statements so enable always. */
#undef NDEBUG /* Must undef above assert.h or other that might include it. */
#endif

#include <assert.h>
#include "ffsox.h"
#if defined (FFSOX_MAP_MEMORY) // [
#include <libavformat/avio.h>
#include <libavutil/file.h>
//#define FFSOX_FFI_FILE_MAP
#if defined (FFSOX_FFI_FILE_MAP) // [
#include <fcntl.h>
#endif // ]
#endif // ]

AVCodec *ffi_find_decoder(enum AVCodecID id)
{
#if 0 // [
  AVCodec *p=AV_CODEC_ID_MP3==id?avcodec_find_decoder_by_name("mp3float"):NULL;

  return NULL==p?avcodec_find_decoder(id):p;
#else // ] [
  AVCodec *p;

  switch (id) {
    case AV_CODEC_ID_MP1:
      p=avcodec_find_decoder_by_name("mp1float");
      DVWRITELN("AV_CODEC_ID_MP1: %p",p);
      break;
    case AV_CODEC_ID_MP2:
      p=avcodec_find_decoder_by_name("mp2float");
      DVWRITELN("AV_CODEC_ID_MP2: %p",p);
      break;
    case AV_CODEC_ID_MP3:
      p=avcodec_find_decoder_by_name("mp3float");
      DVWRITELN("AV_CODEC_ID_MP3: %p",p);
      break;
    default:
      p=NULL;
      break;
  }

  return NULL==p?avcodec_find_decoder(id):p;
#endif // ]
}

#if defined (FFSOX_FFI_DECODE) // [
ffi_st_t *ffi_st_init(ffi_st_t *ist, AVStream *st, int mode,
    ffi_st_cbs_t *cbs, void *data)
#elif defined (FFSOX_AUDIO_DRC) // ] [
ffi_st_t *ffi_st_init(ffi_st_t *ist, AVStream *st, int mode, double *scale)
#else // ] [
ffi_st_t *ffi_st_init(ffi_st_t *ist, AVStream *st, int mode)
#endif // ]
{
#if defined (FFSOX_AUDIO_DRC) // [
  AVDictionary *opts=NULL;
  char drc_scale[128];
#endif // ]
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
  AVCodecParameters *codecpar=st->codecpar;
#else // ] [
  AVCodecContext *codecpar=st->codec;
#endif // ]

//DWRITELN("---------------------------------------------");
//DVWRITELN("mode.stereo=%d mode.drc.enabled=%d",FFI_AUDIO_STEREO&mode,FFI_AUDIO_DRC&mode);
//DWRITELN("---------------------------------------------");
#if defined (FFSOX_FFI_DECODE) // [
  ist->cbs=cbs;
  ist->data=data;
#endif // ]
  ist->st=st;
#if ! defined (FFSOX_FFI_DECODE) // [
#if ! defined (FFSOX_AUDIO_CUST) // [
  ist->mode=mode;
#endif // ]
#endif // ]
  ist->dec=NULL;
  ist->dc1=NULL;
  ist->index=st->index;

  ////
  ist->decoding_needed=0;
  ist->saw_first_ts=0;
  ist->dts=AV_NOPTS_VALUE;
  ist->next_dts=AV_NOPTS_VALUE;
  ////

#if ! defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
  if (AVMEDIA_TYPE_AUDIO==codecpar->codec_type) {
    if (FFI_AUDIO_STEREO&mode) {
      codecpar->request_channel_layout=AV_CH_LAYOUT_STEREO;
    }
    else
      codecpar->request_channel_layout=AV_CH_LAYOUT_NATIVE;

    codecpar->request_sample_fmt=AV_SAMPLE_FMT_FLT;
  }
#endif // ]

  if (FFI_OPEN&mode) {
    // find decoder for the stream.
    if (NULL==(ist->dec=ffi_find_decoder(codecpar->codec_id)))
      goto error;

#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
    // Allocate a codec context for the decoder.
    if (!(ist->dc1=avcodec_alloc_context3(ist->dec)))
      goto error;

    // Copy codec parameters from input stream to output codec context.
    if (avcodec_parameters_to_context(ist->dc1,codecpar)<0)
      goto error;
#else // ] [
    ist->dc1=codecpar;
#endif // ]

#if defined (FFSOX_FFI_DECODE) // [
#if defined (FFSOX_AUDIO_DRC) // [
    if (ffi_st_drc_enabled(ist)) {
      sprintf(drc_scale,"%0.1f",ffi_st_drc_scale(ist));
//DVWRITELN("mode.drc.cale=\"%s\"",drc_scale);

      if (0!=av_dict_set(&opts,"drc_scale",drc_scale,0))
        goto error;
    }
#endif // ]
#elif defined (FFSOX_AUDIO_DRC) // ] [
    if ((mode&FFI_AUDIO_DRC)&&scale) {
      sprintf(drc_scale,"%0.1f",*scale);
//DVWRITELN("mode.drc.cale=\"%s\"",drc_scale);

      if (0!=av_dict_set(&opts,"drc_scale",drc_scale,0))
        goto error;
    }

    if (0!=avcodec_open2(ist->dc1,ist->dec,&opts))
      goto error;

    if (opts)
      av_dict_free(&opts);
#else // ] [
    if (0!=avcodec_open2(ist->dc1,ist->dec,NULL))
      goto error;
#endif // ]
  }

#if 0 // [
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
  ist->dc=st->codec;
#else // ] [
  ist->dc=dc;
#endif // ]
#endif // ]

  return ist;
error:
#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // [
  if (NULL!=opts)
    av_dict_free(&opts);
#endif // ]
  ffi_st_close(ist);

  return NULL;
}

ffi_st_t *ffi_st_close(ffi_st_t *ist)
{
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
  if (NULL!=ist->dc1) {
    avcodec_free_context(&ist->dc1);
    ist->dc1=NULL;
  }
#else // ] [
  if (NULL!=ist->dec&&NULL!=ist->dc1) {
    avcodec_close(ist->dc1);
    ist->dc1=NULL;
  }
#endif // ]

  return ist;
}

//#define FFI_DUMP_META
#if defined (FFI_DUMP_META) // {
void ffi_dump_meta(AVFormatContext *ic)
{
  AVDictionaryEntry *mtag=NULL;
fprintf(stderr,">>>\n");
  while (NULL!=(mtag=av_dict_get(ic->metadata,"",mtag,
      AV_DICT_IGNORE_SUFFIX))) {
fprintf(stderr,"--- \"%s\": \"%s\"\n",mtag->key,mtag->value);
  }
fprintf(stderr,"<<<\n");
}
#endif // }

#if defined (FFSOX_FFI_DECODE) // [
static ffi_t *ffi_open(ffi_t *ffi, int mode, ffi_st_cbs_t *cbs, void *data)
#elif defined (FFSOX_AUDIO_DRC) // [
static ffi_t *ffi_open(ffi_t *ffi, int mode, double *scale)
#else // ] [
static ffi_t *ffi_open(ffi_t *ffi, int mode)
#endif // ]
{
  AVFormatContext *ic=ffi->ic;
#if defined (FFSOX_VST) // {
  int ai=-1,vi=-1,i;
#else // } {
  int ai=-1,i;
#endif // }

  ffi->ts_offset=AV_NOPTS_VALUE==ic->start_time?0.0:-ic->start_time;
  ffi->nb_streams=ic->nb_streams;

  if (NULL==(ffi->ist=calloc(ffi->nb_streams, sizeof *ffi->ist)))
    return NULL;

  for (i=0;i<ffi->nb_streams;++i) {
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
    AVCodecParameters *codecpar=ic->streams[i]->codecpar;
#else // ] [
    AVCodecContext *codecpar=ic->streams[i]->codec;
#endif // ]

#if 0
    if (st->codec->codec_type==AVMEDIA_TYPE_VIDEO)
      vi=i;
    else if (st->codec->codec_type==AVMEDIA_TYPE_AUDIO
        &&(ai<0||2==st->codec->channels)) {

      ai=i;
    }
#endif

    switch (codecpar->codec_type) {
      case AVMEDIA_TYPE_AUDIO:
        if (ai<0||2==codecpar->channels)
          ai=i;
        break;
#if defined (FFSOX_VST) // {
      case AVMEDIA_TYPE_VIDEO:
        if (vi<0)
          vi=i;
        break;
#endif // }
      default:
        break;
    }
  }

  if (ai<0)
    return NULL;

  for (i=0;i<ffi->nb_streams;++i) {
    ffi_st_t *ist=ffi->ist+i;

#if defined (FFSOX_FFI_DECODE) // [
    if ((i==ai||i==vi)&&NULL==ffi_st_init(ist,ic->streams[i],mode,cbs,data))
      return NULL;
#elif defined (FFSOX_AUDIO_DRC) // ] [
    if ((i==ai||i==vi)&&NULL==ffi_st_init(ist,ic->streams[i],mode,scale))
      return NULL;
#else // ] [
    if ((i==ai||i==vi)&&NULL==ffi_st_init(ist,ic->streams[i],mode))
      return NULL;
#endif // ]

    if (i==ai)
      ffi->ast=ist;
#if defined (FFSOX_VST) // {
    else if (i==vi)
      ffi->vst=ist;
#endif // }
  }

  return ffi;
}

#if ! defined (FFO_CLONE_STREAM_META) // {
static void ffi_clone_dict(AVDictionary **od, AVDictionary *id)
{
  AVDictionaryEntry *mtag=NULL;

  while (NULL!=(mtag=av_dict_get(id,"",mtag,AV_DICT_IGNORE_SUFFIX)))
    av_dict_set(od,mtag->key,mtag->value,AV_DICT_DONT_OVERWRITE);
}
#endif // }

#if defined (FFSOX_MAP_MEMORY) // [
static int ffi_read_packet(void *opaque, uint8_t *buf, int buf_size)
{
  int code=AVERROR_EOF;
  ffi_t *ffi=opaque;

  if (!ffi->map.rep)
    goto exit;

  buf_size=FFMIN(buf_size,ffi->map.mp-ffi->map.rp);

  if (!buf_size)
    goto exit;

  /* copy internal buffer data to buf */
  memcpy(buf,ffi->map.rp,buf_size);
  ffi->map.rp+=buf_size;
  code=buf_size;
exit:
  return code;
}

static int64_t ffi_seek(void *opaque, int64_t offset, int whence)
{
  int64_t code=AVERROR_EOF;
  ffi_t *ffi=opaque;
  uint8_t *next;

  switch (whence) {
  case SEEK_SET:
    DVWRITELN("SEEK_SET %I64d",offset);
    next=ffi->map.rep?ffi->map.rep+offset:NULL;
    break;
  case SEEK_CUR:
    DVWRITELN("SEEK_CUR %I64d",offset);
    next=ffi->map.rep?ffi->map.rp+offset:NULL;
    break;
  case SEEK_END:
    DVWRITELN("SEEK_END %I64d",offset);
    next=ffi->map.rep?ffi->map.mp+offset:NULL;
    break;
  case AVSEEK_SIZE:
    if (ffi->map.rep)
      code=ffi->map.size;

    DVWRITELN("AVSEEK_SIZE %I64d",code);
    return code;
  default:
    DWRITELN("default");
    return -1;
  }

  if (next&&ffi->map.rep<=next&&next<ffi->map.mp) {
    ffi->map.rp=next;
    code=0;
  }

  return code;
}

static int ffi_file_map(ffi_t *ffi, const pb_str_t *fn)
{
#if defined (FFSOX_FFI_FILE_MAP) // [
  struct _stat buf;
  int fd;

  if (_wstat(PS_TCS(fn),&buf)<0) {
    DMESSAGE("stat");
    goto stat;
  }

  if ((fd=_wopen(PS_TCS(fn),_O_BINARY))<0) {
    DMESSAGE("open");
    goto open;
  }

  if (!(ffi->map.rep=malloc(buf.st_size))) {
    DMESSAGE("malloc");
    goto malloc;
  }

  if (_read(fd,ffi->map.rep,buf.st_size)<buf.st_size) {
    DMESSAGE("read");
    goto read;
  }

  close(fd);
  ffi->map.size=buf.st_size;

  return 0;
read:
  free(ffi->map.rep);
  ffi->map.rep=NULL;
malloc:
  close(fd);
open:
stat:
  return -1;
#else // ] [
  return av_file_map(PS_UTF8(fn),&ffi->map.rep,&ffi->map.size,0,NULL);
#endif // ]
}

static void ffi_file_unmap(ffi_t *ffi)
{
#if defined (FFSOX_FFI_FILE_MAP) // [
  free(ffi->map.rep);
  ffi->map.rep=NULL;
#else // ] [
  av_file_unmap(ffi->map.rep,ffi->map.size);
#endif // ]
}

#if defined (FFSOX_FFI_DECODE) // [
ffi_t *ffi_init(ffi_t *ffi, pb_str_t *fn, size_t avio_ctx_buffer_size,
    int mode, ffi_st_cbs_t *cbs, void *data)
#elif defined (FFSOX_AUDIO_DRC) // ] [
ffi_t *ffi_init(ffi_t *ffi, pb_str_t *fn, size_t avio_ctx_buffer_size,
    int mode, double *scale)
#else // ] [
ffi_t *ffi_init(ffi_t *ffi, pb_str_t *fn, size_t avio_ctx_buffer_size,
    int mode)
#endif // ]
#else // ] [
#if defined (FFSOX_FFI_DECODE) // [
ffi_t *ffi_init(ffi_t *ffi, pb_str_t *fn, int mode, ffi_st_cbs_t *cbs,
    void *data)
#elif defined (FFSOX_AUDIO_DRC) // ] [
ffi_t *ffi_init(ffi_t *ffi, pb_str_t *fn, int mode, double *scale)
#else // ] [
ffi_t *ffi_init(ffi_t *ffi, pb_str_t *fn, int mode)
#endif // ]
#endif // ]
{
  AVFormatContext *ic=NULL;
#if defined (FFSOX_MAP_MEMORY) // [
  uint8_t *avio_ctx_buffer=NULL;
  //size_t avio_ctx_buffer_size=4096;
#endif // ]
#if ! defined (FFO_CLONE_STREAM_META) // {
  unsigned int i;
#endif // }

  FFSOX_ZERO(ffi);

#if defined (FFSOX_MAP_MEMORY) // [
  if (avio_ctx_buffer_size) {
    // Adapted "<ffmpeg>/doc/examples/avio_reading.c".
    /* slurp file content into buffer */
    if (ffi_file_map(ffi,fn)<0) {
      DVMESSAGE("av_file_map(\"%s\")",PS_UTF8(fn));
      goto try_open_file;
    }

    ffi->map.mp=ffi->map.rep+ffi->map.size;
    ffi->map.rp=ffi->map.rep;

    ffi->avio_ctx_buffer_size=avio_ctx_buffer_size;

    if (!(ic=avformat_alloc_context())) {
      DMESSAGE("avformat_alloc_context()");
      goto error;
    }

    ////////////////
    avio_ctx_buffer=av_malloc(avio_ctx_buffer_size);

    if (!avio_ctx_buffer) {
      DVMESSAGE("av_malloc(%d)",avio_ctx_buffer_size);
      goto error;
    }

    ////////////////
    ffi->avio_ctx=avio_alloc_context(avio_ctx_buffer,avio_ctx_buffer_size,
        0,ffi,ffi_read_packet,NULL,ffi_seek);

    if (!ffi->avio_ctx) {
      DMESSAGE("avio_alloc_context()");
      goto error;
    }

    ic->pb=ffi->avio_ctx;

    ////////////////
    if (0!=avformat_open_input(&ic,NULL,NULL,NULL))
      goto error;
  }
  else {
  try_open_file:
#endif // ]
    if (0!=avformat_open_input(&ic,PS_UTF8(fn),NULL,NULL))
      goto error;
#if defined (FFSOX_MAP_MEMORY) // [
  }
#endif // ]

  ic->flags|=AVFMT_FLAG_GENPTS;
  ffi->ic=ic;

  if (avformat_find_stream_info(ic,NULL)<0)
    goto error;

#if ! defined (FFO_CLONE_STREAM_META) // {
  if (FFI_STREAM_META&mode) {
    for (i=0;i<ic->nb_streams;++i)
      ffi_clone_dict(&ic->metadata,ic->streams[i]->metadata);
  }
#endif // }

#if defined (FFSOX_DEPRECATED_AV_FORMAT_CONTEXT_FILENAME) // [
  if (FFI_CSV&mode)
    pb_csv2avdict(ffi->ic->url,'\t',av_dict_set,&ffi->ic->metadata);
#else // ] [
  if (FFI_CSV&mode)
    pb_csv2avdict(ffi->ic->filename,'\t',av_dict_set,&ffi->ic->metadata);
#endif // ]

//DWRITELN("(FFI_OPEN&mode)=%d",FFI_OPEN&mode);
#if defined (FFSOX_FFI_DECODE) // [
  if ((FFI_OPEN&mode)&&NULL==ffi_open(ffi,mode,cbs,data))
    goto error;
#elif defined (FFSOX_AUDIO_DRC) // ] [
  if ((FFI_OPEN&mode)&&NULL==ffi_open(ffi,mode,scale))
    goto error;
#else // ] [
  if ((FFI_OPEN&mode)&&NULL==ffi_open(ffi,mode))
    goto error;
#endif // ]

  return ffi;
error:
  ffi_cleanup(ffi);

  return NULL;
}

ffi_t *ffi_cleanup(ffi_t *ffi)
{
  if (NULL!=ffi->ist) {
    int i;

    for (i=ffi->nb_streams-1;0<=i;--i)
      ffi_st_close(ffi->ist+i);

    free(ffi->ist);
    ffi->ist=NULL;
    ffi->ast=NULL;
#if defined (FFSOX_VST) // {
    ffi->vst=NULL;
#endif // }
  }

  if (NULL!=ffi->ic) {
    avformat_close_input(&ffi->ic);
    ffi->ic=NULL;
  }

#if defined (FFSOX_MAP_MEMORY) // [
  /* note: the internal buffer could have changed,
     and be != avio_ctx_buffer */
  if (ffi->avio_ctx) {
    av_freep(&ffi->avio_ctx->buffer);
    av_freep(&ffi->avio_ctx);
  }

  if (ffi->map.rep) {
    ffi_file_unmap(ffi);
    ffi->map.rep=NULL;
  }
#endif // ]

  return ffi;
}
