/*
 * sox_reader.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "ffsox.h"

#if 0 // { libav
int av_frame_get_channels(const AVFrame *frame)
{
  return av_get_channel_layout_nb_channels(frame->channel_layout);
}
#endif // }

#if defined (DEBUG) // [
///////////////////////////////////////////////////////////////////////////////
#if 0 // [
#define AR_MNEMONIC(ar) ffsox_ar_state_mnemonic(ar)

#define FFSOX_AR_STATE_MNEMONIC_OR
static const char *ffsox_ar_state_mnemonic(const ffsox_ar_t *ar)
{
  static struct { int state; const char *mnemonic; } map[]={
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
    { FFSOX_AR_READ_PACKET, "FFSOX_AR_READ_PACKET" },
    { FFSOX_AR_RECEIVE_FRAME, "FFSOX_AR_RECEIVE_FRAME" },
#if defined (FFSOX_AUDIO_STEREO) // [
    { FFSOX_AR_CONVERT_FRAME, "FFSOX_AR_CONVERT_FRAME" },
#endif // ]
    { FFSOX_AR_FLUSH, "FFSOX_AR_FLUSH" },
#else // ] [
    { FFSOX_AR_RUN, "FFSOX_AR_RUN" },
    { FFSOX_AR_FLUSH, "FFSOX_AR_FLUSH" },
#endif // ]
    { FFSOX_AR_DONE, "FFSOX_AR_DONE" },
    { FFSOX_AR_ERROR, "FFSOX_AR_ERROR" },
#if ! defined (FFSOX_AR_STATE_MNEMONIC_OR) // [
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
    { FFSOX_AR_MASK, "FFSOX_AR_MASK" },
    { FFSOX_AR_VALID, "FFSOX_AR_VALID" },
#else // ] [
    { FFSOX_AR_VALID, "FFSOX_AR_VALID" },
#endif // ]
#endif // ]
  };

  static int n=(sizeof map)/(sizeof *map);
#if 1 // [
  static char mnemonic[256];
  char *mp=mnemonic;
#endif // ]
  int i;

  for (i=0;i<n;++i) {
#if defined (FFSOX_AR_STATE_MNEMONIC_OR) // [
    if (map[i].state&ar->state) {
      if (mnemonic<mp)
        *mp++='|';

      sprintf(mp,map[i].mnemonic);
      mp+=strlen(mp);
      *mp='\0';
    }
#else // ] [
    if (map[i].state==ar->state)
      return map[i].mnemonic;
#endif // ]
  }

#if defined (FFSOX_AR_STATE_MNEMONIC_OR) // [
  return mnemonic<mp?mnemonic:"ERROR: out of range";
#else // ] [
  return "ERROR: out of range";
#endif // ]
}
#else // ] [
#define AR_MNEMONIC(ar) ""
#endif // ]
#endif // ]

///////////////////////////////////////////////////////////////////////////////
typedef uint8_t u8_t;
typedef int16_t s16_t;
typedef int32_t s32_t;
typedef float f32_t;
typedef double f64_t;

typedef struct priv {
#if defined (FFSOX_LOCAL_AR) // {
  ffsox_ar_t ar;
#else // } {
  ffsox_ar_t *ar;
#endif // }

//#define PRIV_UNION
#if defined (PRIV_UNION) // [
  union {
    // interleaved
    struct {
      uint8_t *rp;
      uint8_t *mrp;
    } i;

    // planar
    struct {
      uint8_t **rp;
      size_t nch;
      size_t bytes;
      size_t n;
      size_t i;
    } p;
  };
#else // ] [
  AVFrame *frame;
  int channel,sample;
#endif // ]
} priv_t;

///////////////////////////////////////////////////////////////////////////////
static void priv_reset_i(priv_t *priv)
{
#if defined (FFSOX_LOCAL_AR) // {
  AVFrame *frame=priv->ar.frame;
#else // } {
  AVFrame *frame=ffsox_ar_frame_internal(priv->ar);
#endif // }
#if defined (PRIV_UNION) // [
  size_t n=frame->nb_samples;
#if defined (FFSOX_DEPRECATED) // [
  size_t nch=frame->channels;
#else // ] [
  size_t nch=av_frame_get_channels(frame);
#endif // ]
  size_t bytes=av_get_bytes_per_sample(frame->format);
  uint8_t *rp=frame->data[0];

  priv->i.rp=rp;
  priv->i.mrp=rp+n*nch*bytes;
#else // ] [
  priv->frame=frame;
  priv->channel=0;
  priv->sample=0;
#endif // ]
}

static void priv_reset_p(priv_t *priv)
{
#if defined (FFSOX_LOCAL_AR) // {
  AVFrame *frame=priv->ar.frame;
#else // } {
  AVFrame *frame=ffsox_ar_frame_internal(priv->ar);
#endif // }

#if defined (PRIV_UNION) // [
  priv->p.rp=frame->data;
#if defined (FFSOX_DEPRECATED) // [
  priv->p.nch=frame->channels;
#else // ] [
  priv->p.nch=av_frame_get_channels(frame);
#endif // ]
  priv->p.bytes=av_get_bytes_per_sample(frame->format);
  priv->p.n=frame->nb_samples;
  priv->p.i=0;
#else // ] [
  priv->frame=frame;
  priv->channel=0;
  priv->sample=0;
#endif // ]
}

void priv_reset(priv_t *priv)
{
#if defined (FFSOX_LOCAL_AR) // {
  AVFrame *frame=priv->ar.frame;
#else // } {
  AVFrame *frame=ffsox_ar_frame_internal(priv->ar);
#endif // }

  switch (frame->format) {
  // interleaved:
  case AV_SAMPLE_FMT_U8:
  case AV_SAMPLE_FMT_S16:
  case AV_SAMPLE_FMT_S32:
  case AV_SAMPLE_FMT_FLT:
  case AV_SAMPLE_FMT_DBL:
    priv_reset_i(priv);
    break;
  // planar:
  case AV_SAMPLE_FMT_U8P:
  case AV_SAMPLE_FMT_S16P:
  case AV_SAMPLE_FMT_S32P:
  case AV_SAMPLE_FMT_FLTP:
  case AV_SAMPLE_FMT_DBLP:
    priv_reset_p(priv);
    break;
  // unsupported:
  default:
    break;
  }
}

///////////////////////////////////////////////////////////////////////////////
static int priv_getopts(sox_effect_t *effp, int argc, char *argv[])
{
  priv_t *priv=effp->priv;
#if defined (FFSOX_LOCAL_AR) // {
  ffi_t *ffi;
#endif // }
  int n=1;

  memset(priv,0,sizeof *priv);

  if (argc<2)
    return SOX_EOF;

#if defined (FFSOX_LOCAL_AR) // {
#error local audio reader not supported any longer.
  ffi=(ffi_t *)argv[n++];

  if (NULL==ffsox_ar_init_ffi(&priv->ar,ffi))
    return SOX_EOF;
#else // } {
  priv->ar=(ffsox_ar_t *)argv[n++];
#endif // }

#if ! defined (PRIV_UNION) // [
  priv->frame=NULL;
  priv->channel=0;
  priv->sample=0;
#endif // ]

  return SOX_SUCCESS;
}

#if defined (FFSOX_LOCAL_AR) // {
static int priv_kill(sox_effect_t *effp)
{
  priv_t *priv=effp->priv;

  ffsox_ar_cleanup(&priv->ar);

  return SOX_SUCCESS;
}
#endif // }

///////////////////////////////////////////////////////////////////////////////
#if defined (FFSOX_LOCAL_AR) // [
#define priv_ar(priv) \
  (&(priv)->ar)
#else // ] [
#define priv_ar(priv) \
  ((priv)->ar)
#endif // ]
#define priv_scale_s16(q) \
 (-(q)/INT16_MIN)
#define priv_scale_s32(q) \
 (-(q)/INT32_MIN)
#define priv_scale_f32(q) \
 (q)
#define priv_scale_f64(q) \
 (q)

///////////////////////////////////////////////////////////////////////////////
#if defined (LIBFFSOX_BS1770) // [
#if defined (PRIV_UNION) // [
#define PRIV_DRAIN_I(id1,id2,ID) \
static int priv_drain_##id1##i(sox_effect_t *effp, int nch, \
    sox_sample_t *obuf, size_t *osamp) \
{ \
  priv_t *priv=effp->priv; \
  ffsox_ar_t *ar=priv_ar(priv); \
  double q1=ar->q; \
  double q2=priv_scale_##id1(q1); \
  uint8_t *rp,*mrp; \
  sox_sample_t *wp=obuf; \
  sox_sample_t *mwp=obuf+(*osamp-=*osamp%nch); \
  int ch; \
  id1##_t x; \
  double y; \
  bs1770_sample_##id2##_t sample; \
  SOX_SAMPLE_LOCALS; \
 \
  if (0==ar->count) { \
    priv->i.rp=NULL; \
    priv->i.mrp=NULL; \
  } \
 \
  rp=priv->i.rp; \
  mrp=priv->i.mrp; \
 \
  while (wp<mwp) { \
    for (ch=0;ch<nch;++ch) { \
      while (rp==mrp) { \
        if (!rp) \
          DWRITELN("buisy waiting ..."); \
 \
        ffsox_ar_next(ar); \
 \
        if (!ffsox_ar_valid(ar)) \
          goto done; \
 \
        priv_reset_i(priv); \
        rp=priv->i.rp; \
        mrp=priv->i.mrp; \
      } \
 \
      memcpy(&x,rp,sizeof x); \
      rp+=sizeof x; \
      sample[ch]=x; \
 \
      if (1.0==q1) \
        *wp++=SOX_##ID##_TO_SAMPLE(x,effp->clips); \
      else { \
        y=q2*x; \
        *wp++=SOX_FLOAT_64BIT_TO_SAMPLE(y,effp->clips); \
      } \
    } \
 \
    if (NULL!=ar->lufs) \
      ar->lufs->vmt->add_sample_##id2(ar->lufs,sample); \
  } \
done: \
 \
  priv->i.rp=rp; \
 \
  return 0<(*osamp=wp-obuf)?SOX_SUCCESS:SOX_EOF; \
}
#else // ] [
#error not implemented yet.
#endif // ]

PRIV_DRAIN_I(s16,i16,SIGNED_16BIT)
PRIV_DRAIN_I(s32,i32,SIGNED_32BIT)
PRIV_DRAIN_I(f32,f32,FLOAT_32BIT)
PRIV_DRAIN_I(f64,f64,FLOAT_64BIT)
#else // ] [
#if defined (PRIV_UNION) // [
#define PRIV_DRAIN_I(id,ID) \
static int priv_drain_##id##i(sox_effect_t *effp, int nch, \
    sox_sample_t *obuf, size_t *osamp) \
{ \
  priv_t *priv=effp->priv; \
  ffsox_ar_t *ar=priv_ar(priv); \
  double q1=ar->q; \
  double q2=priv_scale_##id(q1); \
  uint8_t *rp,*mrp; \
  sox_sample_t *wp=obuf; \
  sox_sample_t *mwp=obuf+(*osamp-=*osamp%nch); \
  id##_t x; \
  double y; \
  SOX_SAMPLE_LOCALS; \
 \
  if (0==ar->count) { \
    priv->i.rp=NULL; \
    priv->i.mrp=NULL; \
  } \
 \
  rp=priv->i.rp; \
  mrp=priv->i.mrp; \
 \
  while (wp<mwp) { \
    while (rp==mrp) { \
      if (!rp) \
        DWRITELN("buisy waiting ..."); \
 \
      ffsox_ar_next(ar); \
 \
      if (!ffsox_ar_valid(ar)) \
        goto done; \
 \
      priv_reset_i(priv); \
      rp=priv->i.rp; \
      mrp=priv->i.mrp; \
    } \
 \
    memcpy(&x,rp,sizeof x); \
    rp+=sizeof x; \
 \
    if (1.0==q1) \
      *wp++=SOX_##ID##_TO_SAMPLE(x,effp->clips); \
    else { \
      y=q2*x; \
      *wp++=SOX_FLOAT_64BIT_TO_SAMPLE(y,effp->clips); \
    } \
  } \
done: \
  \
  priv->i.rp=rp; \
 \
  return 0<(*osamp=wp-obuf)?SOX_SUCCESS:SOX_EOF; \
}
#else // ] [
#define PRIV_DRAIN_I(id,ID) \
static int priv_drain_##id##i(sox_effect_t *effp, int nch, \
    sox_sample_t *obuf, size_t *osamp) \
{ \
  priv_t *priv=effp->priv; \
  AVFrame *frame=priv->frame; \
  ffsox_ar_t *ar=priv_ar(priv); \
  double q1=ar->q; \
  double q2=priv_scale_##id(q1); \
  int nb_samples=frame->channels*frame->nb_samples; \
  int sample; \
  sox_sample_t *wp=obuf; \
  sox_sample_t *mwp=obuf+(*osamp-=*osamp%frame->channels); \
  id##_t x; \
  double y; \
  SOX_SAMPLE_LOCALS; \
 \
  if (0==ar->count) { \
    priv->sample=0; \
  } \
 \
  sample=priv->sample; \
 \
  while (wp<mwp) { \
    while (sample==nb_samples) { \
      ffsox_ar_next(ar); \
 \
      if (!ffsox_ar_valid(ar)) \
        goto done; \
 \
      priv_reset_i(priv); \
      sample=0; \
    } \
 \
    x=((id##_t *)frame->data[0])[sample]; \
    ++sample; \
 \
    if (1.0==q1) \
      *wp++=SOX_##ID##_TO_SAMPLE(x,effp->clips); \
    else { \
      y=q2*x; \
      *wp++=SOX_FLOAT_64BIT_TO_SAMPLE(y,effp->clips); \
    } \
  } \
done: \
  \
  priv->sample=sample; \
 \
  return 0<(*osamp=wp-obuf)?SOX_SUCCESS:SOX_EOF; \
}
#endif // ]

PRIV_DRAIN_I(s16,SIGNED_16BIT)
PRIV_DRAIN_I(s32,SIGNED_32BIT)
PRIV_DRAIN_I(f32,FLOAT_32BIT)
PRIV_DRAIN_I(f64,FLOAT_64BIT)
#endif // }

///////////////////////////////////////////////////////////////////////////////
#if defined (LIBFFSOX_BS1770) // [
#if defined (PRIV_UNION) // [
#define PRIV_DRAIN_P(id1,id2,ID) \
static int priv_drain_##id1##p(sox_effect_t *effp, int nch, \
    sox_sample_t *obuf, size_t *osamp) \
{ \
  priv_t *priv=effp->priv; \
  ffsox_ar_t *ar=priv_ar(priv); \
  double q1=ar->q; \
  double q2=priv_scale_##id1(q1); \
  uint8_t **rp=priv->p.rp; \
  int n,i; \
  sox_sample_t *wp=obuf; \
  sox_sample_t *mwp=obuf+(*osamp-=*osamp%nch); \
  int ch; \
  id1##_t x; \
  double y; \
  bs1770_sample_##id2##_t sample; \
  SOX_SAMPLE_LOCALS; \
 \
  if (0==ar->count) { \
    priv->p.n=0; \
    priv->p.i=0; \
  } \
 \
  n=priv->p.n; \
  i=priv->p.i; \
 \
  while (wp<mwp) { \
    while (i==n) { \
      ffsox_ar_next(ar); \
 \
      if (!ffsox_ar_valid(ar)) \
        goto done; \
 \
      priv_reset_p(priv); \
      rp=priv->p.rp; \
      nch=priv->p.nch; \
      n=priv->p.n; \
      i=priv->p.i; \
    } \
 \
    for (ch=0;ch<nch;++ch) { \
      memcpy(&x,rp[ch]+i*sizeof x,sizeof x); \
      sample[ch]=x; \
 \
      if (1.0==q1) \
        *wp++=SOX_##ID##_TO_SAMPLE(x,effp->clips); \
      else { \
        y=q2*x; \
        *wp++=SOX_FLOAT_64BIT_TO_SAMPLE(y,effp->clips); \
      } \
    } \
 \
    if (NULL!=ar->lufs) \
      ar->lufs->vmt->add_sample_##id2(ar->lufs,sample); \
 \
    ++i; \
  } \
done: \
 \
  priv->p.i=i; \
 \
  return 0<(*osamp=wp-obuf)?SOX_SUCCESS:SOX_EOF; \
}
#else // ] [
#error not implemented yet.
#endif // ]

PRIV_DRAIN_P(s16,i16,SIGNED_16BIT)
PRIV_DRAIN_P(s32,i32,SIGNED_32BIT)
PRIV_DRAIN_P(f32,f32,FLOAT_32BIT)
PRIV_DRAIN_P(f64,f64,FLOAT_64BIT)
#else // ] [
#if defined (PRIV_UNION) // [
#define PRIV_DRAIN_P(id,ID) \
static int priv_drain_##id##p(sox_effect_t *effp, int nch, \
    sox_sample_t *obuf, size_t *osamp) \
{ \
  priv_t *priv=effp->priv; \
  ffsox_ar_t *ar=priv_ar(priv); \
  double q1=ar->q; \
  double q2=priv_scale_##id(q1); \
  uint8_t **rp=priv->p.rp; \
  int n,i; \
  sox_sample_t *wp=obuf; \
  sox_sample_t *mwp=obuf+(*osamp-=*osamp%nch); \
  int ch; \
  id##_t x; \
  double y; \
  SOX_SAMPLE_LOCALS; \
 \
  if (0==ar->count) { \
    priv->p.n=0; \
    priv->p.i=0; \
  } \
 \
  n=priv->p.n; \
  i=priv->p.i; \
 \
  while (wp<mwp) { \
    while (i==n) { \
      if (!n) \
        DWRITELN("buisy waiting ..."); \
 \
      ffsox_ar_next(ar); \
 \
      if (!ffsox_ar_valid(ar)) \
        goto done; \
 \
      priv_reset_p(priv); \
      rp=priv->p.rp; \
      nch=priv->p.nch; \
      n=priv->p.n; \
      i=priv->p.i; \
    } \
 \
    for (ch=0;ch<nch;++ch) { \
      memcpy(&x,rp[ch]+i*sizeof x,sizeof x); \
 \
      if (1.0==q1) \
        *wp++=SOX_##ID##_TO_SAMPLE(x,effp->clips); \
      else { \
        y=q2*x; \
        *wp++=SOX_FLOAT_64BIT_TO_SAMPLE(y,effp->clips); \
      } \
    } \
 \
    ++i; \
  } \
done: \
 \
  priv->p.i=i; \
 \
  return 0<(*osamp=wp-obuf)?SOX_SUCCESS:SOX_EOF; \
}
#else // ] [
#define PRIV_DRAIN_P(id,ID) \
static int priv_drain_##id##p(sox_effect_t *effp, int nch, \
    sox_sample_t *obuf, size_t *osamp) \
{ \
  priv_t *priv=effp->priv; \
  AVFrame *frame=priv->frame; \
  int nb_samples=frame->nb_samples; \
  ffsox_ar_t *ar=priv_ar(priv); \
  double q1=ar->q; \
  double q2=priv_scale_##id(q1); \
  int channel,sample; \
  sox_sample_t *wp=obuf; \
  sox_sample_t *mwp=obuf+(*osamp-=*osamp%frame->channels); \
  id##_t x; \
  double y; \
  SOX_SAMPLE_LOCALS; \
 \
  if (0==ar->count) { \
    priv->channel=0; \
    priv->sample=0; \
  } \
 \
  channel=priv->channel; \
  sample=priv->sample; \
 \
  while (wp<mwp) { \
    while (sample==nb_samples) { \
      ffsox_ar_next(ar); \
 \
      if (!ffsox_ar_valid(ar)) \
        goto done; \
 \
      priv_reset_p(priv); \
      priv->channel=channel=0; \
      priv->sample=sample=0; \
    } \
 \
    for (channel=0;channel<frame->channels;++channel) { \
      x=((id##_t *)frame->data[channel])[sample]; \
 \
      if (1.0==q1) \
        *wp++=SOX_##ID##_TO_SAMPLE(x,effp->clips); \
      else { \
        y=q2*x; \
        *wp++=SOX_FLOAT_64BIT_TO_SAMPLE(y,effp->clips); \
      } \
    } \
 \
    ++sample; \
  } \
done: \
 \
  priv->channel=channel; \
  priv->sample=sample; \
 \
  return 0<(*osamp=wp-obuf)?SOX_SUCCESS:SOX_EOF; \
}
#endif // ]

PRIV_DRAIN_P(s16,SIGNED_16BIT)
PRIV_DRAIN_P(s32,SIGNED_32BIT)
PRIV_DRAIN_P(f32,FLOAT_32BIT)
PRIV_DRAIN_P(f64,FLOAT_64BIT)
#endif // ]

///////////////////////////////////////////////////////////////////////////////
/*
 * drain() is called only once per sox effect. For the sox reader effect it
 * reads all the samples and flows them into the sox chain.
 */
static int priv_drain(sox_effect_t *effp, sox_sample_t *obuf, size_t *osamp)
{
  int code=SOX_EOF;
  priv_t *priv=effp->priv;
#if defined (PRIV_UNION) // [
#if defined (FFSOX_DEPRECATED_STREAM_CODEC) // [
  AVCodecParameters *codecpar=ffsox_ar_stream(priv_ar(priv))->codecpar;
#else // ] [
  AVCodecContext *codecpar=ffsox_ar_stream(priv_ar(priv))->codec;
#endif // ]
  int sample_fmt=codecpar->format;
  int channels=codecpar->channels;
#else // ] [
  ffsox_ar_t *ar=priv_ar(priv);
  AVFrame *frame=ffsox_ar_frame_internal(ar);
  int sample_fmt,channels;

  if (!frame||AV_SAMPLE_FMT_NONE==frame->format) {
		/*
		 * We should always go here and set up the effect, i.e. read a
		 * first frame and based on that decide how to proceed. Subsequent
		 * frames will be read in the format found by that first frame
		 * specific priv_drain_<format>() methods.
		 */
    ffsox_ar_next(ar);

    frame=ffsox_ar_frame_internal(ar);
    priv->channel=0;
    priv->sample=0;
  }

  priv->frame=frame;
  sample_fmt=frame->format;
  channels=frame->channels;
#endif // ]

  switch (sample_fmt) {
  case AV_SAMPLE_FMT_S16:
    DWRITELN("priv_drain_s16i {");
    code=priv_drain_s16i(effp,channels,obuf,osamp);
    DVWRITELN("priv_drain_s16i } code: %d",code);
    break;
  case AV_SAMPLE_FMT_S32:
    DMARKLN();
    code=priv_drain_s32i(effp,channels,obuf,osamp);
    DVWRITELN("code: %d",code);
    break;
  case AV_SAMPLE_FMT_FLT:
    DMARKLN();
    code=priv_drain_f32i(effp,channels,obuf,osamp);
    DVWRITELN("code: %d",code);
    break;
  case AV_SAMPLE_FMT_DBL:
    DMARKLN();
    code=priv_drain_f64i(effp,channels,obuf,osamp);
    DVWRITELN("code: %d",code);
    break;
  // planar.
  case AV_SAMPLE_FMT_S16P:
    DMARKLN();
    code=priv_drain_s16p(effp,channels,obuf,osamp);
    DVWRITELN("code: %d",code);
    break;
  case AV_SAMPLE_FMT_S32P:
    DMARKLN();
    code=priv_drain_s32p(effp,channels,obuf,osamp);
    DVWRITELN("code: %d",code);
    break;
  case AV_SAMPLE_FMT_FLTP:
    DMARKLN();
    code=priv_drain_f32p(effp,channels,obuf,osamp);
    DVWRITELN("code: %d",code);
    break;
  case AV_SAMPLE_FMT_DBLP:
    DMARKLN();
    code=priv_drain_f64p(effp,channels,obuf,osamp);
    DVWRITELN("code: %d",code);
    break;
  default:
    DMARKLN();
    break;
  }

  return code;
}

#if 0 // [
#define PRIV_FLOW
/*
 * We don't need a flow() method because we're an input effect to SoX and as
 * such we're just draining.
 */
static int priv_flow(sox_effect_t *effp, const sox_sample_t *ibuf,
    sox_sample_t *obuf, size_t *isamp, size_t *osamp)
{
  return SOX_SUCCESS;
}
#endif // ]

///////////////////////////////////////////////////////////////////////////////
sox_effect_handler_t const *ffsox_sox_reader()
{
  static sox_effect_handler_t handler={
    // Effect name
    .name="ffsox_sox_reader",
    // Short explanation of parameters accepted by effect
    .usage=NULL,
    // Combination of SOX_EFF_* flags
    .flags=SOX_EFF_MCHAN,
    // Called to parse command-line arguments (called once per effect).
    .getopts=priv_getopts,
    // Called to initialize effect (called once per flow).
    .start=NULL,
    // Called to process samples.
#if defined (PRIV_FLOW) // [
    .flow=priv_flow,
#else // ] [
    .flow=NULL,
#endif // ]
    // Called to finish getting output after input is complete.
    .drain=priv_drain,
    // Called to shut down effect (called once per flow).
    .stop=NULL,
    // Called to shut down effect (called once per effect).
#if defined (FFSOX_LOCAL_AR) // {
    .kill=priv_kill,
#else // } {
    .kill=NULL,
#endif // }
    // Size of private data SoX should pre-allocate for effect
    .priv_size=sizeof (priv_t)
  };

  return &handler;
}
