#include <stdlib.h>
#include <string.h>
#include "pb_xy.h"

///////////////////////////////////////////////////////////////////////////////
static void atom(char **dst, char *src, int n, int *size, char *buf)
{
  if (n<0)
    n=0;

  if (NULL!=buf) {
    memcpy(buf+*size,src,n);
    *dst=buf+*size;
    (*dst)[n]=0;
  }

  *size+=n+1;
}

#if defined (WIN32) && defined (UNICODE)
static void watom(wchar_t **dst, wchar_t *src, int n, int *size, char *buf)
{
  if (n<0)
    n=0;

  if (NULL!=buf) {
    memcpy(buf+*size,src,n*sizeof(wchar_t));
    *dst=(wchar_t *)(buf+*size);
    (*dst)[n]=0;
  }

  *size+=(n+1)*sizeof(wchar_t);
}
#endif

///////////////////////////////////////////////////////////////////////////////
static int __pb_file_cre(pb_str_t *path, pb_file_t *f)
{
  pb_str_t name,ext;
  int i,n,size;
  char *buf;

  pb_str_name(&name,path);
  pb_str_ext(&ext,&name);
  buf=(char *)f;

  if (NULL!=f) {
    f->prev=NULL;
    f->next=NULL;
    f->path=path;
    f->name=name;
    f->ext=ext;
  }

  size=sizeof *f;

  // dn
  if (NULL!=f)
    f->dn=(pb_str_t *)(buf+size);

  size+=sizeof *f->dn;
#if defined (WIN32) && defined (UNICODE)
  watom(f?&f->dn->tcs:NULL,path->tcs,name.tcs-path->tcs-1,&size,buf);
  atom(f?&f->dn->utf8:NULL,path->utf8,name.utf8-path->utf8-1,&size,buf);
  atom(f?&f->dn->oem:NULL,path->oem,name.oem-path->oem-1,&size,buf);
#else
  atom(f?&f->dn->utf8:NULL,path->utf8,name.utf8-path->utf8-1,&size,buf);

  if (NULL!=f) {
    f->dn->oem=f->dn->utf8;
#if defined (WIN3)
    f->dn->tcs=f->dn->utf8;
#endif
  }
#endif

  i=size%4;

  if (0<i)
    size+=4-i;

  // bn
  if (NULL!=f)
    f->bn=(pb_str_t *)(buf+size);

  size+=sizeof *f->bn;
#if defined (WIN32) && defined (UNICODE)
  n=NULL==ext.tcs?wcslen(name.tcs):ext.tcs-name.tcs-1;
  watom(f?&f->bn->tcs:NULL,name.tcs,n,&size,buf);
  n=NULL==ext.utf8?strlen(name.utf8):ext.utf8-name.utf8-1;
  atom(f?&f->bn->utf8:NULL,name.utf8,n,&size,buf);
  n=NULL==ext.oem?strlen(name.oem):ext.oem-name.oem-1;
  atom(f?&f->bn->oem:NULL,name.oem,n,&size,buf);
#else
  n=NULL==ext.utf8?strlen(name.utf8):ext.utf8-name.utf8-1;
  atom(f?&f->bn->utf8:NULL,name.utf8,n,&size,buf);

  if (NULL!=f) {
    f->bn->oem=f->bn->utf8;
#if defined (WIN3)
    f->bn->tcs=f->bn->utf8;
#endif
  }
#endif

  return size;
}

///////////////////////////////////////////////////////////////////////////////
pb_file_t *pb_file_cre(pb_str_t *path)
{
  pb_file_t *f;
  int size;

  size=__pb_file_cre(path,NULL);

  if (NULL==(f=malloc(size)))
    return NULL;

  __pb_file_cre(path,f);

  return f;
}

///////////////////////////////////////////////////////////////////////////////
int pb_file_test(pb_file_t *head, pb_str_t *path)
{
  struct stat s1,s2;

  if (0!=pb_str_stat(path,&s1))
    return 0;

  for (;head;head=head->next)
    if (0!=pb_str_stat(head->path,&s2))
      continue;
	else if (s1.st_ino==s2.st_ino&&s1.st_dev==s2.st_dev)
      return 1;

  return 0;
}

///////////////////////////////////////////////////////////////////////////////
pb_file_t *pb_file_push(pb_file_t **head, pb_file_t **tail, pb_file_t *f)
{
  f->next=NULL;

  if (NULL==(f->prev=*tail))
    *head=f;
  else
    (*tail)->next=f;

  return *tail=f;
}

///////////////////////////////////////////////////////////////////////////////
pb_file_t *pb_file_pop(pb_file_t **head, pb_file_t **tail)
{
  pb_file_t *f=*tail;

  if (NULL==(*tail=f->prev))
    *head=NULL;

  f->prev=NULL;
  f->next=NULL;

  return f;
}

///////////////////////////////////////////////////////////////////////////////
void pb_file_close(pb_file_t **head, pb_file_t **tail)
{
  while (NULL!=*tail)
    free(pb_file_pop(head,tail));
}
