#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include "pb_xy.h"

int pb_str_size(const TCHAR *ts)
{
  int size;

  size=sizeof(pb_str_t);
#if defined (WIN32) && defined (UNICODE)
  size+=(wcslen(ts)+1)*sizeof *ts;
  size+=pb_ws2s_size(ts,CP_UTF8);
  size+=pb_ws2s_size(ts,CP_OEMCP);
#else
  size+=(strlen(ts)+1)*sizeof *ts;
#endif

  return size;
}

int pb_str_copy(pb_str_t *ps, const TCHAR *ts, int size)
{
  int n;
  char *buf;

  n=sizeof(pb_str_t);
  buf=(char *)ps;
  size-=n;
  buf+=n;

#if defined (WIN32) && defined (UNICODE)
  ps->utf8=buf;
  n=pb_ws2s_copy(ps->utf8,ts,CP_UTF8,size);
  size-=n;
  buf+=n;

  ps->oem=buf;
  n=pb_ws2s_copy(ps->oem,ts,CP_OEMCP,size);
  size-=n;
  buf+=n;

  ps->tcs=(wchar_t *)buf;
  n=(wcslen(ts)+1)*sizeof *ts;
  memcpy(ps->tcs,ts,n);
  size-=n;
  buf+=n;
#else
  ps->utf8=buf;
  ps->oem=buf;
#if defined (WIN32)
  ps->tcs=buf;
#endif
  n=strlen(ts)+1;
  memcpy(buf,ts,n);
  size-=n;
  buf+=n;
#endif

  return buf-(char *)ps;
}

pb_str_t *pb_str_dup(pb_str_t *ts)
{
  return pb_str_dup_pfx(NULL,ts);
}

pb_str_t *pb_str_dup_pfx(pb_str_t *pfx, pb_str_t *ts)
{
#if defined (ROLLBACK)
  pb_str_t *ps;
  pb_y_t y;

  pb_y_init(&y);
  pb_y_count_part_ps(&y,pfx,ts);
  pb_y_count_whole(&y);

  if (NULL==(ps=pb_y_alloc(&y,sizeof *ps)))
	return NULL;

  pb_y_add_part_ps(&y,pfx,ts);
  pb_y_add_whole(&y,ps);
  pb_y_close(&y);

  return ps;
#else // ROLLBACK
  pb_str_t *ps;
  pb_z_t z;
  int size;

  size=pb_z_dup_pfx(&z,pfx,ts,NULL);

  if (NULL==(ps=malloc(size)))
	return NULL;

  size=pb_z_dup_pfx(&z,pfx,ts,ps);

  return ps;
#endif // ROLLBACK
}

pb_str_t *pb_str_cre(const TCHAR *ts)
{
  pb_str_t *ps;
  size_t size;

  size=pb_str_size(ts);

  if (NULL==(ps=malloc(size)))
	return NULL;

  pb_str_copy(ps,ts,size);

  return ps;
}

pb_str_t *pb_str_cre_cs(const char *s)
{
#if defined (WIN32) && defined (UNICODE)
  pb_str_t *ps;
  TCHAR *ts;
  int size;

  size=pb_s2ws_size(s,CP_UTF8);

  if (NULL==(ts=malloc(size*sizeof *ts)))
    return NULL;

  pb_s2ws_copy(ts,s,CP_UTF8,size);
  
  if (NULL==(ps=pb_str_cre(ts)))
    goto error;

  free(ts);

  return ps;
error:
  if (NULL!=ts)
    free(ts);

  return NULL;
#else
  return pb_str_cre((char *)s);
#endif
}

pb_str_t *pb_str_cat(TCHAR *ts, ...)
{
#if defined (ROLLBACK) // {
  pb_str_t *ps;
  TCHAR *cur;
  pb_y_t y;
  va_list ap;

  pb_y_init(&y);
  va_start(ap,ts);

  for (cur=ts;cur;cur=va_arg(ap,TCHAR *))
    pb_y_count_part(&y,cur);

  va_end(ap);
  pb_y_count_whole(&y);

  if (NULL==(ps=pb_y_alloc(&y,sizeof *ps)))
	return NULL;

  va_start(ap,ts);

  for (cur=ts;cur;cur=va_arg(ap,TCHAR *))
    pb_y_add_part(&y,cur);

  va_end(ap);
  pb_y_add_whole(&y,ps);
  pb_y_close(&y);

  return ps;
#else // } {
  va_list ap;
  pb_z_t z;
  pb_str_t *ps;
  int size;

  va_start(ap,ts);
  size=pb_z_cat(&z,ts,ap,NULL);
  va_end(ap);

  if (NULL==(ps=malloc(size)))
	return NULL;

  va_start(ap,ts);
  size=pb_z_cat(&z,ts,ap,ps);
  va_end(ap);

  return ps;
#endif // }
}

pb_str_t *pb_str_cat_ps(pb_str_t *ts, ...)
{
#if defined (ROLLBACK)
  pb_str_t *ps;
  pb_str_t *cur;
  pb_y_t y;
  va_list ap;

  pb_y_init(&y);
  va_start(ap,ts);

  for (cur=ts;cur;cur=va_arg(ap,pb_str_t *))
    pb_y_count_part_ps(&y,NULL,cur);

  va_end(ap);
  pb_y_count_whole(&y);

  if (NULL==(ps=pb_y_alloc(&y,sizeof *ps)))
	return NULL;

  va_start(ap,ts);

  for (cur=ts;cur;cur=va_arg(ap,pb_str_t *))
    pb_y_add_part_ps(&y,NULL,cur);

  va_end(ap);
  pb_y_add_whole(&y,ps);
  pb_y_close(&y);

  return ps;
#else // ROLLBACK
  va_list ap;
  pb_z_t z;
  pb_str_t *ps2;
  int size;

  va_start(ap,ts);
  size=pb_z_cat_ps(&z,ts,ap,NULL);
  va_end(ap);

  if (NULL==(ps2=malloc(size)))
	return NULL;

  va_start(ap,ts);
  pb_z_cat_ps(&z,ts,ap,ps2);
  va_end(ap);

  return ps2;
#endif // ROLLBACK
}

pb_str_t **pb_str_cre_n_ps(pb_str_t **ts, int n)
{
#if defined (ROLLBACK)
  pb_x_t x;
  int i;

  pb_x_init(&x);

  for (i=0;i<n;++i)
	pb_x_count_ps(&x,ts[i],NULL);

  if (NULL==pb_x_alloc(&x))
	return NULL;

  for (i=0;i<n;++i)
	pb_x_add_ps(&x,ts[i],NULL);

  pb_x_flush(&x);

  return x.ps;
#else // ROLLBACK
  pb_a_t a;
  pb_str_t **psp;
  int size;

  size=pb_a_str_cre_n_ps(&a,ts,n,NULL);

  if (NULL==(psp=malloc(size)))
	return NULL;

  size=pb_a_str_cre_n_ps(&a,ts,n,psp);

  return psp;
#endif // ROLLBACK
}

pb_str_t **pb_str_dup_v(pb_str_t **ps)
{
#if defined (ROLLBACK)
  pb_str_t **mp;

  for (mp=ps;*mp;++mp)
    ;

  return pb_str_cre_n_ps(ps,mp-ps);
#else // ROLLBACK
  pb_a_t a;
  pb_str_t **psp;
  int size;

  size=pb_a_str_cre_v_ps(&a,ps,NULL);

  if (NULL==(psp=malloc(size)))
	return NULL;

  size=pb_a_str_cre_v_ps(&a,ps,psp);

  return psp;
#endif // ROLLBACK
}

pb_str_t **pb_str_cre_n(TCHAR **ts, int n)
{
#if defined (ROLLBACK)
  pb_x_t x;
  int i;

  pb_x_init(&x);

  for (i=0;i<n;++i)
	pb_x_count(&x,ts[i],NULL);

  if (NULL==pb_x_alloc(&x))
	return NULL;

  for (i=0;i<n;++i)
	pb_x_add(&x,ts[i],NULL);

  pb_x_flush(&x);

  return x.ps;
#else // ROLLBACK
  pb_a_t a;
  pb_str_t **psp;
  int size;

  size=pb_a_str_cre_n(&a,ts,n,NULL);

  if (NULL==(psp=malloc(size)))
	return NULL;

  size=pb_a_str_cre_n(&a,ts,n,psp);

  return psp;
#endif // ROLLBACK
}

pb_str_t **pb_str_cre_v(TCHAR **ts)
{
#if defined (ROLLBACK)
  TCHAR **mp;

  for (mp=ts;*mp;++mp)
	;

  return pb_str_cre_n(ts,mp-ts);
#else // ROLLBACK
  pb_a_t a;
  pb_str_t **psp;
  int size;

  size=pb_a_str_cre_v(&a,ts,NULL);

  if (NULL==(psp=malloc(size)))
	return NULL;

  size=pb_a_str_cre_v(&a,ts,psp);

  return psp;
#endif // ROLLBACK
}

pb_str_t **pb_str_cre_v_cs(char **s)
{
#if defined (ROLLBACK)
  pb_x_t x;
  char **cur;
#if defined (WIN32) && defined (UNICODE)
  int size,max;
  wchar_t *ws;

  max=0;

  for (cur=s;*cur;++cur) {
    size=pb_s2ws_size(*cur,CP_UTF8);

	if (max<size)
	  max=size;
  }

  if (NULL==(ws=malloc(max*sizeof *ws)))
    return NULL;
#endif

  pb_x_init(&x);

  for (cur=s;*cur;++cur) {
#if defined (WIN32) && defined (UNICODE)
    pb_s2ws_copy(ws,*cur,CP_UTF8,max);
	pb_x_count(&x,ws,NULL);
#else
	pb_x_count(&x,*cur,NULL);
#endif
  }

  if (NULL==pb_x_alloc(&x))
	goto error;

  for (cur=s;*cur;++cur) {
#if defined (WIN32) && defined (UNICODE)
    pb_s2ws_copy(ws,*cur,CP_UTF8,max);
	pb_x_add(&x,ws,NULL);
#else
	pb_x_add(&x,*cur,NULL);
#endif
  }

  pb_x_flush(&x);
#if defined (WIN32) && defined (UNICODE)
  free(ws);
#endif

  return x.ps;
error:
#if defined (WIN32) && defined (UNICODE)
  if (NULL!=ws)
    free(ws);
#endif
  return NULL;
#else // ROLLBACK
#if defined (WIN32) && defined (UNICODE)
  TCHAR **ts;
  pb_str_t **ps;

  if (NULL==(ts=pb_s2ws_v(s,CP_UTF8)))
	return NULL;

  ps=pb_str_cre_v(ts);
  free(ts);

  return ps;
#else
  return pb_str_cre_v(s);
#endif
#endif // ROLLBACK
}
