#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <stdarg.h>
#include "pb_xy.h"

#if defined (WIN32)
#  define TCS(s)        (s)->tcs
#  define SEP           _T("\\")
#else
#  define _T(s)         s
#  define _TDIR         DIR
#  define _tdirent      dirent
#  define _topendir     opendir
#  define _tclosedir    closedir
#  define _trewinddir   rewinddir
#  define _treaddir     readdir
#  define _tcscmp       strcmp
#  define TCS(s)        (s)->utf8
#  define SEP           "/"
#endif

#if defined (ROLLBACK)
static void pb_x_readdir(pb_x_t *x, pb_str_t *dir, int filt, _TDIR *dirp,
    pb_x_next_t pb_x_next)
{
  struct _tdirent *e;

  while (NULL!=(e=_treaddir(dirp))) {
	if (filt&&(0==_tcscmp(_T("."),e->d_name)||0==_tcscmp(_T(".."),e->d_name)))
	  continue;
	else if (NULL==dir)
	  pb_x_next(x,e->d_name,NULL);
	else
	  pb_x_next(x,TCS(dir),SEP,e->d_name,NULL);
  }
}

pb_str_t **pb_str_readdir(pb_str_t *dir, int cat, int filt)
{
  _TDIR *dirp;
  pb_str_t *cur;
  pb_x_t x;
  struct _tdirent *e;

  if (NULL==(dirp=_topendir(TCS(dir))))
	goto error;

  pb_x_init(&x);
  pb_x_readdir(&x,cat?dir:NULL,filt,dirp,pb_x_count);

  if (NULL==pb_x_alloc(&x))
	return NULL;

  _trewinddir(dirp);
  pb_x_readdir(&x,cat?dir:NULL,filt,dirp,pb_x_add);
  pb_x_flush(&x);
  _tclosedir(dirp);

  return x.ps;
error:
  if (NULL!=dirp)
	_tclosedir(dirp);

  return NULL;
}
#else // ROLLBACK
int pb_a_readdir(pb_a_t *a, pb_str_t *dir, int cat, int filt, _TDIR *dirp,
    pb_str_t **psp)
{
  struct _tdirent *e;

  pb_a_start(a,psp);

  while (NULL!=(e=_treaddir(dirp))) {
	if (filt&&(0==_tcscmp(_T("."),e->d_name)||0==_tcscmp(_T(".."),e->d_name)))
	  continue;
	else if (NULL==dir)
      pb_a_add(a,e->d_name);
	else
      pb_a_add_v(a,TCS(dir),SEP,e->d_name,NULL);
  }

  return pb_a_finish(a);
}

pb_str_t **pb_str_readdir(pb_str_t *dir, int cat, int filt)
{
  pb_a_t a;
  int size;
  pb_str_t **psp=NULL;
  _TDIR *dirp=NULL;

  if (NULL==(dirp=_topendir(TCS(dir))))
	goto error;

  size=pb_a_readdir(&a,dir,cat,filt,dirp,NULL);

  if (NULL==(psp=malloc(size)))
	goto error;

  _trewinddir(dirp);
  size=pb_a_readdir(&a,dir,cat,filt,dirp,psp);
  _tclosedir(dirp);

  return psp;
error:
  if (NULL!=dirp)
	_tclosedir(dirp);

  return NULL;
}
#endif // ROLLBACK
