#include "pb.h"

#if defined (WIN32)
static DWORD get_ino_fake(const pb_str_t *path)
{
  BY_HANDLE_FILE_INFORMATION fi;
  DWORD dwShareMode=FILE_SHARE_READ|FILE_SHARE_WRITE;
  HANDLE hFile=NULL;
  BOOL b;
  DWORD ino=0;

  memset(&fi,0,sizeof fi);

  hFile=CreateFile(
    path->tcs,		// __in      LPCTSTR lpFileName,
    GENERIC_READ,	// __in      DWORD dwDesiredAccess,
    dwShareMode,	// __in      DWORD dwShareMode,
    NULL,			// __in_opt  LPSECURITY_ATTRIBUTES lpSecurityAttributes,
    OPEN_EXISTING,	// __in      DWORD dwCreationDisposition,
    0,				// __in      DWORD dwFlagsAndAttributes,
    NULL			// __in_opt  HANDLE hTemplateFile
  );

  if (NULL==hFile)
    goto cleanup;

  b=GetFileInformationByHandle(
    hFile,	// __in   HANDLE hFile,
  	&fi		// __out  LPBY_HANDLE_FILE_INFORMATION lpFileInformation
  );

  if (!b)
    goto cleanup;

  ino=fi.nFileIndexLow;
cleanup:
  if (NULL!=hFile)
    CloseHandle(hFile);

  return ino;
}
#endif

int pb_str_stat(const pb_str_t *path, struct stat *buf)
{
#if defined (WIN32)
  // http://pubs.opengroup.org/onlinepubs/007904975/functions/stat.html
  // http://msdn.microsoft.com/en-us/library/14h5k7ff%28v=vs.71%29.aspx
  int code;
  struct _stat wbuf;

  if (0==(code=_tstat(path->tcs,&wbuf))) {
    buf->st_mode=wbuf.st_mode;
    //buf->st_ino=wbuf.st_ino;
    buf->st_ino=get_ino_fake(path);
    buf->st_dev=wbuf.st_dev;
    buf->st_uid=wbuf.st_uid;
    buf->st_gid=wbuf.st_gid;
    buf->st_atime=wbuf.st_atime;
    buf->st_ctime=wbuf.st_ctime;
    buf->st_mtime=wbuf.st_mtime;
    buf->st_nlink=wbuf.st_nlink;
  }

  return code;
#else
  return stat(path->utf8,buf);
#endif
}
