#ifndef __PB_XY_H__
#define __PB_XY_H__

#include <stdarg.h>
#include "pb.h"

#if defined (ROLLBACK)
///////////////////////////////////////////////////////////////////////////////
int pb_str_size(TCHAR *ts);
int pb_str_copy(pb_str_t *ps, TCHAR *ts, int size);

///////////////////////////////////////////////////////////////////////////////
#if defined (WIN32) && defined (UNICODE)
typedef struct pb_y_size {
  int tcs;
  int utf8;
  int oem;
} pb_y_size_t;
#endif

typedef struct pb_y {
  int n;
#if defined (WIN32) && defined (UNICODE)
  pb_y_size_t size;
  pb_str_t prev;
  pb_str_t next;
#else
  int size;
  char *prev;
  char *next;
#endif
} pb_y_t;

void pb_y_init(pb_y_t *y);
void pb_y_close(pb_y_t *y);

void pb_y_count_part(pb_y_t *y, TCHAR *ts);
void pb_y_count_part_ps(pb_y_t *y, pb_str_t *pfx, pb_str_t *ps);
void pb_y_count_whole(pb_y_t *y);

void *pb_y_alloc(pb_y_t *y, int offs);

void pb_y_add_part(pb_y_t *y, TCHAR *ts);
void pb_y_add_part_ps(pb_y_t *y, pb_str_t *pfx, pb_str_t *ps);
void pb_y_add_whole(pb_y_t *y, pb_str_t *ps);

///////////////////////////////////////////////////////////////////////////////
typedef struct pb_x {
  pb_str_t **ps;
  pb_str_t *buf;
  int n;
  pb_y_t y;
} pb_x_t;

typedef void (*pb_x_next_t)(pb_x_t *x, TCHAR *ts, ...);
typedef void (*pb_x_next_ps_t)(pb_x_t *x, pb_str_t *ts, ...);

void pb_x_init(pb_x_t *x);
void pb_x_count(pb_x_t *x, TCHAR *ts, ...);
void pb_x_count_ps(pb_x_t *x, pb_str_t *ts, ...);
pb_str_t **pb_x_alloc(pb_x_t *x);
void pb_x_add(pb_x_t *x, TCHAR *ts, ...);
void pb_x_add_ps(pb_x_t *x, pb_str_t *ts, ...);
void pb_x_flush(pb_x_t *x);
#endif // ROLLBACK

///////////////////////////////////////////////////////////////////////////////
#if defined (WIN32) && defined (UNICODE)
typedef struct pb_z_size {
  int tcs;
  int utf8;
  int oem;
} pb_z_size_t;
#else
typedef int pb_z_size_t;
#endif

typedef struct pb_z {
  pb_z_size_t offs;
  pb_z_size_t size;
} pb_z_t;

void pb_z_init(pb_z_t *z, char *buf);
void pb_z_assign(pb_z_t *z, char *buf, pb_str_t *ps);

int pb_z_size(pb_z_t *z);
void pb_z_part(pb_z_t *z, TCHAR *s, char *buf);
void pb_z_part_ps(pb_z_t *z, pb_str_t *ps, char *buf);
void pb_z_part_ps_pfx(pb_z_t *z, pb_str_t *pfx, pb_str_t *ps, char *buf);
void pb_z_whole(pb_z_t *z, char *buf);

int pb_z_dup_pfx(pb_z_t *z, pb_str_t *pfx, pb_str_t *ts, pb_str_t *ps);
int pb_z_cat(pb_z_t *z, TCHAR *s, va_list ap, pb_str_t *ps);
int pb_z_cat_ps(pb_z_t *z, pb_str_t *ps1, va_list ap, pb_str_t *ps2);

///////////////////////////////////////////////////////////////////////////////
typedef struct pb_a {
  int n;
  int offs1;
  int offs2;
  pb_str_t **psp;
  pb_str_t *ps;
  char *buf;
  pb_z_t z;
} pb_a_t;

void pb_a_start(pb_a_t *a, pb_str_t **psp);
void pb_a_add(pb_a_t *a, TCHAR *s);
void pb_a_add_v(pb_a_t *a, TCHAR *s, ...);
void pb_a_add_ps(pb_a_t *a, pb_str_t *s);
int pb_a_finish(pb_a_t *a);

int pb_a_str_cre_n(pb_a_t *a, TCHAR **s, int n, pb_str_t **psp);
int pb_a_str_cre_n_ps(pb_a_t *a, pb_str_t **s, int n, pb_str_t **psp);

int pb_a_str_cre_v(pb_a_t *a, TCHAR **s, pb_str_t **psp);
int pb_a_str_cre_v_ps(pb_a_t *a, pb_str_t **s, pb_str_t **psp);

#endif // __PB_XY_H__
