#include <stdlib.h>
#include <string.h>
#include "pb_xy.h"

void pb_y_init(pb_y_t *y)
{
  memset(y,0,sizeof *y);
}

void pb_y_close(pb_y_t *y)
{
}

void pb_y_count_part(pb_y_t *y, TCHAR *ts)
{
#if defined (WIN32) && defined (UNICODE)
  y->size.utf8+=pb_ws2s_size(ts,CP_UTF8)-1;
  y->size.oem+=pb_ws2s_size(ts,CP_OEMCP)-1;
  y->size.tcs+=wcslen(ts);
#else
  y->size+=strlen(ts);
#endif
}

void pb_y_count_part_ps(pb_y_t *y, pb_str_t *pfx, pb_str_t *ps)
{
#if defined (WIN32) && defined (UNICODE)
  y->size.utf8+=strlen(ps->utf8);
  y->size.oem+=strlen(ps->oem);
  y->size.tcs+=wcslen(ps->tcs);

  if (NULL!=pfx) {
    y->size.utf8-=strlen(pfx->utf8);
    y->size.oem-=strlen(pfx->oem);
    y->size.tcs-=wcslen(pfx->tcs);
  }
#else
  y->size+=strlen(ps->utf8);

  if (NULL!=pfx)
    y->size-=strlen(pfx->utf8);
#endif
}

void pb_y_count_whole(pb_y_t *y)
{
#if defined (WIN32) && defined (UNICODE)
  ++y->size.utf8;
  ++y->size.oem;
  ++y->size.tcs;
#else
  ++y->size;
#endif
  ++y->n;
}

void *pb_y_alloc(pb_y_t *y, int offs)
{
  int size;
  char *buf;

#if defined (WIN32) && defined (UNICODE)
  size=offs+y->size.utf8+y->size.oem+y->size.tcs*sizeof(TCHAR);
#else
  size=offs+y->size;
#endif

  if (NULL==(buf=malloc(size)))
    return NULL;

#if defined (WIN32) && defined (UNICODE)
  y->prev.utf8=buf+offs;
  y->prev.oem=y->prev.utf8+y->size.utf8;
  y->prev.tcs=(TCHAR *)(y->prev.oem+y->size.oem);
  /*
  y->prev.tcs=(TCHAR *)buf+offs;
  y->prev.utf8=(char *)y->prev.tcs+y->size.utf8*sizeof(TCHAR);
  y->prev.oem=y->prev.utf8+y->size.utf8;
  */
#else
  y->prev=buf+offs;
#endif
  y->next=y->prev;

  return buf;
}

void pb_y_add_part(pb_y_t *y, TCHAR *ts)
{
  int len;

#if defined (WIN32) && defined (UNICODE)
  len=pb_ws2s_copy(y->next.utf8,ts,CP_UTF8,y->size.utf8)-1;
  y->size.utf8-=len;
  y->next.utf8+=len;

  len=pb_ws2s_copy(y->next.oem,ts,CP_OEMCP,y->size.oem)-1;
  y->size.oem-=len;
  y->next.oem+=len;

  len=wcslen(ts);
  memcpy(y->next.tcs,ts,(len+1)*sizeof *ts);
  y->size.tcs-=len*sizeof *ts;
  y->next.tcs+=len;
#else
  len=strlen(ts);
  memcpy(y->next,ts,len+1);
  y->size-=len;
  y->next+=len;
#endif
}

void pb_y_add_part_ps(pb_y_t *y, pb_str_t *pfx, pb_str_t *ps)
{
  int len1,len2;

#if defined (WIN32) && defined (UNICODE)
  len1=NULL==pfx?0:strlen(pfx->utf8);
  len2=strlen(ps->utf8)-len1;
  memcpy(y->next.utf8,ps->utf8+len1,len2+1);
  y->size.utf8-=len2;
  y->next.utf8+=len2;

  len1=NULL==pfx?0:strlen(pfx->oem);
  len2=strlen(ps->oem)-len1;
  memcpy(y->next.oem,ps->oem+len1,len2+1);
  y->size.oem-=len2;
  y->next.oem+=len2;

  len1=NULL==pfx?0:wcslen(pfx->tcs);
  len2=wcslen(ps->tcs)-len1;
  memcpy(y->next.tcs,ps->tcs+len1*sizeof(wchar_t),(len2+1)*sizeof(wchar_t));
  y->size.tcs-=len2*sizeof(wchar_t);
  y->next.tcs+=len2;
#else
  len1=NULL==pfx?0:strlen(pfx->utf8);
  len2=strlen(ps->utf8)-len1;
  memcpy(y->next,ps->utf8+len1,len2+1);
  y->size-=len2;
  y->next+=len2;
#endif
}

void pb_y_add_whole(pb_y_t *y, pb_str_t *ps)
{
#if defined (WIN32) && defined (UNICODE)
  ++y->next.tcs;
  ++y->next.utf8;
  ++y->next.oem;

  *ps=y->prev;
#else
  ++y->next;

  ps->utf8=y->prev;
  ps->oem=y->prev;
#if defined (WIN32)
  ps->tcs=y->prev;
#endif
#endif

  y->prev=y->next;
}
