#include <string.h>
#include <stdarg.h>
#include "pb_xy.h"

void pb_z_init(pb_z_t *z, char *buf)
{
#if defined (WIN32) && defined (UNICODE) // {
  if (NULL==buf) {
    z->offs.tcs=0;
    z->offs.utf8=0;
    z->offs.oem=0;

    z->size.tcs=0;
    z->size.utf8=0;
    z->size.oem=0;
  }
  else {
    z->offs.tcs=0;
    z->offs.utf8=z->offs.tcs+z->size.tcs;
    z->offs.oem=z->offs.utf8+z->size.utf8;
  }
#else // } {
  z->offs=0;

  if (NULL==buf)
	z->size=0;
#endif // }
}

void pb_z_assign(pb_z_t *z, char *buf, pb_str_t *ps)
{
#if defined (WIN32) && defined (UNICODE) // {
  if (NULL!=ps) {
	ps->tcs=(TCHAR *)(buf+z->offs.tcs);
	ps->utf8=buf+z->offs.utf8;
	ps->oem=buf+z->offs.oem;
  }
#else // } {
  if (NULL!=ps) {
#if defined (WIN32) // {
	ps->tcs=buf+z->offs;
#endif // }
	ps->utf8=buf+z->offs;
	ps->oem=buf+z->offs;
  }
#endif // }
}

int pb_z_size(pb_z_t *z)
{
#if defined (WIN32) && defined (UNICODE) // {
  return z->offs.oem+z->size.tcs+z->size.utf8+z->size.oem;
#else
  return z->offs+z->size;
#endif
}

void pb_z_part(pb_z_t *z, TCHAR *s, char *buf)
{
  pb_z_size_t n;
#if defined (WIN32) && defined (UNICODE) // {
  n.tcs=wcslen(s);

  if (NULL==buf) {
    n.utf8=pb_ws2s_size(s,CP_UTF8)-1;
    n.oem=pb_ws2s_size(s,CP_OEMCP)-1;
    z->size.tcs+=n.tcs*sizeof(TCHAR);
    z->size.utf8+=n.utf8;
    z->size.oem+=n.oem;
  }
  else {
	memcpy(buf+z->offs.tcs,s,n.tcs*sizeof(TCHAR));
    n.utf8=pb_ws2s_copy(buf+z->offs.utf8,s,CP_UTF8,z->size.utf8)-1;
    n.oem=pb_ws2s_copy(buf+z->offs.oem,s,CP_OEMCP,z->size.oem)-1;
    z->offs.tcs+=n.tcs*sizeof(TCHAR);
    z->offs.utf8+=n.utf8;
    z->offs.oem+=n.oem;
    z->size.tcs-=n.tcs*sizeof(TCHAR);
    z->size.utf8-=n.utf8;
    z->size.oem-=n.oem;
  }
#else // } {
  n=strlen(s);

  if (NULL==buf)
    z->size+=n;
  else {
	memcpy(buf+z->offs,s,n);
    z->offs+=n;
    z->size-=n;
  }
#endif // }
}

void pb_z_part_ps(pb_z_t *z, pb_str_t *ps, char *buf)
{
  pb_z_size_t n;
#if defined (WIN32) && defined (UNICODE) // {
  n.tcs=wcslen(ps->tcs);
  n.utf8=strlen(ps->utf8);
  n.oem=strlen(ps->oem);

  if (NULL==buf) {
    z->size.tcs+=n.tcs*sizeof(TCHAR);
    z->size.utf8+=n.utf8;
    z->size.oem+=n.oem;
  }
  else {
	memcpy(buf+z->offs.tcs,ps->tcs,n.tcs*sizeof(TCHAR));
	memcpy(buf+z->offs.utf8,ps->utf8,n.utf8);
	memcpy(buf+z->offs.oem,ps->oem,n.oem);
    z->offs.tcs+=n.tcs*sizeof(TCHAR);
    z->offs.utf8+=n.utf8;
    z->offs.oem+=n.oem;
    z->size.tcs-=n.tcs*sizeof(TCHAR);
    z->size.utf8-=n.utf8;
    z->size.oem-=n.oem;
  }
#else // } {
  n=strlen(ps->utf8);

  if (NULL==buf)
    z->size+=n;
  else {
	memcpy(buf+z->offs,ps->utf8,n);
    z->offs+=n;
    z->size-=n;
  }
#endif // }
}

void pb_z_part_ps_pfx(pb_z_t *z, pb_str_t *pfx, pb_str_t *ps, char *buf)
{
  pb_z_size_t offs;
  pb_z_size_t n;
#if defined (WIN32) && defined (UNICODE) // {
  offs.tcs=pfx?wcslen(pfx->tcs):0;
  offs.utf8=pfx?strlen(pfx->utf8):0;
  offs.oem=pfx?strlen(pfx->oem):0;

  n.tcs=wcslen(ps->tcs)-offs.tcs;
  n.utf8=strlen(ps->utf8)-offs.utf8;
  n.oem=strlen(ps->oem)-offs.oem;

  if (n.tcs<0)
    n.tcs=0;

  if (n.utf8<0)
    n.utf8=0;

  if (n.oem<0)
    n.oem=0;

  if (NULL==buf) {
    z->size.tcs+=n.tcs*sizeof(TCHAR);
    z->size.utf8+=n.utf8;
    z->size.oem+=n.oem;
  }
  else {
	memcpy(buf+z->offs.tcs,ps->tcs+offs.tcs,n.tcs*sizeof(TCHAR));
	memcpy(buf+z->offs.utf8,ps->utf8+offs.utf8,n.utf8);
	memcpy(buf+z->offs.oem,ps->oem+offs.oem,n.oem);
    z->offs.tcs+=n.tcs*sizeof(TCHAR);
    z->offs.utf8+=n.utf8;
    z->offs.oem+=n.oem;
    z->size.tcs-=n.tcs*sizeof(TCHAR);
    z->size.utf8-=n.utf8;
    z->size.oem-=n.oem;
  }
#else // } {
  offs=pfx?strlen(pfx->utf8):0;
  n=strlen(ps->utf8)-offs;

  if (n<0)
    n=0;

  if (NULL==buf)
    z->size+=n;
  else {
	memcpy(buf+z->offs,ps->utf8+offs,n);
    z->offs+=n;
    z->size-=n;
  }
#endif // }
}

void pb_z_whole(pb_z_t *z, char *buf)
{
#if defined (WIN32) && defined (UNICODE) // {
  if (NULL==buf) {
    z->size.tcs+=sizeof(TCHAR);
    ++z->size.utf8;
    ++z->size.oem;
  }
  else {
	*((TCHAR *)(buf+z->offs.tcs))=0;
	*(buf+z->offs.utf8)=0;
	*(buf+z->offs.oem)=0;
    z->offs.tcs+=sizeof(TCHAR);
    ++z->offs.utf8;
    ++z->offs.oem;
    z->size.tcs-=sizeof(TCHAR);
    --z->size.utf8;
    --z->size.oem;
  }
#else // } {
  if (NULL==buf)
    ++z->size;
  else {
	*(buf+z->offs)=0;
    ++z->offs;
    --z->size;
  }
#endif // }
}

int pb_z_cat(pb_z_t *z, TCHAR *s, va_list ap, pb_str_t *ps)
{
  TCHAR *x;
  int offs=sizeof *ps;
  char *buf=ps?(char *)ps+offs:NULL;

  pb_z_init(z,buf);
  pb_z_assign(z,buf,ps);

  for (x=s;x;x=va_arg(ap,TCHAR *))
    pb_z_part(z,x,buf);

  pb_z_whole(z,buf);

  return offs+pb_z_size(z);
}

int pb_z_cat_ps(pb_z_t *z, pb_str_t *ps1, va_list ap, pb_str_t *ps2)
{
  pb_str_t *x;
  int offs=sizeof *ps2;
  char *buf=ps2?(char *)ps2+offs:NULL;

  pb_z_init(z,buf);
  pb_z_assign(z,buf,ps2);

  for (x=ps1;x;x=va_arg(ap,pb_str_t *))
    pb_z_part_ps(z,x,buf);

  pb_z_whole(z,buf);

  return offs+pb_z_size(z);
}

int pb_z_dup_pfx(pb_z_t *z, pb_str_t *pfx, pb_str_t *ps1, pb_str_t *ps2)
{
  int offs=sizeof *ps2;
  char *buf=ps2?(char *)ps2+offs:NULL;

  pb_z_init(z,buf);
  pb_z_assign(z,buf,ps2);
  pb_z_part_ps_pfx(z,pfx,ps1,buf);
  pb_z_whole(z,buf);

  return offs+pb_z_size(z);
}
